
/*
 cc -g -c imgdbs_mysql.c && mv imgdbs_mysql.o imgdbs.o
*/


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include <mysql/mysql.h>
#include <mysql/errmsg.h>

#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"

#include "imgsamp.h"
#include "imgexempt.h"

#include "imgdb.h"
#include "imgdbs_mysql.h"


/* ------------------------- ImgdbS mysql wrapper ----------------- */


int Imgdbs_mysql_init(imdbs,flag)
struct ImgdbS *imdbs;
int flag;
{
 imdbs->db= mysql_init(NULL);
 if(imdbs->db==NULL) {
   strcpy(imdbs->imdb->errmsg,"mysql_init() failed. Probably out of memory.");
   return(-1);
 }
 return(1);
}


int Imgdbs_mysql_handle_error(imdbs,flag)
struct ImgdbS *imdbs;
int flag;
/*
 bit2= copy mysql_error() to imdbs->imdb->errmsg
*/
{
 int ret,siz;
 char *msg;
 unsigned int merrno;

 merrno= mysql_errno(imdbs->db);
 if(flag&2) {
   msg= mysql_error(imdbs->db);
   siz= sizeof(imdbs->imdb->errmsg);
   if(strlen(msg)>=siz) {
     strncpy(imdbs->imdb->errmsg,msg,siz-1);
     imdbs->imdb->errmsg[siz-1]= 0;
   } else
     strcpy(imdbs->imdb->errmsg,msg);
 }

 if(merrno==CR_SERVER_GONE_ERROR || merrno==CR_SERVER_LOST) {

   fprintf(stderr,"%s\n... trying to reconnect ...\n",mysql_error(imdbs->db));

   /* shut down */;
   Imgdbs_mysql_free_result(imdbs,0);
   Imgdbs_mysql_close(imdbs,0);
   /* bring up again */;
   ret= Imgdbs_mysql_init(imdbs,0);
   if(ret<=0)
     return(ret);
   /* try to reconnect */;
   ret= Imgdbs_mysql_open(imdbs,1);
   if(ret<=0)
     return(ret);
   return(1);
 }

 return(0);
}


int Imgdbs_mysql_open(imdbs,flag)
struct ImgdbS *imdbs;
int flag;
/*
 bit0= no Imgdbs_mysql_handle_error() in case of error
*/
{
 int retry_count= 0,ret;
 MYSQL *myret;

retry:;
 myret= mysql_real_connect(imdbs->db,imdbs->hostname,
                           imdbs->username,imdbs->password,
                           imdbs->dbname,0,NULL,0);
 if(myret==NULL) {
   if(!flag&1) {
     ret= Imgdbs_mysql_handle_error(imdbs,0);
     if(ret==1) {
       if(retry_count++<=Imgdbs_reconnect_maxtrY)
         goto retry;
     }     
   }     
   sprintf(imdbs->imdb->errmsg,"connection to MySQL failed : %s %s %s \n %s",
                               imdbs->hostname,imdbs->username,imdbs->dbname,
                               mysql_error(imdbs->db));
   return(-1);
 }
 return(1);
}


int Imgdbs_mysql_close(imdbs,flag)
struct ImgdbS *imdbs;
int flag;
{
 mysql_close(imdbs->db);
 imdbs->db= NULL;
 return(1);
}


int Imgdbs_mysql_query(imdbs,query_text,query_length,flag)
struct ImgdbS *imdbs;
char *query_text;
int query_length;
int flag;
{
 int retry_count= 0,ret;
 int qret;

 if(imdbs->opened==0) {
   ret= Imgdbs_open(imdbs,0);
   if(ret<=0)
     return(ret);
 }
retry:;
 qret= mysql_real_query(imdbs->db,query_text,query_length);
 if(qret!=0) {
   ret= Imgdbs_mysql_handle_error(imdbs,2);
   if(ret==1) {
     if(retry_count++<=Imgdbs_reconnect_maxtrY)
       goto retry;
   }     
   return(0);
 }
 return(1);
}


int Imgdbs_mysql_request_result(imdbs,flag)
struct ImgdbS *imdbs;
int flag;
/*
 bit4= buffer whole reply so query can be left open while intermediate query
*/
{
 int retry_count= 0,ret;

retry:;
 if(imdbs->fetch_mode==0 || (flag&16)) {
   imdbs->query_result= mysql_store_result(imdbs->db);
 } else {
   imdbs->query_result= mysql_use_result(imdbs->db);
 }
 if(imdbs->query_result==NULL) {
   Imgdbs_mysql_handle_error(imdbs,2);
   return(0);
 }
 return(1);
}


int Imgdbs_mysql_esc_string(imdbs,target,source,source_length,flag)
struct ImgdbS *imdbs;
char *target;
char *source;
int source_length;
int flag;
{
 int l;

 l=  mysql_escape_string(target,source,source_length);
 return(l);
}


int Imgdbs_mysql_end_query(imdbs,flag)
struct ImgdbS *imdbs;
int flag;
{
 int ret;
 MYSQL_ROW row;

 do {
   row= mysql_fetch_row(imdbs->query_result);
 } while(row!=NULL);
 if(mysql_errno(imdbs->db)!=0) {
   ret= Imgdbs_mysql_handle_error(imdbs,2);
   if(ret!=1)
     return(0);
 }
 return(1);
}


int Imgdbs_mysql_free_result(imdbs,flag)
struct ImgdbS *imdbs;
int flag;
{
 if(imdbs->query_result==NULL)
   return(0);
 mysql_free_result(imdbs->query_result);
 imdbs->query_result= NULL;
 return(1);
}


int Imgdbs_mysql_unique_seed(imdbs,seed,flag)
/* produces a unique id or a seed value which is converted to a unique id
   by the RDB server */
struct ImgdbS *imdbs;
unsigned long *seed;
int flag;
{
 *seed= 0;
 return(1);
}


int Imgdbs_mysql_get_unique_id(imdbs,id,flag)
/* reads the actual unique id after the RDB server processed the seed */
struct ImgdbS *imdbs;
unsigned long *id;
int flag;
{
 /* >>> this only works up to 32 bit, the MySQl type is 64 bit as it seems */
 /* >>> how to use 64 bit in a portable way ? */
 *id= mysql_insert_id(imdbs->db);
 return(1);
}


int Imgdbs_mysql_get_fieldcount(imdbs,flag)
struct ImgdbS *imdbs;
int flag;
{
 int fieldcount;

 fieldcount= mysql_num_fields(imdbs->query_result);
 return(fieldcount);
}


int Imgdbs_mysql_get_row(imdbs,fields,lengths,flag)
struct ImgdbS *imdbs;
char ***fields;
unsigned long **lengths;
int flag;
{
 int retry_count= 0,ret;
 unsigned int *fieldlengths;
 MYSQL_ROW row;

retry:;
 row= mysql_fetch_row(imdbs->query_result);
 if(row==NULL) {
   if(mysql_errno(imdbs->db)!=0) {
     Imgdbs_mysql_handle_error(imdbs,2);
     return(-1);
   }
   return(0);
 }
 *lengths= mysql_fetch_lengths(imdbs->query_result);
 *fields= (char **) row;
 return(1);
}


/* ------------------------- ImgdbS general ----------------------- */


int Imgdbs_new(imdbs,imdb,hostname,username,password,
                          dbname,sample_table,sample_format,exempt_table,flag)
struct ImgdbS **imdbs;
struct ImgdB *imdb;
char *hostname,*username,*password;
char *dbname;                                   /* proposal : "similar_db" */
char *sample_table;
int sample_format;
char *exempt_table;
int flag;
{
 int ret;
 struct ImgdbS *m;

 *imdbs= m= TSOB_FELD(struct ImgdbS,1);
 if(*imdbs==NULL)
   return(-1);
 m->imdb= imdb;
 m->hostname= NULL;
 m->username= NULL;
 m->password= NULL;
 m->dbname= NULL;
 m->db= NULL;
 m->fetch_mode= 1;
 m->opened= 0;
 m->has_similar_udf= 0;
 m->has_color_avg= 1;
 m->sample_format= sample_format;
 m->sample_table= NULL;
 strcpy(m->id_fieldname,"id");
 strcpy(m->adr_fieldname,"name");
 strcpy(m->sample_fieldname,"smp17");
 m->exempt_table= NULL;
 strcpy(m->eid_fieldname,"eid");
 strcpy(m->eboss_fieldname,"eboss");
 strcpy(m->eadr_fieldname,"ename");
 m->maxeid[0]= 0;
 m->eid_list= NULL;
 m->eid_count= 0;
 m->eid_readpos= 0;
 m->query_result= NULL;
 m->query_token= NULL;
 m->search_where= NULL;
 m->search_result= NULL;
 m->search_token= NULL;

 if(Sregex_string(&(m->hostname),hostname,0)<=0)
   goto failed;
 if(Sregex_string(&(m->username),username,0)<=0)
   goto failed;
 if(Sregex_string(&(m->password),password,0)<=0)
   goto failed;
 if(Sregex_string(&(m->dbname),dbname,0)<=0)
   goto failed;
 if(Sregex_string(&(m->sample_table),sample_table,0)<=0)
   goto failed;
 if(Sregex_string(&(m->exempt_table),exempt_table,0)<=0)
   goto failed;
 return(1);
failed:;
 Imgdbs_destroy(imdbs,0);
 return(-1);
}


int Imgdbs_destroy(imdbs,flag)
struct ImgdbS **imdbs;
int flag;
{
 struct ImgdbS *m;
 
 m= *imdbs;
 if(m==NULL)
   return(0);
 Imgdbs_close(m,0);
 Sregex_string(&(m->hostname),NULL,0);
 Sregex_string(&(m->username),NULL,0);
 Sregex_string(&(m->password),NULL,0);
 Sregex_string(&(m->dbname),NULL,0);
 Sregex_string(&(m->sample_table),NULL,0);
 Sregex_string(&(m->exempt_table),NULL,0);
 Sregex_string(&(m->search_where),NULL,0);
 Smem_freE((char *) m);
 *imdbs= NULL;
 return(1);
}


char *Imgdbs__dbtype(flag)
int flag;
{
 return(Imgdbs_dbtypE);
}


int Imgdbs_read_conf(imdbs_return,imdb,names,contents,count,flag)
struct ImgdbS **imdbs_return;
struct ImgdB *imdb;
char **names;
char **contents;
int count;
int flag;
{
 int ret,i;
 char hostname[SfileadrL],username[SfileadrL],password[SfileadrL];
 char dbname[SfileadrL],filename[SfileadrL];
 char sample_tablename[SfileadrL],exempt_tablename[SfileadrL];
 char varname[81],text[81];
 struct ImgdbS *imdbs;

 varname[0]= 0;

 strcpy(varname,"hostname");
 ret= Sfile_find_var(varname,names,contents,count,hostname,sizeof(hostname),0);
 if(ret<=0)
   goto ex;
 strcpy(varname,"username");
 ret= Sfile_find_var(varname,names,contents,count,username,sizeof(username),0);
 if(ret<=0)
   goto ex;
 strcpy(varname,"dbname");
 ret= Sfile_find_var(varname,names,contents,count,dbname,sizeof(dbname),0);
 if(ret<=0)
   goto ex;
 varname[0]= 0;

 ret= Sfile_find_var("password",names,contents,count,password,
                     sizeof(password),0);
 if(ret<0)
   goto ex;
 /* if password is empty : look for password file */
 if(password[0]==0) {
   ret= Imgdb_get_mapname(imdb,filename,8);
   if(ret<=0)
     goto ex;
   ret= Sfile_file_to_text(filename,password,sizeof(password),1|2);
   if(ret<=0)
     goto ex;
   if(ret==2)
     password[0]= 0;
 }
 strcpy(sample_tablename,"similar_samples");
 ret= Sfile_find_var("sample_tablename",names,contents,count,
                     sample_tablename,sizeof(sample_tablename),2);
 if(ret<0)
   goto ex;
 strcpy(exempt_tablename,"similar_exempts");
 ret= Sfile_find_var("exempt_tablename",names,contents,count,
                     exempt_tablename,sizeof(exempt_tablename),2);
 if(ret<0)
   goto ex;

 /* try wether there is a SQL adapter at all */
 ret= Imgdbs_new(imdbs_return,imdb,hostname,username,password,dbname,
                 sample_tablename,3,exempt_tablename,0);
 for(i=0;i<SfileadrL;i++)
   password[i]= 0;
 if(ret<=0)
   goto ex;

 /* set special server and database properties */
 ret= Sfile_find_var("has_similar_udf",names,contents,count,
                     text,sizeof(text),2);
 if(ret<0)
   goto ex;
 if(ret>0 && (strcmp(text,"yes")==0 || strcmp(text,"true")==0
              || strcmp(text,"1")==0 || strcmp(text,"on")==0))
   (*imdbs_return)->has_similar_udf= 1;
 ret= Sfile_find_var("has_color_avg",names,contents,count,
                     text,sizeof(text),2);
 if(ret<0)
   goto ex;
 if(ret>0 && (strcmp(text,"no")==0 || strcmp(text,"false")==0
              || strcmp(text,"0")==0 || strcmp(text,"off")==0))
   (*imdbs_return)->has_color_avg= 0;
 ret= Sfile_find_var("write_native",names,contents,count,
                     text,sizeof(text),2);
 if(ret<0)
   goto ex;
 (*imdbs_return)->sample_format= 3;
 if(ret>0 && (strcmp(text,"no")==0 || strcmp(text,"false")==0
              || strcmp(text,"0")==0 || strcmp(text,"off")==0))
   (*imdbs_return)->sample_format= 0;
 if(ret>0 && strcmp(text,"swapped")==0)
   (*imdbs_return)->sample_format= 4;

 /* eventually adjust column names */
 imdbs= *imdbs_return;
 ret= Sfile_find_var("sample_idname",names,contents,count,
                     imdbs->id_fieldname,sizeof(imdbs->id_fieldname),2);
 if(ret<0)
   goto ex;
 ret= Sfile_find_var("sample_adrname",names,contents,count,
                     imdbs->adr_fieldname,sizeof(imdbs->adr_fieldname),2);
 if(ret<0)
   goto ex;
 ret= Sfile_find_var("sample_dataname",names,contents,count,
                     imdbs->sample_fieldname,sizeof(imdbs->sample_fieldname),2);
 if(ret<0)
   goto ex;
 ret= Sfile_find_var("exempt_idname",names,contents,count,
                     imdbs->eid_fieldname,sizeof(imdbs->eid_fieldname),2);
 if(ret<0)
   goto ex;
 ret= Sfile_find_var("exempt_bossname",names,contents,count,
                     imdbs->eboss_fieldname,sizeof(imdbs->eboss_fieldname),2);
 if(ret<0)
   goto ex;
 ret= Sfile_find_var("exempt_adrname",names,contents,count,
                     imdbs->eadr_fieldname,sizeof(imdbs->eadr_fieldname),2);
 if(ret<0)
   goto ex;

 ret= 1;
ex:;
 if(ret==0 && varname[0]!=0) {
   sprintf(imdb->errmsg,"Required variable missing or content too long :  %s",
                        varname); 
 }
 return(ret);
}


int Imgdbs_copy(source,target,flag)
struct ImgdbS *source,*target;
int flag;
{
 int ret;

 Imgdbs_close(target,0);

 if(Sregex_string(&(target->hostname),source->hostname,0)<=0)
   return(-1);
 if(Sregex_string(&(target->username),source->username,0)<=0)
   return(-1);
 if(Sregex_string(&(target->password),source->password,0)<=0)
   return(-1);
 if(Sregex_string(&(target->dbname),source->dbname,0)<=0)
   return(-1);
 if(Sregex_string(&(target->sample_table),source->sample_table,0)<=0)
   return(-1);
 strcpy(target->id_fieldname,source->id_fieldname);
 strcpy(target->adr_fieldname,source->adr_fieldname);
 strcpy(target->sample_fieldname,source->sample_fieldname);
 if(Sregex_string(&(target->exempt_table),source->exempt_table,0)<=0)
   return(-1);
 strcpy(target->eid_fieldname,source->eid_fieldname);
 strcpy(target->eboss_fieldname,source->eboss_fieldname);
 strcpy(target->eadr_fieldname,source->eadr_fieldname);
 strcpy(target->maxeid,source->maxeid);
 if(Sregex_string(&(target->search_where),source->search_where,0)<=0)
   return(-1);
 if(Imgdbs_is_open(source,0)){
   ret= Imgdbs_open(target,0);
   if(ret<=0)
     return(ret);
 }
 return(1);
}


int Imgdbs_open(imdbs,flag)
struct ImgdbS *imdbs;
int flag;
/*
 >>> bit14= create table if necessary
*/
{
 int ret;

 ret= Imgdbs_mysql_init(imdbs,0);
 if(ret<=0)
   {ret= -1; goto ex;}
 ret= Imgdbs_mysql_open(imdbs,0);
 if(ret<=0)
   {ret= -1; goto ex;}

 /* >>> check wether table is available and of usable format */

 imdbs->opened= 1;
 ret= 1;
ex:;
 return(ret);
}


int Imgdbs_close(imdbs,flag)
struct ImgdbS *imdbs;
int flag;
{
 if(imdbs->opened==0) 
   return(2);
 Imgdbs_end_query(imdbs,1);
 Imgdbs_end_query(imdbs,0);
 if(imdbs->eid_list!=NULL)
   Smem_freE(imdbs->eid_list);
 Imgdbs_mysql_close(imdbs,0);
 imdbs->opened= 0;
 return(1);
}


int Imgdbs_is_open(imdbs,flag)
struct ImgdbS *imdbs;
int flag;
{
 return(imdbs->opened);
}


int Imgdbs_db_exists(imdbs,flag)
struct ImgdbS *imdbs;
int flag;
{
 int ret;
 char where[81],what[81],*result= NULL;

 /* SELECT COUNT($id) FROM $table */
 sprintf(what,"COUNT(%s)",imdbs->id_fieldname);
 where[0]= 0;
 ret= Imgdbs_simple_select(imdbs,what,imdbs->sample_table,where,&result,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:
 Sregex_string(&result,NULL,0);
 return(ret);
}


int Imgdbs_report_query(imdbs,pretext,query_text,flag)
struct ImgdbS *imdbs;
char *pretext;
char *query_text;
int flag;
{
 int size,ret;
 char *buff= NULL;

 if(query_text==NULL) {
   if(Sregex_string(&buff,"<NULL>",0)<=0)
     {ret= -1; goto ex;}
 } else {
   size= strlen(query_text)+1;
   buff= (char *) Smem_malloC(size);
   if(buff==NULL)
     {ret= -1; goto ex;}
   strcpy(buff,query_text);
   Sfile_esc_filename(buff,sizeof(buff),64);
 }
 fprintf(stderr,"%s%s\n",pretext,buff);
 ret= 1;
ex:;
 Sregex_string(&buff,NULL,0);
 return(ret);
}


int Imgdbs_issue_query(imdbs,query_text,query_length,flag)
struct ImgdbS *imdbs;
char *query_text;
int query_length;
int flag;
/*
 bit0= this is a query which finally needs to be closed (i.e. SELECT)
 bit4= with bit0: buffer whole reply so query can be left open
                  while intermediate query
*/
{
 int qret,ret;
 char *msg;

 Imgdbs_end_query(imdbs,0);

 if(imdbs->imdb->debug_mode>0)
   Imgdbs_report_query(imdbs,"------ Query : ",query_text,0);

 ret= Imgdbs_mysql_query(imdbs,query_text,query_length,0);
 if(ret<=0)
   goto ex;
 if(flag&1) {
   ret= Imgdbs_mysql_request_result(imdbs,flag&16);
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:;
 if((flag&1) && ret>0) {
   if(Sregex_string(&(imdbs->query_token),"imgdbs_query: ",0)<=0)
     {ret= -1; goto ex;}
   if(Sregex_string(&(imdbs->query_token),query_text,1)<=0)
     {ret= -1; goto ex;}
 }
 return(ret);
}


int Imgdbs_preserve_searchpos(imdbs,flag)
struct ImgdbS *imdbs;
int flag;
{
 if(imdbs->search_result!=NULL & imdbs->search_result!=imdbs->query_result)
   Imgdbs_end_query(imdbs,1);
 imdbs->search_result= imdbs->query_result;
 imdbs->search_token= imdbs->query_token;
 return(1);
}


int Imgdbs_activate_searchpos(imdbs,flag)
struct ImgdbS *imdbs;
int flag;
{
 int ret;
 if(imdbs->search_result!=imdbs->query_result)
   Imgdbs_end_query(imdbs,0);
 imdbs->query_result= imdbs->search_result;
 imdbs->query_token= imdbs->search_token;
 return(1);
}


int Imgdbs_make_udf_search(imdbs,samp,result,result_size,flag)
/* make search expression using custom SQL function SIMILAR_IMG */
struct ImgdbS *imdbs;
struct ImgsamplE *samp;
char *result;
int result_size;
int flag;
{
 int ret,l= 0,search_method,max_misses;
 double eps;
 char *samp_text= NULL;
 struct ImgmatcH *matcher;

 /* SIMILAR_IMG( $samplename , '$samp' ,
                     tolerance,tolerance,ignore,ignore, 'method_number') */;
 if(result_size<=l+16+strlen(imdbs->sample_fieldname))
   {ret= -1; goto ex;}
 sprintf(result+l,"SIMILAR_IMG(%s,",imdbs->sample_fieldname);
 l+= strlen(result+l);
 ret= Imgdbs_format_samp(imdbs,samp,&samp_text,8);
 if(ret<=0)
   goto ex;
 if(result_size<=l+16+strlen(imdbs->sample_fieldname))
   {ret= -1; goto ex;}
 strcat(result+l,samp_text);
 l+= strlen(result+l);
 if(result_size<=l+2*32+3*16)
   {ret= -1; goto ex;}
 ret= Imgmap_get_matcher(imdbs->imdb->imp,&matcher,0);
 if(ret<=0)
   goto ex;
 Imgmatch_choose_tolerances(matcher,samp,0);
 Imgmatch_get_searchpar(matcher,&search_method,&eps,&max_misses,0);
 sprintf(result+l,",%lf,%lf,%ld,%ld,'%ld')=1 ",
                  eps,eps,max_misses,max_misses,search_method);
 l+= strlen(result+l);
 ret= 1;
ex:
 Sregex_string(&samp_text,NULL,0);
 return(ret); 
}


int Imgdbs_init_search_query(imdbs,samp,name,flag)
struct ImgdbS *imdbs;
struct ImgsamplE *samp;
char *name;
int flag;
/*
 bit0= do not issue query yet (probably never issue it)
 bit1= sample contains image data for possible pre-selection
 bit2= name contains fileaddress for possible pre-selection
 bit4= prepare for possible intermediate queries
*/
{
 int ret,l=0,i;
 double minmax[6];
 char text[2*SfileadrL+1024];
 static char colors[3][8]= {"red","green","blue"};

 Imgdbs_end_query(imdbs,0);

 /* preset selection name or sample average */;
 Sregex_string(&(imdbs->search_where),NULL,0);
 l= 0;
 text[0]= 0;
 if(flag&2) {

   if(imdbs->has_similar_udf) {

     /* use custom SQL function SIMILAR_IMG */;
     if(l>0) {
       strcat(text,"AND ");
       l+= 4;
     }
     ret= Imgdbs_make_udf_search(imdbs,samp,text+l,sizeof(text)-l,0);
     if(ret<=0)
       goto ex;

   } else if(imdbs->has_color_avg) {
     ret= Imgmap_avg_limits(imdbs->imdb->imp,samp,minmax,0);
     if(ret<=0)
       goto ex;
     for(i=0;i<6;i++) {
       if(minmax[i]>0 && minmax[i]<Imgsample_max_coorD/Imgsamp_scalE) {
         if(l>0) {
           strcat(text,"AND ");
           l+= 4;
         }
         sprintf(text+l,"%s_%s %s %ld ",imdbs->sample_fieldname,colors[i/2],
                                        (i%2 ? "<=" : ">=" ), 
                                        (int) (minmax[i]*Imgsamp_scalE));
         l+= strlen(text+l);
       }
     }
   }
 }
 if(flag&4) {
   if(l>0) {
     strcat(text,"AND ");
     l+= 4;
   }
   sprintf(text+l,"%s = '",imdbs->adr_fieldname);
   l+= strlen(text+l);
   l+= Imgdbs_mysql_esc_string(imdbs,text+l,name,strlen(name),0);
   text[l++]= '\'';
   text[l]= 0;
 }
 if(flag&(2|4) && strlen(text)>0) {
   if(Sregex_string(&(imdbs->search_where)," WHERE ",0)<=0)
     {ret= -1; goto ex;}
   if(Sregex_string(&(imdbs->search_where),text,1)<=0)
     {ret= -1; goto ex;}
 }
 if(!(flag&1)) {
   ret= Imgdbs_read_samp(imdbs,samp,2|4|(flag&16));
   if(ret<=0)
     goto ex;
 }

 Imgdbs_preserve_searchpos(imdbs,0);
 ret= 1;
ex:;
 return(ret);
}


int Imgdbs_end_query(imdbs,flag)
struct ImgdbS *imdbs;
int flag;
/*
 bit0= delete search_query
*/
{
 MYSQL_RES *query_result_mem= NULL;
 char *query_token_mem= NULL;
 int ret;

 if(imdbs->query_token==NULL)
   return(2);

 if((flag&1)) {
   if(imdbs->query_result!=imdbs->search_result) {
     query_result_mem= imdbs->query_result;
     query_token_mem= imdbs->query_token;
     imdbs->query_result= imdbs->search_result;
     imdbs->query_token= imdbs->search_token;
   }
   imdbs->search_result= NULL;
   imdbs->search_token= NULL;
 }
 if(imdbs->query_result!=NULL) {
   if(imdbs->query_result==imdbs->search_result) {
     if(!(flag&1)) { /* dummy end of map search query */
       imdbs->query_result= NULL;
       imdbs->query_token= NULL;
       ret= 1; goto ex;
     }
   }
   ret= Imgdbs_mysql_end_query(imdbs,0);
   if(ret<=0)
     goto ex;
   Imgdbs_mysql_free_result(imdbs,0);
   if(imdbs->imdb->debug_mode>1)
     Imgdbs_report_query(imdbs,"------ ended : ",imdbs->query_token,0);
 }
 ret= 1;
ex:;
 if(imdbs->imdb->debug_mode>2)
   if(imdbs->query_token!=NULL)
     Imgdbs_report_query(imdbs,"------ token removed : ",imdbs->query_token,0);
 Sregex_string(&(imdbs->query_token),NULL,0);
 if(query_result_mem!=NULL) {
   imdbs->query_result= query_result_mem;
   imdbs->query_token= query_token_mem;
 }
 return(ret);
}


int Imgdbs_format_nam(imdbs,filename,result,flag)
struct ImgdbS *imdbs;
char *filename;
char **result;          /* must be NULL or a valid freeable memory address */
int flag;
/*
 bit0= append to result rather than overwrite it
*/
{
 char fileadr[2*SfileadrL];
 int l;
 
 l= 0;
 fileadr[l++]= '\'';
 l+= Imgdbs_mysql_esc_string(imdbs,fileadr+l,filename,strlen(filename),0);
 fileadr[l++]= '\'';
 fileadr[l]= 0;
 if(Sregex_string(result,fileadr,flag&1)<=0)
   return(-1);
 return(1);
}


int Imgdbs_format_intcoord(imdbs,colorname,index,coord,buff,buffl,flag)
struct ImgdbS *imdbs;
char *colorname;      /* "red", "green" or "blue" */
int index;
Imgsamp_coorD coord;
char *buff;         /* must provide at least 81 bytes memory */
int *buffl;
int flag;
/*
 bit0= prepend field names and '=' (for UPDATE)
 bit1= do only produce a list of fieldnames 
 bit2= do not prepend index
*/
{
 int l;

 l= *buffl; 
 if(flag&1) {
   if(flag&4) 
     sprintf(buff+l,"%s_%s",imdbs->sample_fieldname,colorname);
   else
     sprintf(buff+l,"%s_%s_%ld",imdbs->sample_fieldname,colorname,index);
   l+= strlen(buff+l);
   if(!(flag&2))
     buff[l++]= '=';
 }
 if(!(flag&2)) {
   sprintf(buff+l,"%lu",coord);
   l+= strlen(buff+l);
 }
 *buffl= l;
 return(1);
}


int Imgdbs_format_samp(imdbs,samp,result,flag)
struct ImgdbS *imdbs;
struct ImgsamplE *samp;
char **result;          /* must be NULL or a valid freeable memory address */
int flag;
/*
 bit0= prepend field names and '=' (for UPDATE)
 bit1= do only produce a list of fieldnames 
 bit2= append to result rather than replace it
 bit3= under no circumstances append the three average values
*/
{
 char std[2+12+(Imgsamp_max_sizE+1)*3*sizeof(Imgsamp_coorD)+1];
 char buff[10*3*(Imgsamp_max_sizE+1)+3*(Imgsamp_max_sizE+1)*80+8];
 char revision[8];
 int i,j,k,coord_size,l,b,smp_size=0,avg[3],stdl,ret,hflag;

 coord_size= sizeof(Imgsamp_coorD);
 if(flag&2)
   flag|= 1;

 if(!(flag&2)) {
   if(imdbs->sample_format==0)
     hflag= 0;
   else if(imdbs->sample_format==3)
     hflag= 2;
   else if(imdbs->sample_format==4)
     hflag= 2|4;
   else
     hflag= 1;
   ret= Imgsample_smp17_format(samp,std,&smp_size,avg,hflag);
   if(ret<=0)
     return(ret);
 }
 l= 0;
 buff[l]= 0;
 if(imdbs->sample_format==0 || imdbs->sample_format==3 ||
    imdbs->sample_format==4) {
   if(flag&1) {
     sprintf(buff,"%s",imdbs->sample_fieldname);
     l= strlen(buff);
     if(!(flag&2))
       buff[l++]= '=';
   }
   if(!(flag&2)) {
     buff[l++]= '\'';
     l+= Imgdbs_mysql_esc_string(imdbs,buff+l,std,smp_size,0);
     buff[l++]= '\'';
   }
   buff[l]= 0;
 } else if(imdbs->sample_format==1) {
   for(i=0;i<samp->spoints;i++) {
     if(i>0)
       buff[l++]= ',';
     if(flag&1) {
       sprintf(buff,"%s_%ld",imdbs->sample_fieldname,i);
       l= strlen(buff);
       if(!(flag&2))
         buff[l++]= '=';
     }
     if(!(flag&2)) {
       buff[l++]= '\'';
       l+= Imgdbs_mysql_esc_string(imdbs,buff+l,std+i*3*coord_size,
                                   3*coord_size,0);
       buff[l++]= '\'';
     }
   }
   buff[l]= 0;
 } else if(imdbs->sample_format==2) {
   for(i=0;i<samp->spoints;i++) {
     if(i>0)
       buff[l++]= ',';
     Imgdbs_format_intcoord(imdbs,"red",i,samp->red[i],buff,&l,flag&3);
     buff[l++]= ',';
     Imgdbs_format_intcoord(imdbs,"green",i,samp->green[i],buff,&l,flag&3);
     buff[l++]= ',';
     Imgdbs_format_intcoord(imdbs,"blue",i,samp->blue[i],buff,&l,flag&3);
     buff[l]= 0;
   }
 } else {
   sprintf(imdbs->imdb->errmsg,"PROGRAM ERROR : unknown sample_format %ld",
                               imdbs->sample_format);
   return(-1);
 }

 if(imdbs->has_color_avg &&!(flag&8)) {
   /* append average pixel as 3 integer columns */
   buff[l++]= ',';
   Imgdbs_format_intcoord(imdbs,"red",0,(Imgsamp_coorD) avg[0],buff,&l,
                          4|(flag&3));
   buff[l++]= ',';
   Imgdbs_format_intcoord(imdbs,"green",0,(Imgsamp_coorD) avg[1],buff,&l,
                          4|(flag&3));
   buff[l++]= ',';
   Imgdbs_format_intcoord(imdbs,"blue",0,(Imgsamp_coorD) avg[2],buff,&l,
                          4|(flag&3));
 }

 if(Sregex_string(result,buff,(flag>>2)&1)<=0)
   return(-1);
 return(1);
}


int Imgdbs_deformat_bincoord(imdbs,std,coord,flag)
struct ImgdbS *imdbs;
char *std;
Imgsamp_coorD *coord;
int flag;
{
 int b;

 *coord= 0;
 for(b=0;b<sizeof(Imgsamp_coorD);b++) {
   (*coord)<<= 8;
   (*coord)|= std[b]&0xff;
 }
 return(1);
}


int Imgdbs_deformat_samp(imdbs,fieldcount,fields,fieldlengths,samp,flag)
struct ImgdbS *imdbs;
int fieldcount;
char **fields;
unsigned long *fieldlengths;
struct ImgsamplE *samp;
int flag;
{
 int binary= 0,pixel_length,i,j,ok,b,coord_size,smp_size,stdl,ret;
 char std[Imgsamp_max_sizE*3*sizeof(Imgsamp_coorD)+1],revision[8],num[81];
 char message[256];
 double d;


 if(imdbs->sample_format==0 || imdbs->sample_format==3 ||
    imdbs->sample_format==4) {
   if(fieldcount <1 || fieldlengths[0]==0) {
     sprintf(imdbs->imdb->errmsg,"Sample size mismatch : expected some bytes");
     return(-1);
   }
   ret= Imgsample_decode(samp,fields[0],fieldlengths[0],message,0);
   if(ret<=0) {
     strcpy(imdbs->imdb->errmsg,message);
     return(-1);
   }
   goto done;
 } else if(imdbs->sample_format==1) {
   pixel_length= 3*sizeof(Imgsamp_coorD);
   ok= 0;
   if(fieldcount>=samp->spoints+3) {
     for(i=0;i<samp->spoints;i++)
       if(fieldlengths[i]!=pixel_length)
     break;
     if(i>=samp->spoints)
       ok= 1;
   }
   if(!ok) {
     sprintf(imdbs->imdb->errmsg,
             "Sample size mismatch : expected %ldx%ld bytes + 3 numbers",
             samp->spoints,pixel_length);
     return(-1);
   }
   for(i=0;i<samp->spoints;i++)
     for(j=0;j<pixel_length;j++)
       std[i*pixel_length+j]= fields[i][j];
   stdl= samp->spoints*pixel_length;
   /* append average data from three integer columns */
   for(j=0;j<3;j++) {
     Imgdbs_read_field(imdbs,fields[samp->spoints+j],
                       fieldlengths[samp->spoints+j],num,sizeof(num),0);
     sscanf(num,"%lf",&d);
     if(d>Imgsample_max_coorD||d<0) 
       d= 0;
     for(b=0;b<sizeof(Imgsamp_coorD);b++)
       std[stdl+j*sizeof(Imgsamp_coorD)+b]= 
                             (((int) d)>>(8*(sizeof(Imgsamp_coorD)-1-b)))&0xff;
   }
   binary= 1;
 } else if(imdbs->sample_format==2) {
   if(fieldcount<3*samp->spoints) {
format_2_mismatch:;
     sprintf(imdbs->imdb->errmsg,
             "Sample size mismatch : expected %ld numbers between 0 and %lu",
             3*samp->spoints,Imgsample_max_coorD);
     return(-1);
   }
   for(i=0;i<samp->spoints;i++) {
     Imgdbs_read_field(imdbs,fields[i*3+0],fieldlengths[i*3+0],
                       num,sizeof(num),0);
     sscanf(num,"%lf",&d);
     if(d>Imgsample_max_coorD||d<0) 
       goto format_2_mismatch;
     samp->red[i]= d;
     Imgdbs_read_field(imdbs,fields[i*3+1],fieldlengths[i*3+1],
                       num,sizeof(num),0);
     sscanf(num,"%lf",&d);
     if(d>Imgsample_max_coorD||d<0) 
       goto format_2_mismatch;
     samp->green[i]= d;
     Imgdbs_read_field(imdbs,fields[i*3+2],fieldlengths[i*3+2],
                       num,sizeof(num),0);
     sscanf(num,"%lf",&d);
     if(d>Imgsample_max_coorD||d<0) 
       goto format_2_mismatch;
     samp->blue[i]= d;
   }
 }
 if(binary) {
   coord_size= sizeof(Imgsamp_coorD);
   for(i=0;i<samp->spoints;i++) {
     Imgdbs_deformat_bincoord(imdbs,std+i*3*coord_size+0*coord_size,
                              &(samp->red[i]),0);
     Imgdbs_deformat_bincoord(imdbs,std+i*3*coord_size+1*coord_size,
                              &(samp->green[i]),0);
     Imgdbs_deformat_bincoord(imdbs,std+i*3*coord_size+2*coord_size,
                              &(samp->blue[i]),0);
   }
 }
done:;
 samp->invalid&= ~8;
 return(1);
}


int Imgdbs_fieldnames(imdbs,samp,result,flag)
struct ImgdbS *imdbs;
struct ImgsamplE *samp; /* mainly for sample->spoints */
char **result;          /* must be NULL or a valid freeable memory address */
int flag;
/*
 bit0= append to result rather than overwrite
*/
{
 int ret;

 if(Sregex_string(result,imdbs->id_fieldname,flag&1)<=0)
   return(-1);
 if(Sregex_string(result,",",1)<=0)
   return(-1);
 if(Sregex_string(result,imdbs->adr_fieldname,1)<=0)
   return(-1);
 if(Sregex_string(result,",",1)<=0)
   return(-1);
 ret= Imgdbs_format_samp(imdbs,samp,result,2|4);
 if(ret<=0)
   return(ret);
 return(1);
}


int Imgdbs_write_samp(imdbs,writesample,flag)
struct ImgdbS *imdbs;
struct ImgsamplE *writesample;
int flag;
/*
 bit14= with bit15: also delete all samples with higher id
 bit15= delete sample from database
*/
{
 int i,ret,head,sp,qret;
 char *query_text= NULL,id_text[81],*msg,op[8];

 strcpy(op,"=");
 if(flag&(1<<15)) {
   writesample->invalid= -1;
   /* DELETE FROM $table 
             WHERE id=$writesample->id */
   if(Sregex_string(&query_text,"DELETE FROM ",0)<=0)
     {ret= -1; goto ex;}
   if(Sregex_string(&query_text,imdbs->sample_table,1)<=0)
     {ret= -1; goto ex;}
   if(flag&(1<<14))
     strcpy(op,">=");
 } else {
   /* UPDATE $table SET sample=$writesample 
                    WHERE id=$writesample->id */
   if(Sregex_string(&query_text,"UPDATE ",0)<=0)
     {ret= -1; goto ex;}
   if(Sregex_string(&query_text,imdbs->sample_table,1)<=0)
     {ret= -1; goto ex;}
   if(Sregex_string(&query_text," SET ",1)<=0)
     {ret= -1; goto ex;}
   ret= Imgdbs_format_samp(imdbs,writesample,&query_text,1|4);
   if(ret<=0) {
     sprintf(imdbs->imdb->errmsg,"Formatting of binary sample data failed\n");
     return(ret);
   }
 }
 if(Sregex_string(&query_text," WHERE ",1)<=0)
   {ret= -1; goto ex;}

 if(writesample->id==0) {

   /* >>> try to find out id by name */;

 }

 sprintf(id_text,"%s%s%ld",imdbs->id_fieldname,op,writesample->id);
 if(Sregex_string(&query_text,id_text,1)<=0)
   {ret= -1; goto ex;}
 ret= Imgdbs_issue_query(imdbs,query_text,strlen(query_text),0);
 if(ret<=0) 
   goto ex;
 ret= 1;
ex:;
 Sregex_string(&query_text,NULL,0);
 return(ret);
}


int Imgdbs_rename(imdbs,samp,new_name,flag)
struct ImgdbS *imdbs;
struct ImgsamplE *samp;
char *new_name;
int flag;
{
 int i,ret,head,sp,qret;
 char *query_text= NULL,id_text[81];

 /* UPDATE $table SET name=new_name 
                  WHERE id=$samp->id */
 if(Sregex_string(&query_text,"UPDATE ",0)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&query_text,imdbs->sample_table,1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&query_text," SET ",1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&query_text,imdbs->adr_fieldname,1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&query_text,"=",1)<=0)
   {ret= -1; goto ex;}
 ret= Imgdbs_format_nam(imdbs,new_name,&query_text,1);
 if(ret<=0) 
   goto ex;
 if(Sregex_string(&query_text," WHERE ",1)<=0)
   {ret= -1; goto ex;}

 if(samp->id==0) {
   /* >>> try to find out id by name */;
 }

 sprintf(id_text,"%s=%ld",imdbs->id_fieldname,samp->id);
 if(Sregex_string(&query_text,id_text,1)<=0)
   {ret= -1; goto ex;}
 ret= Imgdbs_issue_query(imdbs,query_text,strlen(query_text),0);
 if(ret<=0) 
   goto ex;
 ret= 1;
ex:;
 Sregex_string(&query_text,NULL,0);
 return(ret);
}


int Imgdbs_insert(imdbs,table,fieldnames,values,insert_id,flag)
struct ImgdbS *imdbs;
char *table;
char *fieldnames;
char *values;
unsigned long *insert_id;
int flag;
{
 int ret;
 char *query_text= NULL;

 /* INSERT INTO $table ($fieldname) VALUES($values) */;
 if(Sregex_string(&query_text,"INSERT INTO ",0)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&query_text,table,1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&query_text," (",1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&query_text,fieldnames,1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&query_text,") VALUES(",1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&query_text,values,1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&query_text,")",1)<=0)
   {ret= -1; goto ex;}
 ret= Imgdbs_issue_query(imdbs,query_text,strlen(query_text),0);
 if(ret<=0) 
   goto ex;
 Imgdbs_mysql_get_unique_id(imdbs,insert_id,0);

 ret= 1;
ex:;
 Sregex_string(&query_text,NULL,0);
 return(ret);

write_err:;
 sprintf(imdbs->imdb->errmsg,"cannot write sample to map table : ");
 sprintf(imdbs->imdb->errmsg+strlen(imdbs->imdb->errmsg),
         "hostname=%s , database=%s , table=%s , username=%s\n",
         imdbs->hostname,imdbs->dbname,table,imdbs->username);
 ret= 0; goto ex;
}


int Imgdbs_simple_select(imdbs,what,table,where,result,flag)
struct ImgdbS *imdbs;
char *what,*table,*where;
char **result;          /* must be NULL or a valid freeable memory address */
int flag;
/*
 bit0= just start query : do not read row or end query
 bit1= accept NULL result as empty sting
*/
{
 int ret;
 char *query_text=  NULL;
 unsigned long fieldcount,*fieldlengths;
 char **row;

 if(Sregex_string(&query_text,"SELECT ",0)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&query_text,what,1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&query_text," FROM ",1)<=0)
   {ret= -1; goto ex;}
 if(Sregex_string(&query_text,table,1)<=0)
   {ret= -1; goto ex;}
 if(where[0]!=0) {
   if(Sregex_string(&query_text," WHERE ",1)<=0)
     {ret= -1; goto ex;}
   if(Sregex_string(&query_text,where,1)<=0)
     {ret= -1; goto ex;}
 }
 ret= Imgdbs_issue_query(imdbs,query_text,strlen(query_text),1);
 if(ret<=0) 
   goto ex;
 /* get next row from last query */;
 fieldcount= Imgdbs_mysql_get_fieldcount(imdbs,0);
 if(fieldcount<1)
   {ret= 0; goto ex;}
 if(flag&1)
   {ret= 1; goto ex;}
 ret= Imgdbs_mysql_get_row(imdbs,&row,&fieldlengths,0);
 if(ret<=0)
   {ret= 0; goto ex;}
 if(fieldlengths[0]==0)
   {ret= 0; goto ex;}
 if(Sregex_string_cut(result,row[0],fieldlengths[0],0)<=0)
   {ret= -1; goto ex;}
 ret= 1;
ex:;
 if(ret<=0 || !(flag&1))
   Imgdbs_end_query(imdbs,0);
 if(ret==0 && (flag&2)) {
   if(Sregex_string(result,"",0)<=0)
     ret= -1;
   else
     ret= 1;
 }  
 Sregex_string(&query_text,NULL,0);
 return(ret);
}


int Imgdbs_record(imdbs,filename,writesample,flag)
struct ImgdbS *imdbs;
char *filename;
struct ImgsamplE *writesample;
int flag;
{
 int ret,i,l,qret;
 char *value_text= NULL,*fieldnames= NULL;
 char fileadr[2*SfileadrL],seedtext[81];
 char *msg;
 unsigned long seed,insert_id;

 /* INSERT INTO $table ($idname,$adrname,$samplenames)
               VALUES(0,$writesample->filename,$writesampledata) */;

 ret= Imgdbs_fieldnames(imdbs,writesample,&fieldnames,0);
 if(ret<=0)
   goto ex;

 ret= Imgdbs_mysql_unique_seed(imdbs,&seed,0);
 if(ret<=0)
   goto ex;
 sprintf(seedtext,"%ld,",seed);
 if(Sregex_string(&value_text,seedtext,0)<=0)
   {ret= -1; goto ex;}
 ret= Imgdbs_format_nam(imdbs,filename,&value_text,1);
 if(ret<=0)
   goto ex;
 if(Sregex_string(&value_text,",",1)<=0)
   {ret= -1; goto ex;}
 ret= Imgdbs_format_samp(imdbs,writesample,&value_text,4);
 if(ret<=0) {
   sprintf(imdbs->imdb->errmsg,"Formatting of binary sample data failed\n");
   return(ret);
 }

 ret= Imgdbs_insert(imdbs,imdbs->sample_table,fieldnames,value_text,
                    &insert_id,0);
 if(ret<=0)
   goto ex;
 /* Imgsample_set_samp_adr(writesample,insert_id,1); */
 Imgsample_set_id(writesample,insert_id);
 ret= 1;
ex:;
 Sregex_string(&value_text,NULL,0);
 Sregex_string(&fieldnames,NULL,0);
 return(ret);
}


int Imgdbs_read_field(imdbs,raw,fieldlength,field,fieldsize,flag)
struct ImgdbS *imdbs;
char *raw;
int fieldlength;
char *field;
unsigned long fieldsize;
int flag;
{
 int long fl;

 fl= fieldlength;
 if(fl>=fieldsize)
   fl= fieldsize-1;
 strncpy(field,raw,fl);
 field[fl]= 0;
 return(fl);
}


int Imgdbs_read_samp(imdbs,readsample,flag)
struct ImgdbS *imdbs;
struct ImgsamplE *readsample;
int flag;
/*
  bit0= address with readsample->id (i.e. start a new query)
  bit1= only start query, do not read first result line
  bit2= use prepared search_where clause
  bit4= prepare for possible intermediate queries
        (automatically done if not started iyet and not bit0|bit1)
*/
{
 int i,sp,ret,start_new_query= 0;
 unsigned long fieldcount,*fieldlengths;
 char *query_text= NULL,*where_text= NULL,*fieldnames= NULL;
 char *msg,id_text[81],num[81],filename[SfileadrL];
 char **row;
 double d;

 if(flag&1) {
   start_new_query= 1;
   /* WHERE id = $readsample->id */;
   if(Sregex_string(&where_text," WHERE ",0)<=0)
     {ret= -1; goto ex;}
   sprintf(id_text,"%s=%ld",imdbs->id_fieldname,readsample->id);
   if(Sregex_string(&where_text,id_text,1)<=0)
     {ret= -1; goto ex;}
 } else if(imdbs->imdb->query_token==NULL || imdbs->query_result==NULL) {
   start_new_query= 1;
   if(!(flag&2))
     flag|= 16;
   if(Sregex_string(&where_text,"",0)<=0)
     {ret= -1; goto ex;}
   if((flag&4) && imdbs->search_where!=NULL) {
     if(Sregex_string(&where_text,imdbs->search_where,1)<=0)
       {ret= -1; goto ex;}
   }
   /* ORDER BY id */
   if(Sregex_string(&where_text," ORDER BY ",1)<=0)
     {ret= -1; goto ex;}
   if(Sregex_string(&where_text,imdbs->id_fieldname,1)<=0)
     {ret= -1; goto ex;}
 }
 if(start_new_query) {
   /* SELECT $fieldnames FROM $table [WHERE <see above> | ORDER BY id] */;
   if(Sregex_string(&query_text,"SELECT ",0)<=0)
     {ret= -1; goto ex;}
   ret= Imgdbs_fieldnames(imdbs,readsample,&query_text,1);
   if(ret<=0)
     goto ex;
   if(Sregex_string(&query_text," FROM ",1)<=0)
     {ret= -1; goto ex;}
   if(Sregex_string(&query_text,imdbs->sample_table,1)<=0)
     {ret= -1; goto ex;}
   if(where_text[0]!=0)
     if(Sregex_string(&query_text,where_text,1)<=0)
       {ret= -1; goto ex;}
   ret= Imgdbs_issue_query(imdbs,query_text,strlen(query_text),1|(flag&16));
   if(ret<=0) 
     goto ex;
 }
 if(imdbs->query_result==NULL)
   {ret= 0; goto ex;}
 fieldcount= Imgdbs_mysql_get_fieldcount(imdbs,0);
 if(fieldcount==0)
   {ret= 0; goto ex;}
 if(fieldcount<6) {
   sprintf(imdbs->imdb->errmsg,
           "Record size mismatch : expected at least six columns");
   ret= -1; goto ex;
 }
 if(flag&2)
   {ret= 1; goto ex;}

 /* get next row from last query */;
 ret= Imgdbs_mysql_get_row(imdbs,&row,&fieldlengths,0);
 if(ret<=0)
   {ret= 0; goto ex;}

 /* transfer data into readsample */
 Imgdbs_read_field(imdbs,row[0],fieldlengths[0],num,sizeof(num),0);
 sscanf(num,"%lf",&d);

 /* >>> again the 64 bit problem */
 if(d>((double) 0xffffffff) || d<0) {
   sprintf(imdbs->imdb->errmsg,
           "Id mismatch : expected a number between 1 and %lu",0xffffffff);
   ret= -1; goto ex;
 }
 /* Imgsample_set_samp_adr(readsample,(unsigned long) d,1); */
 Imgsample_set_id(readsample,(unsigned long) d,0);
 if(fieldlengths[1]==0) {
   sprintf(imdbs->imdb->errmsg,
           "Fileaddress mismatch : expected at least one character");
   ret= -1; goto ex;
 }
 Imgdbs_read_field(imdbs,row[1],fieldlengths[1],filename,sizeof(filename),0);
 ret= Imgsample_set_filename(readsample,filename,0);
 if(ret<=0)
   goto ex;
 ret= Imgdbs_deformat_samp(imdbs,fieldcount-2,row+2,fieldlengths+2,
                           readsample,0);
 if(ret<=0)
   goto ex;

 ret= 1;
ex:;
 Sregex_string(&where_text,NULL,0);
 Sregex_string(&query_text,NULL,0);
 Sregex_string(&fieldnames,NULL,0);
 return(ret);
}


int Imgdbs_read_nam(imdbs,readsample,flag)
struct ImgdbS *imdbs;
struct ImgsamplE *readsample;
int flag;
{
 /* the filename has already been read with Imgdbs_read_samp() */
 return(1);
}


int Imgdbs_del_back(imdbs,samp,flag)
/* delete sample and everything that has a higher address */
struct ImgdbS *imdbs;
struct ImgsamplE *samp;
int flag;
{
 int ret;

 ret= Imgdbs_write_samp(imdbs,samp,(1<<14)|(1<<15));
 return(ret);
}


int Imgdbs_rm_sample(imdbs,samp,flag)
struct ImgdbS *imdbs;
struct ImgsamplE *samp;
int flag;
{
 int ret;

 ret= Imgdbs_write_samp(imdbs,samp,(1<<15));
 return(ret);
}


int Imgdbs_add_exempt(imdbs,filename,lookalikes,flag)
struct ImgdbS *imdbs;
char *filename;
struct LstrinG **lookalikes;
int flag;
{
 int ret;
 struct LstrinG *ls;
 char *value_text= NULL,eid[81],fieldnames[161];
 unsigned long insert_id;

 /* INSERT INTO $table ($eidname,$ebossname,$eadrname)
           VALUE(0,1,$filename) */
 sprintf(fieldnames,"%s,%s,%s",imdbs->eid_fieldname,imdbs->eboss_fieldname,
                               imdbs->eadr_fieldname);
 if(Sregex_string(&value_text,"0,1,",0)<=0)
   {ret= -1; goto ex;}
 ret= Imgdbs_format_nam(imdbs,filename,&value_text,1);
 if(ret<=0)
   goto ex;
 ret= Imgdbs_insert(imdbs,imdbs->exempt_table,fieldnames,value_text,
                    &insert_id,0);
 if(ret<=0)
   goto ex;
 sprintf(eid,"%lu,0,",insert_id);
 for(ls= *lookalikes;ls!=NULL;ls=ls->next) {
   /* INSERT INTO $table ($eidname,$ebossname,$eadrname)
             VALUE($eid,0,$ls->text) */;
   if(Sregex_string(&value_text,eid,0)<=0)
     {ret= -1; goto ex;}
   ret= Imgdbs_format_nam(imdbs,ls->text,&value_text,1);
   if(ret<=0)
     goto ex;
   ret= Imgdbs_insert(imdbs,imdbs->exempt_table,fieldnames,value_text,
                      &insert_id,0);
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:;
 Sregex_string(&value_text,NULL,0);
 return(ret);
}


int Imgdbs_exemptlist_is_changed(imdbs,list_seal,seal,flag)
struct ImgdbS *imdbs;
unsigned long list_seal;
unsigned long *seal;
int flag;
{
 int ret,count;
 unsigned long maxexemt= 0,crc;
 char where[81],what[81],*result= NULL;

 /* SELECT MAX($eidname) FROM $table WHERE $ebossname = 1 */
 sprintf(what,"MAX(%s)",imdbs->eid_fieldname);
 sprintf(where,"%s = 1",imdbs->eboss_fieldname);
 ret= Imgdbs_simple_select(imdbs,what,imdbs->exempt_table,where,&result,2);
 if(ret<=0)
   {ret= -1; goto ex;}
 if(result[0]!=0)
   sscanf(result,"%lu",&maxexemt);
 ret= Imgdbs_count_exempts(imdbs,&count,1);
 if(ret<=0)
   {ret= -1; goto ex;}
 Sregex_crc32(&maxexemt,sizeof(unsigned long),&crc,1);
 Sregex_crc32(&count,sizeof(unsigned long),&crc,0);
 *seal= crc;
 if(list_seal == *seal)
   ret= 0;
 else
   ret= 1;
ex:
 Sregex_string(&result,NULL,0);
 return(ret);
}


int Imgdbs_count_exempts(imdbs,count,flag)
struct ImgdbS *imdbs;
int *count;
int flag;
/*
 bit0= do not add 1 for security
*/
{
 int ret;
 unsigned long maxexemt= 0,crc;
 char where[81],what[81],*result= NULL;

 /* SELECT COUNT($ebossname) FROM $table WHERE $ebossname = 1 */
 sprintf(what,"COUNT(%s)",imdbs->eboss_fieldname);
 sprintf(where,"%s = 1",imdbs->eboss_fieldname);
 ret= Imgdbs_simple_select(imdbs,what,imdbs->exempt_table,where,&result,0);
 if(ret<=0)
   goto ex;
 sscanf(result,"%lu",count);
 ret= 1;
ex:
 Sregex_string(&result,NULL,0);
 return(ret);
}


int Imgdbs_read_exempt(imdbs,listname,filename,lstring,flag)
struct ImgdbS *imdbs;
char *listname;
char filename[SfileadrL];
struct LstrinG **lstring;
int flag;
/*
 bit0= resume reading
*/
/*
 return:  1=more to read
          0=everything is read (or cannot open file)
         -1=fatal failure
*/
{
 int ret,count,i;
 char where[81],what[161],*result= NULL,num[81],lbuffer[SfileadrL];
 unsigned long fieldcount,*fieldlengths;
 double d;
 char **row;

 if(!(flag&1)) {
   /* fetch list of exempt ids */
   imdbs->eid_count= 0;
   if(imdbs->eid_list!=NULL) {
     Smem_freE(imdbs->eid_list);
     imdbs->eid_list= NULL;
   }
   ret= Imgdbs_count_exempts(imdbs,&count,0);
   if(ret<=0)
     {ret= -1; goto ex;}
   /* SELECT $eidname FROM $table WHERE $ebossname = 1 */;
   sprintf(where,"%s = 1",imdbs->eboss_fieldname);
   ret= Imgdbs_simple_select(imdbs,imdbs->eid_fieldname,imdbs->exempt_table,
                             where,result,1);
   if(ret<=0)
     {ret= -1; goto ex;}
   /* transfer result to eid_list,eid_count */
   imdbs->eid_list= TSOB_FELD(unsigned long,count+1);
   if(imdbs->eid_list==NULL)
     {ret= -1; goto ex;}
   for(i=0;i<count;i++) {
     ret= Imgdbs_mysql_get_row(imdbs,&row,&fieldlengths,0);
     if(ret<0)
       goto ex;
     if(ret==0)
   break;    
     Imgdbs_read_field(imdbs,row[0],fieldlengths[0],num,sizeof(num),0);
     sscanf(num,"%lf",&d);
     if(d<0 || d>0xFFFFFFFF)
   break;
     imdbs->eid_list[imdbs->eid_count]= d;
     imdbs->eid_count++;
   }
   Imgdbs_end_query(imdbs,0);
   imdbs->eid_readpos= 0;
 }

 /* are there still ids unread ? */
 if(imdbs->eid_readpos >= imdbs->eid_count || imdbs->eid_list==NULL)
   {ret= 0; goto ex;}

 /* SELECT $ebossname,$eadrname FROM $table WHERE $eidname = $eid */; 
 sprintf(what,"%s,%s",imdbs->eboss_fieldname,imdbs->eadr_fieldname);
 sprintf(where,"%s = %lu",imdbs->eid_fieldname,
                          imdbs->eid_list[imdbs->eid_readpos]);
 ret= Imgdbs_simple_select(imdbs,what,imdbs->exempt_table,where,result,1);
 if(ret<=0)
   {ret= -1; goto ex;}
 fieldcount= Imgdbs_mysql_get_fieldcount(imdbs,0);
 if(fieldcount<2) {
   sprintf(imdbs->imdb->errmsg,
           "Format mismatch while reading an exempt : expected 2 fields");
   ret= -1; goto ex;
 }
 do {
   ret= Imgdbs_mysql_get_row(imdbs,&row,&fieldlengths,0);
   if(ret<0)
     goto ex;
   if(ret==0)
 break;
   if(fieldlengths[0]==0)
 break; 
   if(fieldlengths[0]>0 && strncmp(row[0],"1",fieldlengths[0])==0) {
     Imgdbs_read_field(imdbs,row[1],fieldlengths[1],
                       filename,SfileadrL,0);
   } else {
     Imgdbs_read_field(imdbs,row[1],fieldlengths[1],lbuffer,sizeof(lbuffer),0);
     ret= Lstring_new(lstring,lbuffer,*lstring,1);
     if(ret<=0)
       {ret= -1; goto ex;}
   }
 } while(1);
 imdbs->eid_readpos++;
 ret= 1;
ex:;
 Imgdbs_end_query(imdbs,0);
 Sregex_string(&result,NULL,0);
 return(ret); 
}


int Imgdbs_garbage_coll(imdbs,flag)
struct ImgdbS *imdbs;
int flag;
{
 int ret;

 /* >>> ??? renumbering of ids ??? */;

 ret= 1;
ex:;
 return(ret);
}


