
/*
  cc -g -c imgexempt.c
*/

#include <unistd.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>



#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"

#include "imgexempt.h"


/* ----------------------------- ImgexempT ------------------------------ */


int Imgexempt_new(exempt,filename,lookalikes,flag)
struct ImgexempT **exempt;
char *filename;
struct LstrinG **lookalikes;
int flag;
{
 struct ImgexempT *e;

 *exempt= e= TSOB_FELD(struct ImgexempT,1);
 if(e==NULL)
   return(-1);
 e->filename= NULL;
 e->lookalikes= NULL;
 if(Sregex_string(&(e->filename),filename,0)<=0)
   goto failed;
 e->lookalikes= *lookalikes;
 *lookalikes= NULL;            /* owner is the new exempt object now */
 return(1);
failed:
 Imgexempt_destroy(exempt,0);
 return(-1);
}


int Imgexempt_destroy(exempt,flag)
struct ImgexempT **exempt;
int flag;
{
 if((*exempt)==NULL)
   return(0);
   
 Sregex_string(&((*exempt)->filename),NULL,0);
 Lstring_destroy_all(&((*exempt)->lookalikes),0);
 Smem_freE((char *) *exempt);
 *exempt= NULL;
 return(1);
}


int Imgexempt_cmp_qsort(left,right)
/* to be used by qsort() */
struct ImgexempT **left,**right;
{
 return(strcmp((*left)->filename,(*right)->filename));
}


int Imgexempt_format(exempt,result,flag)
struct ImgexempT *exempt;
char **result;                 /* *result must be NULL or a freeable address */
int flag;
{
 int ret;
 char delimiter[SfileadrL];
 struct LstrinG *s;

 /* find a suitable delimiter */
 strcpy(delimiter,"+");
 ret= Sfile_make_delimiter(exempt->filename,delimiter,sizeof(delimiter),0);
 if(ret<=0)
   return(ret);
 for(s=exempt->lookalikes;s!=NULL;s= s->next) {
   ret= Sfile_make_delimiter(s->text,delimiter,sizeof(delimiter),0);
   if(ret<=0)
     return(ret);
 }
 /* write result */
 if(Sregex_string(result,"-exempt-",0)<=0)
   return(-1);
 if(Sregex_string(result,delimiter,1)<=0)
   return(-1);
 if(Sregex_string(result,"-",1)<=0)
   return(-1);
 if(Sregex_string(result,exempt->filename,1)<=0)
   return(-1);
 for(s=exempt->lookalikes;s!=NULL;s= s->next) {
   if(Sregex_string(result,delimiter,1)<=0)
     return(-1);
   if(Sregex_string(result,s->text,1)<=0)
     return(-1);
 }
 return(1);
}


int Imgexempt_deformat(exempt,text,flag)
struct ImgexempT **exempt;
char *text;
int flag;
{
 int ret;
 char delimiter[SfileadrL],filename[SfileadrL],lname[SfileadrL],*dpt,*cpt;
 static char head[]= {"-exempt-"};
 struct LstrinG *lookalikes= NULL;

 *exempt= NULL;
 cpt= text;

 /* check format and obtain delimiter text */
 if(strncmp(cpt,head,strlen(head))!=0)
   return(0);
 cpt+= strlen(head);
 dpt= strchr(cpt,'-');
 if(dpt==NULL) 
   {ret= 0; goto ex;}
 if(dpt-cpt>=sizeof(delimiter)-1 || dpt==cpt)
   {ret= 0; goto ex;}
 strncpy(delimiter,cpt,dpt-cpt);
 delimiter[dpt-cpt]= 0;
 cpt= dpt+1;

 /* obtain filename and lookalikes */
 dpt= strstr(cpt,delimiter);
 if(dpt==NULL)
   {ret= 0; goto ex;}
 if(dpt-cpt>=sizeof(filename) || dpt==cpt)
   {ret= 0; goto ex;}
 strncpy(filename,cpt,dpt-cpt);
 filename[dpt-cpt]= 0;
 cpt= dpt+strlen(delimiter);
 while(strlen(cpt)) {
   dpt= strstr(cpt,delimiter);
   if(dpt==NULL) {
     /* append last lookalike */;
     dpt= cpt+strlen(cpt);
   }
   if(dpt-cpt>=sizeof(lname) || dpt==cpt)
     {ret= 0; goto ex;}
   strncpy(lname,cpt,dpt-cpt);
   lname[dpt-cpt]= 0;
   if(Lstring_new(&lookalikes,lname,lookalikes,0)<=0)
     {ret= -1; goto ex;}
   cpt= dpt;
   if(*cpt!=0)
     cpt+= strlen(delimiter);
 }
 if(lookalikes==NULL)
   {ret= 0; goto ex;}
 ret= Imgexempt_new(exempt,filename,&lookalikes,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 Lstring_destroy_all(&lookalikes,0);
 return(ret);
}


/* --------------------------- ImgexemptlisT ------------------------------ */


int Imgexemptlist_new(exemptlist,imp,listname,flag)
struct ImgexemptlisT **exemptlist;
struct ImgmaP *imp;
char *listname;
int flag;
/*
 bit0= do not read exemptlist from file
*/
{
 int ret,i;
 struct ImgexemptlisT *e;

 *exemptlist= e= TSOB_FELD(struct ImgexemptlisT,1);
 if(e==NULL)
   return(-1);
 e->imp= imp;
 e->refcount= 1;
 e->exempt= NULL;
 e->count= 0;
 for(i=0;i<Imgexempt_unsorted_addsizE;i++)
   e->added_exempts[i]= NULL;
 e->added_counter= 0;
 e->listname= NULL;
 e->list_seal= 0;

 if(!(flag&1)) {
   ret= Imgmap_read_exemptlist(e->imp,0);
   if(ret<=0)
     goto failed;
 }

 return(1);
failed:;
 Imgexemptlist_destroy(exemptlist,0);
 return(ret);
}


int Imgexemptlist_destroy(exemptlist,flag)
struct ImgexemptlisT **exemptlist;
int flag;
{
 int i;
 struct ImgexemptlisT *e;

 e= *exemptlist;
 if((e)==NULL)
   return(0);
 e->refcount--;
 if(e->refcount>0)
   return(2);
 Imgexemptlist_destroy_exempts(e,0);
 Sregex_string(&(e->listname),NULL,0);
 Smem_freE((char *) e);
 *exemptlist= NULL;
 return(1);
}


int Imgexemptlist_destroy_exempts(exemptlist,flag)
struct ImgexemptlisT *exemptlist;
int flag;
{
 int i;

 for(i=0;i<exemptlist->added_counter;i++)
   Imgexempt_destroy(&(exemptlist->added_exempts[i]),0);
 exemptlist->added_counter= 0;
 if(exemptlist->exempt!=NULL) {
   for(i=0;i<exemptlist->count;i++)
     Imgexempt_destroy(&(exemptlist->exempt[i]),0);
   Smem_freE((char *) exemptlist->exempt);
   exemptlist->exempt= NULL;
 }
 exemptlist->count= 0;
 return(1);
}


int Imgexemptlist_refer(exemptlist,referer,flag)
struct ImgexemptlisT *exemptlist;
struct ImgexemptlisT **referer;
int flag;
{
 if(exemptlist==NULL)
   return(0);
 exemptlist->refcount++;
 *referer= exemptlist;
 return(1);
}


int Imgexemptlist_new_exempts(exemptlist,count,flag)
struct ImgexemptlisT *exemptlist;
int count;
int flag;
{
 int i;

 Imgexemptlist_destroy_exempts(exemptlist,0);
 if(count==0)
   return(2);
 exemptlist->exempt= TSOB_FELD(struct ImgexempT *,count);
 if(exemptlist->exempt==NULL)
   return(-1);
 for(i=0;i<count;i++) 
   exemptlist->exempt[i]= NULL;
 return(1);
}


int Imgexemptlist_set_seal(exemptlist,seal,flag)
struct ImgexemptlisT *exemptlist;
unsigned long seal;
int flag;
{
 exemptlist->list_seal= seal;
 return(1);
}


int Imgexemptlist_set_count(exemptlist,count,flag)
struct ImgexemptlisT *exemptlist;
int count;
int flag;
{
 exemptlist->count= count;
 return(1);
}


int Imgexemptlist_set_listname(exemptlist,listname,flag)
struct ImgexemptlisT *exemptlist;
char *listname;
int flag;
{
 int ret;

 ret= Sregex_string(&(exemptlist->listname),listname,0);
 return(ret);
}


int Imgexemptlist_sort(exemptlist,flag)
struct ImgexemptlisT *exemptlist;
int flag;
{
 if(exemptlist->count<=0)
   return(2);
 qsort(exemptlist->exempt,exemptlist->count,sizeof(struct ImgexempT *),
       Imgexempt_cmp_qsort);
 return(1);
}


int Imgexemptlist_search_idx(exemptlist,name,result_idx,flag)
struct ImgexemptlisT *exemptlist;
char *name;
int *result_idx;
int flag;
{
 int ret,lb,ub,step,idx,cmp;

 if(exemptlist->count<=0) {
   *result_idx= 0;
   return(0);
 }
 lb= 0;
 ub= exemptlist->count-1;

 while(1) {
   step= ub-lb; 
   idx= lb+step/2+step%2;
   cmp= strcmp(exemptlist->exempt[idx]->filename,name);
   if(cmp<0) {
     lb= idx;
   } else if(cmp>0) {
     if(ub==idx)
       ub= lb;
     else
       ub= idx;
   } else {
     *result_idx= idx;
     return(1);
   }
   if(step==0)
 break;
 }
 *result_idx= idx;
 if(strcmp(exemptlist->exempt[idx]->filename,name)<0)
   *result_idx++;
 return(0);
}


int Imgexemptlist_is_exempted(exemptlist,name1,name2,flag)
struct ImgexemptlisT *exemptlist;
char *name1,*name2;
int flag;
{
 int ret,idx;

 if(exemptlist==NULL)
   return(0);
 ret= Imgexemptlist_search_idx(exemptlist,name1,&idx,0);
 if(ret<0)
   return(ret);
 if(ret==0)
   goto not_in_exempt;

 /* look for first of possibly multiple exempts for name1 */
 while(idx>0) {
   if(strcmp(exemptlist->exempt[idx-1]->filename,name1)!=0)
 break;
   idx--;
 }
 /* check all exempts of name1 for existence of name2 */
 while(strcmp(exemptlist->exempt[idx]->filename,name1)==0) {
   if(Lstring_is_listed(exemptlist->exempt[idx]->lookalikes,name2,0)>0)
     return(1);
   if(idx<exemptlist->count-1)
     idx++;
   else
 break;
 }
not_in_exempt:;
 /* test wether exempt was recently added */
 for(idx=0;idx<exemptlist->added_counter;idx++)
   if(strcmp(exemptlist->added_exempts[idx]->filename,name1)==0)
     if(Lstring_is_listed(exemptlist->added_exempts[idx]->lookalikes,name2,0)>0)
       return(1);
 return(0);
}


int Imgexemptlist_can_still_add(exemptlist,flag)
struct ImgexemptlisT *exemptlist;
int flag;
{
 if(exemptlist==NULL)
   return(0);
 return(exemptlist->added_counter<Imgexempt_unsorted_addsizE);
}


int Imgexemptlist_add(exemptlist,filename,lookalikes,flag)
struct ImgexemptlisT *exemptlist;
char *filename;
struct LstrinG **lookalikes;
int flag;
/*
 bit0= do not copy modification time from file
*/
{
 int ret;
 struct ImgexempT *exempt;
 struct stat stbuf;

 if(!Imgexemptlist_can_still_add(exemptlist,0))
   return(0);
 ret= Imgexempt_new(&exempt,filename,lookalikes,0);
 if(ret<=0)
   goto ex;
 exemptlist->added_exempts[exemptlist->added_counter]= exempt;
 exemptlist->added_counter++;
 if(flag&1)
   {ret= 1; goto ex;} 
 Imgmap_exemptlist_is_changed(exemptlist->imp,exemptlist->list_seal,
                              &(exemptlist->list_seal),0);
 ret= 1;
ex:;
 return(ret);
}


int Imgexemptlist_next_match(exemptlist,sregex,exempt,idx,flag)
struct ImgexemptlisT *exemptlist;
struct SregeX *sregex;
struct ImgexempT **exempt;
int *idx;
int flag;
{
 int i,ret;

 for(i= *idx;i<exemptlist->count;i++) {
   if(Sregex_is_match(sregex,exemptlist->exempt[i]->filename,0)) {
     *idx= i+1;
     *exempt= exemptlist->exempt[i];
     return(1);
   }
 }
 *idx= exemptlist->count;
 return(0);
}


