
/*
 cc -g -c imgrgb.c
*/


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <setjmp.h>

/* libjpeg */
#include <jpeglib.h>
#include <jerror.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"

/* this enables a lot of ugly inner details in  imgrgb.h */
#define Imgrgb_itselF 
#include "imgrgb.h"


/* ------------------------ ImgrgB  --------------------------------- */


int Imgrgb_new(rgb,flag)
struct ImgrgB **rgb;
int flag;
{
 struct ImgrgB *r;

 *rgb= r= TSOB_FELD(struct ImgrgB,1);
 if(r==NULL)
   return(-1);
 Imgrgb_init(r,0);
 return(1);
failed:;
 Imgrgb_destroy(rgb,0);
 return(-1);
}


int Imgrgb_destroy(rgb,flag)
struct ImgrgB **rgb;
int flag;
{
 if(*rgb==NULL)
   return(0);
 Imgrgb_close(*rgb,0);
 Smem_freE((char *) (*rgb));
 *rgb= NULL;
 return(1); 
}


int Imgrgb_init(rgb,flag)
struct ImgrgB *rgb;
int flag;
{
 rgb->fp= NULL;
 rgb->error_handle.reaction= 0;
 rgb->error_handle.chef= rgb;
 rgb->buffer= NULL;
 rgb->red= rgb->green= rgb->blue= NULL;
 rgb->started= 0;
 rgb->errmsg[0]= 0;
 rgb->originalname[0]= 0;
 rgb->filename[0]= 0;
 strcpy(rgb->tempdir,Imgrgb_tempfile_diR);
 rgb->tempfile= 0;
 rgb->frameno= 0;
 rgb->current_frame= 0;
 rgb->app_linebuffer_valid= 0;
 return(1);
}


char *Imgrgb_get_openname(rgb,flag)
struct ImgrgB *rgb;
int flag;
{
 return(rgb->openname);
}


char *Imgrgb_get_errmsg(rgb,flag)
struct ImgrgB *rgb;
int flag;
{
 return(rgb->errmsg);
}


int Imgrgb_sizeof(rgb,flag)
struct ImgrgB *rgb;  
int flag;
{
 return(sizeof(*rgb));
}


double *Imgrgb_get_red(rgb,flag)
struct ImgrgB *rgb;
int flag;
{
 if(!rgb->app_linebuffer_valid)
   Imgrgb_app_line(rgb,0);
 return(rgb->red);
}


double *Imgrgb_get_green(rgb,flag)
struct ImgrgB *rgb;
int flag;
{
 if(!rgb->app_linebuffer_valid)
   Imgrgb_app_line(rgb,0);
 return(rgb->green);
}


double *Imgrgb_get_blue(rgb,flag)
struct ImgrgB *rgb;
int flag;
{
 if(!rgb->app_linebuffer_valid)
   Imgrgb_app_line(rgb,0);
 return(rgb->blue);
}


void Imgrgb_error_exit (cinfo)
                     /* this function is called by libjpeg in case of error */
j_common_ptr cinfo;
{
 struct Imgrgb_error_handlE *errh;

 errh= (struct Imgrgb_error_handlE *) cinfo->err;

 /* transmit error message to ImgrgB */
 (*cinfo->err->format_message) (cinfo, errh->chef->errmsg);

 if(errh->reaction==1) 
   longjmp(errh->setjmp_buffer, 1);
 exit(1);
}


int Imgrgb_frames(rgb,flag)
struct ImgrgB *rgb;
int flag;
/*
 bit0-3= mode:
  0= count
  1= delete frame 0 up to first missing one
  2= delete from 0 to rgb->frameno-1
*/
{
 char name[SfileadrL];
 struct stat stbuf;
 int i,og,mode,counter=0,existing;

 mode= flag&15;
 if(mode==2){
   og= rgb->frameno;
 }else{
   og= 1000000;
 }
 for(i=0;i<og;i++){
   sprintf(name,"%s.%ld",rgb->filename,i);
   existing= stat(name,&stbuf);
   if(existing!=-1) counter++;
   if(mode==0){
     if(existing==-1)
 break;
   }else if(mode==1){
     if(existing==-1)
 break;
     unlink(name);
   }else if(mode==2){
     if(existing==-1)
 continue;
     unlink(name);
   }
 }
 return(counter);
}


int Imgrgb_count_frames(rgb,flag)
struct ImgrgB *rgb;
int flag;
{
 int ret;
 ret= Imgrgb_frames(rgb,0);
 return(ret);
}


int Imgrgb_del_frames(rgb,flag)
struct ImgrgB *rgb;
int flag;
/*
 bit0= delete frames 0 to rgb->frameno-1
       (else delete 0 up to first missing one)
*/
{
 int ret;
 ret= Imgrgb_frames(rgb,1+(flag&1));
 return(ret);
}


int Imgrgb_temp_unlink(rgb,flag)
struct ImgrgB *rgb;
int flag;
{
 int i;

 if(rgb->tempfile!=1)
   return(0);
 if(strncmp(rgb->filename,rgb->tempdir,
            strlen(rgb->tempdir))!=0)
   return(0);
 if(rgb->frameno>0) {
   Imgrgb_del_frames(rgb,1);
 }else{
   unlink(rgb->filename);
 }
 rgb->tempfile= 2;
 return(1);
}


int Imgrgb__is_jpeg(name,flag)
int flag;
{
 unsigned char magic[4];

 if(Sfile_file_to_bin(name,magic,4,1|2)==1) 
   if(magic[0]==0xff && magic[1]==0xd8 && magic[2]==0xff && magic[3]==0xe0)
      return(1);
 return(0);
}


int Imgrgb_may_be_convertible(rgb,name,flag)
struct ImgrgB *rgb;
char *name;
int flag;
/*
 return:
  <0 = fatal error
   0 = will probably not be readable in a senseful way
   1 = try convert
   2 = is already JPEG
*/
{
 int ret;

 ret= Imgrgb__estimate_readable(name,flag);
 if(ret==3)
   ret= 1;
 return(ret);
}


int Imgrgb__estimate_readable(name,flag)
char *name;
int flag;
/*
 return:
  <0 = fatal error
   0 = will probably not be readable in a senseful way
   1 = try convert
   2 = is already JPEG
   3 = no idea
*/
{
 int l,suffl;
 char *dpt,magic[4];

 l= strlen(name);
 dpt= strrchr(name,'.');
 if(dpt==NULL)
   return(3);
 suffl= l-(dpt-name)-1;
 if(suffl<3)
   return(3);
 if(suffl==3) {
   if(toupper(dpt[1])=='J' && toupper(dpt[2])=='P' && toupper(dpt[3])=='G') {
     if(Imgrgb__is_jpeg(name,0))
       return(2);
     return(1);
   }
   if(toupper(dpt[1])=='C' && toupper(dpt[2])=='S' && toupper(dpt[3])=='V')
     return(0);
   if(toupper(dpt[1])=='T' && toupper(dpt[2])=='X' && toupper(dpt[3])=='T')
     return(0);
   if(toupper(dpt[1])=='Z' && toupper(dpt[2])=='I' && toupper(dpt[3])=='P')
     return(0);
   if(toupper(dpt[1])=='E' && toupper(dpt[2])=='N' && toupper(dpt[3])=='C')
     return(0);
   if(toupper(dpt[1])=='M' && toupper(dpt[2])=='P' && toupper(dpt[3])=='G')
     return(0);
   if(toupper(dpt[1])=='A' && toupper(dpt[2])=='V' && toupper(dpt[3])=='I')
     return(0);
   if(toupper(dpt[1])=='H' && toupper(dpt[2])=='T' && 
      toupper(dpt[3])=='M')
     return(0);
   if(toupper(dpt[1])=='G' && toupper(dpt[2])=='I' && 
      toupper(dpt[3])=='F')
     return(1);
   if(toupper(dpt[1])=='T' && toupper(dpt[2])=='I' && 
      toupper(dpt[3])=='F')
     return(1);
   if(toupper(dpt[1])=='P' && toupper(dpt[2])=='N' && 
      toupper(dpt[3])=='G')
     return(1);
   if(toupper(dpt[1])=='B' && toupper(dpt[2])=='M' && 
      toupper(dpt[3])=='P')
     return(1);
   if(toupper(dpt[1])=='R' && toupper(dpt[2])=='G' && 
      toupper(dpt[3])=='B')
     return(1);
   if(toupper(dpt[1])=='X' && toupper(dpt[2])=='B' && 
      toupper(dpt[3])=='M')
     return(1);
   if(toupper(dpt[1])=='X' && toupper(dpt[2])=='P' && 
      toupper(dpt[3])=='M')
     return(1);
 } else if(suffl==4) {
   if(toupper(dpt[1])=='J' && toupper(dpt[2])=='P' && 
      toupper(dpt[3])=='E' && toupper(dpt[4])=='G') {
     if(Imgrgb__is_jpeg(name,0))
       return(2);
     return(1);
   }
   if(toupper(dpt[1])=='H' && toupper(dpt[2])=='T' && 
      toupper(dpt[3])=='M' && toupper(dpt[4])=='L')
     return(0);
   if(toupper(dpt[1])=='M' && toupper(dpt[2])=='P' && 
      toupper(dpt[3])=='E' && toupper(dpt[4])=='G')
     return(0);
   if(toupper(dpt[1])=='T' && toupper(dpt[2])=='I' && 
      toupper(dpt[3])=='F' && toupper(dpt[4])=='F')
     return(0);
   if(toupper(dpt[1])=='H' && toupper(dpt[2])=='P' && 
      toupper(dpt[3])=='G' && toupper(dpt[4])=='L')
     return(0);
 }
 return(3);  
}


int Imgrgb_convert(rgb,name,flag)
struct ImgrgB *rgb;
char *name;
int flag;
{
 int l,ltdir,i,ret;
 char cmd[11*SfileadrL],*cpt,escd[5*SfileadrL];
 struct stat stbuf;

 l= strlen(name);
 ret= Imgrgb_may_be_convertible(rgb,name,0);
 if(ret!=1)
   return(ret);
 cpt= strrchr(name,'/');
 if(cpt==NULL) cpt= name;
 else          cpt++;
 sprintf(rgb->filename,"%s/imgrgb%X_",rgb->tempdir,getpid());
 
 ltdir= strlen(rgb->filename);
 for(i=0;cpt[i]!='.'&&cpt[i]!=0;i++) rgb->filename[ltdir+i]= cpt[i];
 rgb->filename[ltdir+i]= 0;
 strcat(rgb->filename,".jpg");


 Imgrgb_del_frames(rgb,0);
 /*  sprintf(cmd,"convert -quality 90 %s jpeg:%s",name,rgb->filename); */
 if(strchr(name,'[')!=NULL) {
   sprintf(rgb->errmsg,"Filename not usable for command convert : %s\n",name);
   return(0);
 }
 Sfile_esc_shell(name,escd,0);
 sprintf(cmd,"convert -quality 90 %s ",escd);
 Sfile_esc_shell(rgb->filename,escd,0);
 sprintf(cmd+strlen(cmd),"jpeg:%s",escd);
 ret= system(cmd);
 if(ret==0) {
   rgb->tempfile= 1;
 } else {
   sprintf(rgb->errmsg,"conversion failed:  system(%s) = %ld\nerrno=%ld",
                       cmd,ret,errno);
   return(0);
 }
 rgb->frameno= Imgrgb_count_frames(rgb,0);
 return(1);
}


int Imgrgb_open(rgb,name,flag)
struct ImgrgB *rgb;
char *name;
int flag;
{ 
 struct jpeg_decompress_struct *cinfo;
 JSAMPARRAY buffer;
 int width,height,pixelkomp,row_stride,ret;

 Imgrgb_init(rgb,0);
 strcpy(rgb->originalname,name);

 /* eventually convert file from other formats */
 strcpy(rgb->filename,name);
 if(Imgrgb_convert(rgb,name,0)<=0) 
   strcpy(rgb->filename,name);      /* stay with old file if it did not work */

 cinfo= &(rgb->cinfo);
 if(rgb->frameno==0) {
   strcpy(rgb->openname,rgb->filename);
 }else{
   sprintf(rgb->openname,"%s.0",rgb->filename);
 }
 rgb->fp= fopen(rgb->openname,"rb");
#ifdef Imgrgb_line_early_unlinK
 if(rgb->frameno==0 && rgb->tempfile)
   Imgrgb_temp_unlink(rgb,0);
#endif
 if(rgb->fp==NULL) {
   sprintf(rgb->errmsg,"Error while opening : %ld",errno);
   return(0);
 }
 rgb->buffer= NULL;

 /* Install error handler */
 /* this relies on :     &rgb->error_handle == &rgb->error_handle.jerr  :-| */
 cinfo->err = jpeg_std_error(&(rgb->error_handle.jerr));
 rgb->error_handle.jerr.error_exit= Imgrgb_error_exit;
 if (setjmp(rgb->error_handle.setjmp_buffer)) {
   /* here the program pointer materializes after an error has been caught */
   Imgrgb_close(rgb,0);
   {ret= 0; goto ex;}
 }
 rgb->error_handle.reaction= 1;                        /* activate longjump */

 /* create decompression object and attach to file pointer */
 jpeg_create_decompress(cinfo);
 jpeg_stdio_src(cinfo,rgb->fp);

 /* Start decompression */
 jpeg_read_header(cinfo, TRUE);
 jpeg_start_decompress(cinfo);
 rgb->started= 1;
 if(cinfo->output_components!=1 && cinfo->output_components!=3) {
   sprintf(rgb->errmsg,"Unusable number of color channels %ld",
                       cinfo->output_components); 
   ret= 0; goto ex;
 }

 /* provide appropriate line buffer */
 Imgrgb_get_sizes(rgb,&height,&width,&pixelkomp,0);
 row_stride= width*pixelkomp;
#ifdef Imgrgb_line_buffer_liB
 /* Make a one-row-high sample array that will go away when done with image */
 rgb->buffer = (*cinfo->mem->alloc_sarray)
                 ((j_common_ptr) cinfo,JPOOL_IMAGE,row_stride,1);
 if(rgb->buffer==NULL)
   {ret= -1; goto ex;}
#else
 /* make storage by normal means */
 rgb->buffer= (JSAMPARRAY) Smem_malloC(sizeof(JSAMPROW));
 if(rgb->buffer==NULL)
   {ret= -1; goto ex;}
 rgb->buffer[0]= (JSAMPROW) Smem_malloC(width*pixelkomp*sizeof(JSAMPLE));
#endif
 rgb->red= (double *) Smem_malloC(width*sizeof(double));
 rgb->green= (double *) Smem_malloC(width*sizeof(double));
 rgb->blue= (double *) Smem_malloC(width*sizeof(double));
 if(rgb->buffer[0]==NULL || 
    rgb->red==NULL || rgb->green==NULL || rgb->blue==NULL)
   {ret= -1; goto ex;}
 ret= 1;
ex:;
 rgb->error_handle.reaction= 0;                       /* deactivate longjump */
 return(ret);
}


int Imgrgb_get_sizes(rgb,height,width,pixelkomp,flag)
struct ImgrgB *rgb;
int *height,*width,*pixelkomp;
int flag;
{
 *height= rgb->cinfo.output_height;
 *width= rgb->cinfo.output_width;
 *pixelkomp= rgb->cinfo.output_components;
 return(1);
}


int Imgrgb_app_line(rgb,flag)
struct ImgrgB *rgb;
int flag;
{
 int i;

 if(rgb->cinfo.output_components==3){
   for(i=0;i<rgb->cinfo.output_width;i++){
     rgb->red[i]= rgb->buffer[0][i*3];
     rgb->green[i]= rgb->buffer[0][i*3+1];
     rgb->blue[i]= rgb->buffer[0][i*3+2];
   }
 }else{
   for(i=0;i<rgb->cinfo.output_width;i++){
     rgb->red[i]= rgb->green[i]= rgb->blue[i]= rgb->buffer[0][i];
   }
 }
 rgb->app_linebuffer_valid= 1;
 return(1);
}


int Imgrgb_read_line(rgb,flag)
struct ImgrgB *rgb;
int flag;
/*
 bit1= fill application style linebuffer
*/
{
 int ret,i;

 rgb->app_linebuffer_valid= 0;
 if(rgb->cinfo.output_scanline>=rgb->cinfo.output_height)
   return(0);
 if (setjmp(rgb->error_handle.setjmp_buffer)) {
   /* here the program pointer materializes after an error has been caught */
   Imgrgb_close(rgb,0);
   return(0);
 }
 rgb->error_handle.reaction= 1;                         /* activate longjump */
 ret= jpeg_read_scanlines(&(rgb->cinfo),rgb->buffer,1);
 rgb->error_handle.reaction= 0;                       /* deactivate longjump */
 if(ret<=0) {
   sprintf(rgb->errmsg,"no scanline read in #%ld of %ld",
                       rgb->cinfo.output_scanline+1,rgb->cinfo.output_height);
   return(0);
 }
 if(!(flag&1))
    return(1);
 Imgrgb_app_line(rgb,0);
 return(1);
}


int Imgrgb_get_pixel(rgb,idx,red,green,blue,flag)
struct ImgrgB *rgb;
int idx;
double *red,*green,*blue;
int flag;
{
 if(idx>=rgb->cinfo.output_width) {
   sprintf(rgb->errmsg,"pixel index too large : %ld (max. %ld)",
                       idx,rgb->cinfo.output_width-1);
   return(0);
 }
 if(rgb->cinfo.output_components==3){
   *red= rgb->buffer[0][idx*3];
   *green= rgb->buffer[0][idx*3+1];
   *blue= rgb->buffer[0][idx*3+2];
 } else {
   *red= *green= *blue= rgb->buffer[0][idx];
 }
 return(1);
}


int Imgrgb_close(rgb,flag)
struct ImgrgB *rgb;
int flag;
{
 if(rgb->fp==NULL) {
   sprintf(rgb->errmsg,"Imgrgb object not active");
   return(0);
 }
 jpeg_destroy_decompress(&(rgb->cinfo));
 fclose(rgb->fp); rgb->fp= NULL;

 if(rgb->tempfile) 
   Imgrgb_temp_unlink(rgb,0);
#ifndef Imgrgb_line_buffer_liB
 if(rgb->buffer[0]!=NULL)
   Smem_freE((char *) rgb->buffer[0]);
 if(rgb->buffer!=NULL)
    Smem_freE((char *) rgb->buffer);
#endif
 rgb->buffer= NULL;
 if(rgb->red!=NULL)
   Smem_freE((char *) rgb->red);
 if(rgb->green!=NULL)
   Smem_freE((char *) rgb->green);
 if(rgb->blue!=NULL)
   Smem_freE((char *) rgb->blue);
 rgb->red= rgb->green= rgb->blue= NULL;
 return(1);
}


