
/*
 cc -g -c imgsamp.c
*/


#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"

#include "imgsamp.h"



/* ------------------------------- ImgsamplE ------------------------------ */


int Imgsample_new(sample,flag)
struct ImgsamplE **sample;
int flag;
{
 int i;
 struct ImgsamplE *s;

 *sample= s= TSOB_FELD(struct ImgsamplE,1);
 if(s==NULL)
   return(-1);
 s->invalid= 15;
 s->id= 0;
 s->samp_adr= 0;
 s->nam_adr= 0;
 s->filename= NULL;
 s->filename_is_reference= 0;
 s->crc= 0;
 s->spoints= Imgsamp_max_sizE;
#ifdef Imgsample_pointer_layeR
 Imgsample_point_rgb(s,s->rgb,0);
#endif /* Imgsample_pointer_layeR */
 for(i=0;i<Imgsamp_max_sizE;i++) 
   s->red[i]= s->green[i]= s->blue[i]= 0;
 s->prev= s->next= NULL;
 return(1);
}


int Imgsample_destroy(sample,flag)
struct ImgsamplE **sample;
int flag;
{
 if(*sample==NULL)
   return(0);
 if((*sample)->filename!=NULL && !(*sample)->filename_is_reference)
   Sregex_string(&((*sample)->filename),NULL,0);
 Smem_freE((char *) *sample);
 *sample= NULL;
 return(1);
}


int Imgsample_copy(source,target,flag)
struct ImgsamplE *source;
struct ImgsamplE *target;
int flag;
/*
 bit0= establish reference link to filename rather than copying it
*/
{
 int i,ret;

 target->invalid= source->invalid;
 target->id= source->id;
 target->samp_adr= source->samp_adr;
 target->nam_adr= source->nam_adr;
 ret= Imgsample_set_filename(target,source->filename,flag&1);
 if(ret<=0)
   return(ret);
 target->crc= source->crc;
 target->spoints= source->spoints;
 for(i=0;i<Imgsamp_max_sizE;i++) {
   target->red[i]= source->red[i];
   target->green[i]= source->green[i];
   target->blue[i]= source->blue[i];
 }
 return(ret);
}


int Imgsample_point_rgb(samp,data,flag)
struct ImgsamplE *samp;
Imgsamp_coorD *data;
int flag;
{
#ifdef Imgsample_pointer_layeR
 samp->red=   data+0*Imgsamp_max_sizE;
 samp->green= data+1*Imgsamp_max_sizE;
 samp->blue=  data+2*Imgsamp_max_sizE;
#endif /* Imgsample_pointer_layeR */
 return(1);
}


int Imgsample__set_coord(coord,value,flag)
Imgsamp_coorD *coord;
double value;
int flag;
{
 if(value>=Imgsamp_scalE)
   *coord= Imgsample_max_coorD;
 else if(value<0.0)
   *coord= 0;
 else
   *coord= value*Imgsamp_scalE;
 return(1);
}


double Imgsample_get_coord(sample,color,idx,flag)
struct ImgsamplE *sample;
int color; /* 0=red , 1=green , 2=blue */
int idx;
int flag;
{
 if(idx>=sample->spoints || idx<0)
   return(0.0);
 switch (color) {
 case 0: return(((double) sample->red[idx])/((double) Imgsamp_scalE));
 case 1: return(((double) sample->green[idx])/((double) Imgsamp_scalE));
 case 2: return(((double) sample->blue[idx])/((double) Imgsamp_scalE));
 }
 return(0.0);
}


int Imgsample_get_colorvector(sample,color,vector,flag)
struct ImgsamplE *sample;
int color; /* 0=red , 1=green , 2=blue */
Imgsamp_coorD **vector;
int flag;
{
 switch (color) {
        case 0: *vector= sample->red;
 break; case 1: *vector= sample->green;
 break; case 2: *vector= sample->blue;
 break; default: return(0);
 }
 return(1);
}


int Imgsample_set_rgb(sample,spoints,red,green,blue,flag)
struct ImgsamplE *sample;
int spoints;
double *red,*green,*blue;
int flag;
{
 int i;

 if(spoints>Imgsamp_max_sizE)
   return(0);
 sample->spoints= spoints;
 for(i=0;i<spoints;i++){
   Imgsample__set_coord(&(sample->red[i]),red[i],0);
   Imgsample__set_coord(&(sample->green[i]),green[i],0);
   Imgsample__set_coord(&(sample->blue[i]),blue[i],0);
 }
 sample->invalid&= ~8;
 return(1);
}


int Imgsample_set_nam_adr(sample,nam_adr,flag)
struct ImgsamplE *sample;
unsigned long nam_adr;
int flag;
{
 sample->nam_adr= nam_adr;
 sample->invalid&= ~2;
 return(1);
}


int Imgsample_set_samp_adr(sample,samp_adr,flag)
struct ImgsamplE *sample;
unsigned long samp_adr;
int flag;
/*
 bit0= set nam_adr to 0 and mark as valid too
*/
{
 sample->samp_adr= samp_adr;
 sample->invalid&= ~1;
 if(flag&1) {
   sample->nam_adr= 0;
   sample->invalid&= ~2;
 }
 return(1);
}


int Imgsample_set_id(sample,id,flag)
struct ImgsamplE *sample;
unsigned long id;
int flag;
{
 sample->id= id;
 return(1);
}


int Imgsample_set_filename(sample,filename,flag)
struct ImgsamplE *sample;
char *filename;
int flag;
/*
 bit0= establish reference link to filename rather than copying it
*/
{
 if(flag&1) {
   if(!sample->filename_is_reference)
     Sregex_string(&(sample->filename),NULL,0);
   sample->filename= filename;
   sample->filename_is_reference= 1;
 } else {
   if(sample->filename_is_reference)
     sample->filename= NULL;
   sample->filename_is_reference= 0;
   if(Sregex_string(&(sample->filename),filename,0)<=0)
     return(-1);
 }
 return(1);
}


int Imgsample_copy_adr(source,target,flag)
struct ImgsamplE *source,*target;
int flag;
{
 target->id= source->id;
 target->samp_adr= source->samp_adr;
 target->nam_adr= source->nam_adr;
 target->invalid&= ~(1|2);
 target->invalid|= source->invalid&(1|2);
 return(1);
}


int Imgsample_serialize(sample,text,flag)
struct ImgsamplE *sample;
char **text;               /* *text must be NULL or point to freaable memory */
int flag;
/*
 bit0= do not append filename (but still do append '+')
*/
{
 char *format= NULL;
 int ret, size, i;

 size= 20+20+sample->spoints*13+100;
 format= Smem_malloC(size);
 if(format==NULL)
   {ret= -1; goto ex;} 
 Sregex_string(text,NULL,0);
 sprintf(format,"-smp%ld-%8.8X",sample->spoints,sample->invalid|3);
 sprintf(format+strlen(format),"-%8.8X-%8.8X",sample->id,sample->crc);
 for(i=0;i<sample->spoints;i++) {
   sprintf(format+strlen(format),"-%4.4X%4.4X%4.4X",
                              sample->red[i],sample->green[i],sample->blue[i]);
 }
 *text= format;
 format= NULL;
 if(Sregex_string(text,"+",1)<=0)
   {ret= -1; goto ex;}
 if(!(flag&1) && sample->filename!=NULL)
   if(Sregex_string(text,sample->filename,1)<=0)
     {ret= -1; goto ex;}
 ret= 1;
ex:;
 if(format!=NULL)
   Smem_freE(format);
 return(ret);
}


int Imgsample__read_hex(rpt,value,length,flag)
char **rpt;
unsigned long value;
int length;
int flag;
/*
 bit0= expect preceding '-'
*/
{
 int ret;

 if(flag&1) {
   if(**rpt!='-')
     return(0);
   (*rpt)++;
 }
 ret= Sregex_hex_value(*rpt,value,length,0);
 if(ret<=0)
   return(ret);
 (*rpt)+= length;
 return(1);
}


int Imgsample_deserialize(sample,text,status,flag)
struct ImgsamplE *sample;
char *text;
int *status;
/*
  0= successfully completed
  1= trying to determine format
  2= trying to confirm format by reading 'invalid' , 'id' , 'crc'
  3= trying to read sample values
  4= trying to read file address
*/
int flag;
{
 int ret,i;
 unsigned long value;
 char expected[80],*rpt;

 *status= 1;
 rpt= text;
 sprintf(expected,"-smp%ld",sample->spoints);
 if(strncmp(expected,text,strlen(expected))!=0)
   {ret= 0; goto ex;}
 rpt+= strlen(expected);
 *status= 2;
 ret= Imgsample__read_hex(&rpt,&(sample->invalid),8,1);
 if(ret<=0)
   goto ex;
 ret= Imgsample__read_hex(&rpt,&(sample->id),8,1);
 if(ret<=0)
   goto ex;
 ret= Imgsample__read_hex(&rpt,&(sample->crc),8,1);
 if(ret<=0)
   goto ex;
 *status= 3;
 for(i=0;i<sample->spoints;i++) {
   ret= Imgsample__read_hex(&rpt,&value,4,1);
   if(ret<=0)
     goto ex;
   sample->red[i]= value;
   ret= Imgsample__read_hex(&rpt,&value,4,0);
   if(ret<=0)
     goto ex;
   sample->green[i]= value;
   ret= Imgsample__read_hex(&rpt,&value,4,0);
   if(ret<=0)
     goto ex;
   sample->blue[i]= value;
 }
 *status= 4;
 if(*rpt!='+')
   {ret= 0; goto ex;}
 ret= Imgsample_set_filename(sample,rpt+1,0);
 if(ret<=0)
   goto ex;
 *status= 0;
 ret= 1;
ex:;
 return(ret);
}


int Imgsample_smp17_format(samp,text,smp_size,avg,flag)
struct ImgsamplE *samp;
char text[108];
int *smp_size;
int avg[3];
int flag;
/*
 bit0= do not prepend revision header (and flags in native format)
 bit1= native format
 bit2= with bit1 : swap bytes when done
*/
{
 int i,j,k,coord_size,b,stdl;
 Imgsamp_coorD savg[3];
 char tr;

 coord_size= sizeof(Imgsamp_coorD);

 /* compute average pixel */
 for(j=0;j<3;j++)
   avg[j]= 0;
 for(i=0;i<samp->spoints;i++) {
   avg[0]+= samp->red[i];
   avg[1]+= samp->green[i];
   avg[2]+= samp->blue[i];
 }
 for(j=0;j<3;j++)
   avg[j]/= samp->spoints;

 stdl= 0;
 if(flag&2) {
   if(!(flag&1)) {
     *((unsigned short *) text)= 2;
     text[2]= text[3]= 0;
     stdl= 4;
   }
   if(coord_size!=2) {
     /* >>> buffer in 2-byte array and swap according to own bytesex */;
     /* <<< up to then : make noise */
     fprintf(stderr,"### imgsamp.c ### PANIC ### Wrong coord_size : %ld\n",
                    coord_size); 
     return(-1);
   }
   memcpy(text+stdl,samp->red,samp->spoints*2);
   stdl+= samp->spoints*2;
   memcpy(text+stdl,samp->green,samp->spoints*2);
   stdl+= samp->spoints*2;
   memcpy(text+stdl,samp->blue,samp->spoints*2);
   stdl+= samp->spoints*2;
   /* append average pixel */
   for(i=0;i<3;i++)
     savg[i]= avg[i];
   memcpy(text+stdl,savg,6);
   stdl+= 6;
   if(flag&4) {
     for(i=0;i<stdl;i+=2) {
       tr= text[i]; text[i]= text[i+1]; text[i+1]= tr;
     }
   }
 } else {
   if(!(flag&1)) {
     text[0]= 0;
     text[1]= 1;
     stdl= 2;
     /* write flags,id,crc */;
     for(i=0;i<4;i++) {
       text[stdl+i]= ((samp->invalid|3)>>(24-i*8))&0xff;
       text[stdl+4+i]= ((samp->id)>>(24-i*8))&0xff;
       text[stdl+8+i]= ((samp->crc)>>(24-i*8))&0xff;
     }
     stdl+= 12;
   }
   for(i=0;i<samp->spoints;i++) {
     for(b=0;b<coord_size;b++)
       text[stdl+i*3*coord_size+0*coord_size+b]=
                                     (samp->red[i]>>(8*(coord_size-1-b)))&0xff;
     for(b=0;b<coord_size;b++)
       text[stdl+i*3*coord_size+1*coord_size+b]=
                                   (samp->green[i]>>(8*(coord_size-1-b)))&0xff;
     for(b=0;b<coord_size;b++)
       text[stdl+i*3*coord_size+2*coord_size+b]=
                                    (samp->blue[i]>>(8*(coord_size-1-b)))&0xff;
   }
   /* append average pixel */
   stdl+= samp->spoints*3*coord_size;
   for(j=0;j<3;j++)
     for(b=0;b<coord_size;b++)
       text[stdl+j*coord_size+b]= (avg[j]>>(8*(coord_size-1-b)))&0xff;
   stdl+= 3*coord_size;
 }
 text[stdl]= 0;
 *smp_size= stdl;
 return(1);
}


int Imgsample_smp17_deformat_coord(samp,std,coord,flag)
struct ImgsamplE *samp;
char *std;
Imgsamp_coorD *coord;
int flag;
/*
 bit0= swapped
*/
{
 int b;

 if(sizeof(Imgsamp_coorD)==2) {
   if(flag&1)
     (*coord)= ((std[1]&0xff)<<8)|(std[0]&0xff);
   else
     (*coord)= ((std[0]&0xff)<<8)|(std[1]&0xff);
   return(1);
 }

 /* more general case for higher color resolutions */
 *coord= 0;
 for(b=0;b<sizeof(Imgsamp_coorD);b++) {
   (*coord)<<= 8;
   (*coord)|= std[b]&0xff;
 }
 if(flag&1) {
   char tr;
   for(b=0;b<sizeof(Imgsamp_coorD)/2;b++) {
     tr= ((char *) coord)[b];
     ((char *) coord)[b]= ((char *) coord)[sizeof(Imgsamp_coorD)-b];  
     ((char *) coord)[sizeof(Imgsamp_coorD)-b]= tr;
   }
 }
 return(1);
}


int Imgsample_smp17_deformat(samp,data,datalen,flag)
struct ImgsamplE *samp;
char *data;
int datalen;
int flag;
/*
 bit1= this is surely unswapped native format
*/{
 int ret,smp_size,swapped= 0,i;
 char *rpt,tr,revision[8];

 smp_size= 2+(samp->spoints+1)*3*sizeof(Imgsamp_coorD);
 if(flag&2)
   goto native_format;
 if(datalen<smp_size) 
   {ret= -1000-smp_size; goto ex;}
 revision[0]= data[0];
 revision[1]= data[1];
 if(revision[0]!=0) {
   swapped= 1;
   revision[1]= revision[0];
 }
 if(revision[1]==2)
   goto native_format;
 if(revision[1]!=1)
   {ret= -1-revision[1]; goto ex;} 

 /* ??? load flags,id,crc ?? */

 rpt= data+2+12;
 for(i=0;i<samp->spoints;i++) {
   Imgsample_smp17_deformat_coord(samp,rpt,&(samp->red[i]),swapped);
   rpt+= sizeof(Imgsamp_coorD);
   Imgsample_smp17_deformat_coord(samp,rpt,&(samp->green[i]),swapped);
   rpt+= sizeof(Imgsamp_coorD);
   Imgsample_smp17_deformat_coord(samp,rpt,&(samp->blue[i]),swapped);
   rpt+= sizeof(Imgsamp_coorD);
 }

 /* ??? what about the average pixel ? */

 ret= 1;
ex:;
 return(ret);

native_format:;
 if(sizeof(Imgsamp_coorD)!=2) {
   /* >>> buffer in 2-byte array and swap according to own bytesex */;
   /* <<< up to then : make noise */
   fprintf(stderr,"### imgsamp.c ### PANIC ### Wrong coord_size : %ld\n",
                  sizeof(Imgsamp_coorD)); 
   return(-1);
 }
 if(datalen<smp_size+2) 
   {ret= -1000-smp_size-2; goto ex;}

 /* ??? interpret flag bits in data[2] and data[3] */

 rpt= data+4;
#ifdef Imgsample_pointer_layeR
 memcpy(samp->rgb,rpt,samp->spoints*2*3);
 rpt+= samp->spoints*2*3;
#else /* Imgsample_pointer_layeR */
 memcpy(samp->red,rpt,samp->spoints*2);
 rpt+= samp->spoints*2;
 memcpy(samp->green,rpt,samp->spoints*2);
 rpt+= samp->spoints*2;
 memcpy(samp->blue,rpt,samp->spoints*2);
 rpt+= samp->spoints*2;
#endif /* ! Imgsample_pointer_layeR */

 /* ??? what about the average pixel ? */

 /* look wether it was not the own bytesex */
 if(*((unsigned short *) data)!=2) {
   for(i=0;i<samp->spoints;i++) {
     rpt= (char *) &(samp->red[i]);
     tr= *rpt; *rpt= *(rpt+1); *(rpt+1)= tr;
     rpt= (char *) &(samp->green[i]);
     tr= *rpt; *rpt= *(rpt+1); *(rpt+1)= tr;
     rpt= (char *) &(samp->blue[i]);
     tr= *rpt; *rpt= *(rpt+1); *(rpt+1)= tr;
   }
 }
 ret= 1;
 goto ex;
}


int Imgsample_decode(samp,data,datalen,message,flag)
struct ImgsamplE *samp;
char *data;
int datalen;
char message[256];
int flag;
{
 int ret,status,i;
 char text[2*SfileadrL];

 /* determine format */;
 if(datalen<96 || datalen>=sizeof(text))
   {ret= 0; goto ex;}
 if(data[0]==0||data[1]==0) {
   /* a binary format */;
   if(data[0]==1||data[1]==1) {
     /* standard binary , possibly swapped */;
     sprintf(message,"sample format : binary smp17");
     ret= Imgsample_smp17_deformat(samp,data,datalen,0);
     if(ret<=0)
       goto ex;
   } else if(data[0]==2||data[1]==2) {
     /* native color layers */;
     sprintf(message,"sample format : native smp17");
     ret= Imgsample_smp17_deformat(samp,data,datalen,2);
     if(ret<=0)
       goto ex;
   } else {
     ret= 0; goto ex;
   }
 } else if(strncmp(data,"-smp16-",7)==0) {
   /* do not assume data to be 0 terminated */
   for(i=0;i<datalen;i++)
     text[i]= data[i];
   text[datalen]= 0;
   sprintf(message,"sample format : ascii smp16");
   ret= Imgsample_deserialize(samp,text,&status,0);
   if(ret==0 && status==4)                    /* only fileaddress is missing */
     ret= 1;
   if(ret<=0) {
     goto ex;
   }
 } else {
   sprintf(message,"sample format not recognized");
   ret= 0; goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}



