

/*
 cc -g -c imvmover.c
*/

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"

#include "../as/asevent.h"
#include "imvguard.h"
#include "imvhistory.h"
#include "imvfiles.h"
#include "../as/asagent.h"
#include "imvapp.h"

#include "imvmover.h"


/* -------------------------------- ImvmoveR ------------------------------ */


int Imvmover_new(mover,agent,flag)
struct ImvmoveR **mover;
struct AsagenT *agent;
int flag;
{
 struct ImvmoveR *m;
 char *cpt,line[SfileadrL];
 int ret;

 *mover= m= TSOB_FELD(struct ImvmoveR,1);
 if(m==NULL)
   return(-1);
 m->recursion_level= 0;
 m->recursion_max= 1000;
 m->agent= agent;
 m->app= NULL;
 if(agent!=NULL)
   m->app= (struct ImvapP *) agent->app;
 m->event= NULL;
 m->history= NULL;
 m->files= NULL;
 m->current_adr[0]= 0;
 m->fstep= 0;

 ret= Imvfiles_new(&(m->files),m->agent,256,0);
 if(ret<=0)
   goto failed;

 return(1);
failed:;
 Imvmover_destroy(mover,0);
 return(ret);
}


int Imvmover_destroy(mover,flag)
struct ImvmoveR **mover;
int flag;
{
 if(*mover==NULL)
   return(0);
 Asevent_destroy(&((*mover)->event),0);
 Imvhistory_destroy(&((*mover)->history),0);
 Imvfiles_destroy(&((*mover)->files),0);
 Smem_freE((char *) *mover);
 *mover= NULL;
 return(1);
}


int Imvmover_reset_cursor(mover,flag)
struct ImvmoveR *mover;
int flag;
{
 Imvfiles_set_idx(mover->files,0,0);
 return(1);
}


int Imvmover_stdout(mover,text,flag)
struct ImvmoveR *mover;
char *text;
int flag;
/*
 bit0= append \n
*/
{
 int ret;
 ret= Asagent_stdout(mover->agent,text,flag&1);
 return(ret);
}


int Imvmover_stderr(mover,text,flag)
struct ImvmoveR *mover;
char *text;
int flag;
/*
 bit0= append \n
*/
{
 int ret;

 ret= Asagent_stderr(mover->agent,text,flag&1);
 return(ret);
}


int Imvmover_save_current_adr(mover,handle,flag)
struct ImvmoveR *mover;
char **handle;
int flag;
{
 int ret;

 ret= Sregex_string(handle,mover->current_adr,0);
 if(ret<=0)
   return(ret);
 return(1);  
}


int Imvmover_restore_current_adr(mover,handle,flag)
struct ImvmoveR *mover;
char **handle;
int flag;
{
 int ret;

 if(*handle==NULL)
   return(0);
 strcpy(mover->current_adr,*handle);
 Sregex_string(handle,NULL,0);
 return(1);
}


int Imvmover_show_current(mover,flag)
struct ImvmoveR *mover;
int flag;
/*
 bit0= print file info
*/
{
 int ret,idx;
 
 Imvfiles_get_idx(mover->files,&idx,0);
 ret= Asagent_show_current(mover->agent,mover->current_adr,idx,flag&1);
 return(ret);
}


int Imvmover_show_doubles(mover,flag)
struct ImvmoveR *mover;
int flag;
/*
 bit0= print file info
*/
{
 int ret;

 ret= Asagent_show_doubles(mover->agent,mover->current_adr,flag&1);
 return(ret);
}


int Imvmover_copy_to_trash(mover,prefix,flag)
struct ImvmoveR *mover;
char *prefix;
int flag;
/*
 bit1= insert new history object (to be undone after main one)
 bit7+8= overwrite_and_delete (see Imvmover_cmd_moveto)
*/
{
 char *adr,*basename,errmsg[SfileadrL],target[SfileadrL];
 int ret,errn;

 adr= mover->current_adr;
 basename= strrchr(adr,'/');
 if(basename==NULL) {
   basename= adr;
 }else{
   basename++;
 }
 sprintf(target,"%s/%s%s",mover->agent->trashdir,prefix,basename);
 Sfile_unique_adr(target,0);
 ret= Sfile_copy_file(adr,target,&errn,errmsg,0);
 if(ret<=0) {
   Imvmover_stderr(mover,"--- while copying to trashdir :",1);
   Imvmover_stderr(mover,errmsg,1);
   return(ret);
 }
 Imvhistory_set_replaced_by(mover->history,NULL,8);
 if(flag&2)
   ret= Imvmover_push_history(mover,mover->current_adr,target,flag&(128|256));
 else {
   ret= Imvhistory_set_current(mover->history,target,0);
   Imvhistory_set_overwrite_and_delete(mover->history,(flag>>7)&3);
 }
 if(ret<=0)
   return(ret);
 return(1);
}


int Imvmover_jpgb_convert(mover,flag)
struct ImvmoveR *mover;
int flag;
/*
 bit0= just return the effective quality value
*/
{
 char cmd[SfileadrL],*adr;
 int ret,quality,status;

 adr= mover->current_adr;
 Sfile_home_adr(".imv_jpgb_quality_value",cmd,0);
 ret= Sfile_file_to_int(cmd,&quality,2);
 if(ret<=0)
   quality= 80;
 if(quality<25)
   quality= 25;
 if(quality>98)
   quality= 98;
 if(flag&1)
   return(quality);
 sprintf(cmd,"convert -comment '' -quality %ld ",quality);
 Sfile_esc_shell(adr,cmd+strlen(cmd),0);
 strcat(cmd," jpeg:");
 Sfile_esc_shell(adr,cmd+strlen(cmd),0);
 Imvmover_stdout(mover,cmd,1);
 Asagent_shell(mover->agent,cmd,&status,0);
 return(1);
}


int Imvmover_push_history(mover,source_adr,current_adr,flag)
struct ImvmoveR *mover;
char *source_adr,*current_adr;
int flag;
/*
 for Imvhistory_new :
 bit1= checked_in
 bit2= checked_out
 bit3= replaced by a descendent of itself
 bit4= no undo possible with this operation
 bit5= nothing has happened in this operation
 bit7+8= overwrite_and_delete (see Imvmover_cmd_moveto)
*/
{
 struct ImvhistorY *history;
 int ret,idx;

 ret= Imvfiles_get_idx(mover->files,&idx,0);
 if(ret<=0)
   return(ret);
 ret= Imvhistory_new(&history,idx,source_adr,current_adr,NULL,mover->event,
                     flag&(63|128|256));
 if(ret<=0)
   return(ret);
 Imvhistory_link_sub(mover->history,history,0);
 return(1);
}


int Imvmover_eval(mover,in_text,result_pt,flag)
struct ImvmoveR *mover;
char *in_text;
char **result_pt;
int flag;
/*
 bit0= evaluation for shell purposes
*/
{
 int ret;

 ret= Asagent_eval(mover->agent,in_text,result_pt,flag&1);
 return(ret);
}


int Imvmover_cmd_undo(mover,flag)
struct ImvmoveR *mover;
int flag;
/*
 return:
  <=0 error
  1   undone
  2   nothing done
*/
{
 int ret,idx;
 struct ImvhistorY *history= NULL;
 char *cmd= NULL,*target= NULL,*current_adr= NULL;

 ret= Asagent_get_permission(mover->agent,"undo","undo",0);
 if(ret<=0)
   goto ex;
 Imvmover_stdout(mover,"\nundo :\n",1);
 Imvhistory_set_no_undo_possible(mover->history,1,0);
 ret= Imvfiles_pop_history(mover->files,&history,0);
 if(ret<0)
   return(ret);
 if(ret==0) {
   Imvmover_stderr(mover,"--- all actions undone as far as possible.",1);
   return(2);
 }
 Imvhistory_get_idx(history,&idx,0);
 Imvfiles_set_idx(mover->files,idx,0);
 ret= Imvfiles_get_adr(mover->files,mover->current_adr,0);
 if(ret<=0)
   goto ex;
 if(Imvhistory_no_undo_possible(history,0)) {
   Imvmover_stderr(mover,"--- sorry, this action cannot be undone :",1);
   Imvmover_stderr(mover,"    ",0);
   Asevent_get_cmd(history->event,&cmd,0);
   if(cmd!=NULL) 
     Imvmover_stderr(mover,cmd,1);
   else
     Imvmover_stderr(mover,"",1);
   ret= 2; goto ex;
 }
 if(Imvhistory_nothing_happened(history,0))
   { ret= 2; goto ex; }

 /* check wether the current file list item is also the undo subject */
 ret= Imvhistory_get_current(history,&current_adr,0);
 if(ret<=0)
   goto ex;
 if(strcmp(current_adr,mover->current_adr)==0) {
   ret= Imvhistory_get_source(history,&target,0);
   if(ret<=0)
     goto ex;
 } else
   Sregex_string(&target,NULL,0);

 ret= Imvmover_perform_undo(mover,history,0);
 if(ret<=0)
   goto ex;

 /* if the current file list item was also the undo subject */
 if(target!=NULL){
   ret= Imvfiles_set_adr(mover->files,target,0);
   if(ret<=0)
     goto ex;
 }

 ret= 1;
ex:;
 Sregex_string(&cmd,NULL,0);
 Sregex_string(&target,NULL,0);
 Sregex_string(&current_adr,NULL,0);
 Imvhistory_destroy(&history,0);
 return(ret);
}


int Imvmover_perform_undo(mover,history_in,flag)
struct ImvmoveR *mover;
struct ImvhistorY *history_in;
int flag;
/*
 return:
  <=0 error
  1   undone
  2   nothing done
*/
{
 struct ImvhistorY *history;
 char *current_adr= NULL,*old_adr= NULL,*saved_adr= NULL;
 int ret,ovw_del;

 history= history_in;

 for(history= history_in ; history!=NULL ; history= history->sub) {
   ret= Imvhistory_get_current(history,&current_adr,0);
   if(ret<=0)
     goto ex;
   ret= Imvhistory_get_source(history,&old_adr,0);
   if(ret<=0)
     goto ex;
   ovw_del= Imvhistory_get_overwrite_and_delete(history,0);
   /* undo a overwrite+del by a copy, undo a copy by a overwrite+del */
   if(ovw_del!=0)
     ovw_del= 3-ovw_del;
   ovw_del= ovw_del<<7;
   ret= Imvmover_save_current_adr(mover,&saved_adr,0);
   if(ret<=0)
     goto ex;
   strcpy(mover->current_adr,current_adr);
   ret= Imvmover_cmd_moveto(mover,old_adr,4|8|64|ovw_del);
   Imvmover_restore_current_adr(mover,&saved_adr,0);
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:;
 Sregex_string(&current_adr,NULL,0);
 Sregex_string(&saved_adr,NULL,0);
 Sregex_string(&old_adr,NULL,0);
 return(ret);
}


int Imvmover_noop(mover,flag)
struct ImvmoveR *mover;
int flag;
{
 int ret;

 Imvhistory_set_nothing_happened(mover->history,1,0);
 return(1);
}


int Imvmover_cmd_jpgb(mover,flag)
struct ImvmoveR *mover;
int flag;
{
 char *adr,copy_adr[SfileadrL];
 int ret;
 struct ImvguarD *guard;

 guard= mover->app->guard;
 adr= mover->current_adr;
 ret= Imvguard_open_adr(guard,adr,0);
 if(ret<=0)
   return(ret);
 ret= Imvmover_copy_to_trash(mover,"_imvjpgb_",0);
 if(ret<=0)
   return(ret);
 ret= Imvmover_jpgb_convert(mover,0);
 if(ret<=0)
   return(ret);
 ret= Imvguard_refresh(guard,"",1);
 if(ret<=0)
   return(ret);
 return(1);
}


int Imvmover_cmd_preserve(mover,flag)
struct ImvmoveR *mover;
int flag;
{
 int ret;

 ret= Asagent_get_permission(mover->agent,"move","preserve",0);
 if(ret<=0)
   return(ret);
 ret= Imvmover_copy_to_trash(mover,"_preserve_",0);
 if(ret<=0)
   return(ret);
 return(1);
}


int Imvmover_cursor_advance(mover,flag)
struct ImvmoveR *mover;
int flag;
/*
 bit0= mover->fstep is the absolute index
*/
{
 int ret;

 ret= Imvfiles_set_idx(mover->files,mover->fstep,!(flag&1));
 mover->fstep= 0;
 return(ret);
}


int Imvmover_cmd_rm(mover,flag)
struct ImvmoveR *mover;
int flag;
/*
  bit1= this is a sidekick: do not manipulate mover->files
  bit2= do not ask for permission to delete the file
*/
{
 char *adr;
 char cmd[SfileadrL];
 int ret,yes,status;
 struct ImvguarD *guard;

 guard= mover->app->guard;
 adr= mover->current_adr;
 sprintf(cmd,"rm -f %s",adr);
 if(!(flag&4)){
   Imvmover_stderr(mover,"???   ",0);
   Imvmover_stderr(mover,cmd,0);
   Imvmover_stderr(mover,"   ???",1);
   Imvmover_stdout(mover,"Do you really want to delete this file forever ?",1);
   ret= Asagent_fetch_yesno(mover->agent,
                             "Do you really want to delete this file forever ?",
                             &yes,1);
   if(ret<=0)
     return(ret);
   if(!yes)
     return(2);
 } else
   Imvmover_stderr(mover,cmd,1);
 ret= Imvguard_check_out(guard,adr,0);
 if(ret<=0)
   return(ret);
 if(!(flag&2))
   Imvfiles_set_invalid(mover->files,0);
 Imvhistory_set_no_undo_possible(mover->history,1,0);
 Asagent_shell(mover->agent,cmd,&status,0);
 return(1);
}


int Imvmover_fetch_collision(mover,new_adr,reply,flag)
struct ImvmoveR *mover;
char *new_adr;
char *reply;
int flag;
/*
 bit0= do really ask user (else: answer autorename) 
*/
{
 int ret,c,asked=0;
 char question[SfileadrL],*agent_reply= NULL;

 if(!(flag&1)) {
   strcpy(reply,"autorename");
   ret= 1;
   goto ex;
 }  
ask:;
 Asagent_ask(mover->agent,"char",
              "Rename Autorename -Delete Overwrite ^back","RrAa-DdOo^",1);
 asked= 1;
 ret= Asagent_fetch_token(mover->agent,&agent_reply,1|2|4);
 Asagent_ask(mover->agent,"cancel","","",2);
 if(ret<=0)
   goto ex;
 c= agent_reply[0];
 if(c=='R' || c=='r') {
   strcpy(reply,"rename ");
   sprintf(question,"New name for  %s  ?",new_adr);
   ret= Asagent_fetch_line(mover->agent,question,&agent_reply,0);
   if(ret<=0)
     goto ex;
   strcat(reply,agent_reply);
 } else if(c=='A' || c=='a') {
   strcpy(reply,"autorename");
 } else if(c=='-' || c=='D' || c=='d') {
   strcpy(reply,"delete");
 } else if(c=='O' || c=='o') {
   strcpy(reply,"overwrite");
 } else if(c=='^') {
   strcpy(reply,"redo");
/*
 } else if(c=='_') {
   strcpy(reply,"rm");
*/
 } else {
   if(c>=32)
     sprintf(question,"undefined reply : '%c'  %ld  0x%-2.2X",c,c,c);
   else
     sprintf(question,"undefined reply : ctl  %ld  0x%-2.2X",c,c);
   Imvmover_stderr(mover,question,1);
   goto ask;
 }

 ret= 1;
ex:;
 Sregex_string(&agent_reply,NULL,0);
 return(ret);
}


int Imvmover_target_collision(mover,new_adr,flag)
struct ImvmoveR *mover;
char *new_adr;
int flag;
/*
 flag:
  bit0= do not ask but use autorename
  bit1= this is a sidekick: do not manipulate mover->files
                            make additional history object
  bit2= do not ask but use overwrite_by_rm
  bit6= sidekick without history
 return values: 
  <=0 = error
  1 = retry moving with collision test
  2 = redo user interaction
  3 = snafu (nothing to be done with this item any more)
*/
{
 int ret=0,appnum,status;
 char adr[SfileadrL],reply[SfileadrL],line[SfileadrL],esc_file[SfileadrL];
 char appendix[80],*saved_adr;
 struct stat stbuf;

 if(flag&1) {
   strcpy(reply,"autorename");
 } else if(flag&4) {
   strcpy(reply,"overwrite_by_rm");
 } else {
   Imvmover_stderr(mover,"Already existing :",1);
   sprintf(line,"ls -l %s >&2",Sfile_esc_shell(new_adr,esc_file,0));
   Asagent_shell(mover->agent,line,&status,0);
   Imvmover_stderr(mover,"",0);                    /* flush text output */
   ret= Imvmover_fetch_collision(mover,new_adr,reply,0);
   if(ret<=0)
     return(0);
 }
 if(strcmp(reply,"autorename")==0) {
   appnum= time(0);
   do {
     strcpy(adr,new_adr);
     sprintf(appendix,"_%-8.8x",appnum++);
     Sfile_addto_filename(adr,appendix,0);
   } while(lstat(adr,&stbuf)!=-1);
   strcpy(new_adr,adr);
   ret= 1; 
   goto ex;
 } else if(strcmp(reply,"delete")==0) {
   ret= Imvmover_cmd_moveto(mover,mover->agent->trashdir,1|(flag&(2|64)));
   if(ret<=0)
     goto ex;
   ret= 3;
   goto ex;
 } else if(strcmp(reply,"overwrite")==0) {
   ret= Imvmover_save_current_adr(mover,&saved_adr,0);
   if(ret<=0)
     goto ex;
   strcpy(mover->current_adr,new_adr);
   Imvmover_stdout(mover,"moving file to trashdir : ",0);
   Imvmover_stdout(mover,new_adr,1);
   ret= Imvmover_cmd_moveto(mover,mover->agent->trashdir,1|2|(flag&64));
   Imvmover_restore_current_adr(mover,&saved_adr,0);
   if(ret<=0)
     goto ex;
   ret= 1; 
   goto ex;
 } else if(strcmp(reply,"overwrite_by_rm")==0) {
   unlink(new_adr);
   ret= 1;
   goto ex;
 } else if (strcmp(reply,"redo")==0) {
   ret= 2;
   goto ex;
 } else if (strncmp(reply,"rename ",strlen("rename "))==0) {
   strcpy(new_adr,reply+strlen("rename "));
   ret= 1;
   goto ex;
 } else if (strcmp(reply,"rm")==0){
   ret= Imvmover_cmd_rm(mover,flag&(2|4));
   if(ret<=0)
     goto ex;
   ret= 3;
   goto ex;
 }
ex:;
 return(ret);
}


int Imvmover_cmd_moveto(mover,in_target,flag)
struct ImvmoveR *mover;
char *in_target;
int flag;
/*
 flag:
  bit0= on collision, use autorename
  bit1= this is a sidekick: do not manipulate mover->files...[idx]
                            insert new history object under current history
  bit2= permission to unlink target file if existing
  bit3= target is the complete new adress
  bit4= report in format :  source > target
  bit5= accept in_target without interpretation
  bit6= sidekick without history (see bit1)
  bit7+8= overwrite_and_delete :
        0= rename
        1= use overwrite+delete rather than rename
           permission to overwrite existing files
           ??? preserve copy of original target
        2= overwrite target but do not unlink source
           (i.e. copy rather than move)

 return values:
  <=0 = error
  1 = all went normal
  2 = redo user interaction
*/
{
 int ret,count=0,errn,preserve_basename,idx,sflag;
 char *adr,new_adr[SfileadrL],*basename,line[2*SfileadrL+160],dir[SfileadrL];
 char *target= NULL;
 struct stat stbuf;
 ino_t inode_new;
 struct ImvguarD *guard;

 guard= mover->app->guard;
 if(flag&32)
   ret= Sregex_string(&target,in_target,0);
 else
   ret= Imvmover_eval(mover,in_target,&target,0);
 if(ret<=0)
   goto ex;

 adr= mover->current_adr;
 basename= strrchr(adr,'/');
 if(basename==NULL)
   basename= adr;
 else
   basename++;
 if(basename[0]=='.') {
   Imvmover_stderr(mover,"hidden filename (leading dot) may not move :",1);
   Imvmover_stderr(mover,basename,1);
   ret= 0; goto ex;
 }

 if(flag&8) {
   strcpy(new_adr,target);
 } else {
   preserve_basename= 0;
   if(stat(target,&stbuf)!=-1) 
     if(S_ISDIR(stbuf.st_mode)) 
       preserve_basename= 1;
   if(preserve_basename) {
     sprintf(new_adr,"%s/%s",target,basename);
   } else {
     strcpy(new_adr,target);
   }
 }
try_move:;
 if(Imvfiles_is_bad_name(mover->files,new_adr,0)){
   Imvmover_stderr(mover,"illegal filename given as move target :",1);
   Imvmover_stderr(mover,new_adr,1);
   ret= 0; goto ex;
 }
 count++;
 if(lstat(new_adr,&stbuf)!=-1) {
   inode_new= stbuf.st_ino;
   if(lstat(mover->current_adr,&stbuf)!=-1)
     if(inode_new==stbuf.st_ino) {
       Imvmover_stderr(mover,"--- note : source and target are identical",1);
       Imvmover_stderr(mover,"   ",0);
       Imvmover_stderr(mover,mover->current_adr,1);
       Imvmover_stderr(mover," > ",0);
       Imvmover_stderr(mover,new_adr,1);
       ret= 1;
       goto ex;
     }
   if(flag&128) {

     /* ??? make a copy */

     if(!(flag&64))
       /* with sub-history */;

   } else {
     if((flag&(1|4))&&count>1) {
       Imvmover_stderr(mover,"--- cannot circumvent existing file ",0);
       Imvmover_stderr(mover,new_adr,1);
       ret= 0; goto ex;
     }
     ret= Imvmover_target_collision(mover,new_adr,flag&7);
     if(ret<=0)
       goto ex;
     if(ret==1) 
       goto try_move;
     if(ret==2)
       goto ex;
     if(ret==3) {
       ret= 1;
       goto ex;
     }
   }
 }
 if(flag&4) {
   Imvmover_stdout(mover,mover->current_adr,0);
   Imvmover_stdout(mover,"  >  ",0);
   Imvmover_stdout(mover,new_adr,1);
 }
 ret= Imvguard_check_out(guard,adr,2);
 if(ret<=0)
   goto ex;
 sflag= 0;
 if(flag&128)
   sflag|= 128|1;
 sflag|= flag&256;
 ret= Sfile_move_file(adr,new_adr,&errn,line,sflag);
 if(ret<=0){
   Imvmover_stderr(mover,line,1);
   Imvguard_check_in(guard,adr,4);
   ret= 0;
   goto ex;
 }
 if(ret==2 && !(flag&128))
   Imvmover_stdout(mover,
                  "note : cross filesystem move performed by copy+rm",1);
 if(flag&64) {
   ; /* sidekick without history */
 } else if(flag&2) {
   ret= Imvmover_push_history(mover,mover->current_adr,new_adr,0);
   if(ret<=0)
     goto ex;
 } else {
   ret= Imvhistory_set_current(mover->history,new_adr,0);
   if(ret<=0)
     goto ex;
   Imvhistory_set_overwrite_and_delete(mover->history,(flag>>7)&3,0);
 }
 strcpy(mover->current_adr,new_adr);
 if(!(flag&(2|64))){
   ret= Imvfiles_set_adr(mover->files,new_adr,0);
   if(ret<=0)
     goto ex;
 }
 if(count>1 && (flag&4))
   ret= Imvguard_refresh(guard,new_adr,0);
 else
   ret= Imvguard_check_in(guard,new_adr,2);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:
 Sregex_string(&target,NULL,0);
 return(ret);
}


int Imvmover_cmd_doubleto(mover,target,flag)
struct ImvmoveR *mover;
char *target;
int flag;
/*
 bit5= accept target without interpretation
*/
{
 int ret,r_ret;
 char *saved_adr= NULL,*source= NULL;
 struct ImvhistorY *history_sub= NULL;

 ret= Asagent_get_first_double(mover->agent,mover->current_adr,&source,0);
 if(ret==0)
   {ret= 2; goto ex;}
 if(ret<0)
   goto ex;
 ret= Imvmover_save_current_adr(mover,&saved_adr,0);
 if(ret<=0)
   goto ex;
 strcpy(mover->current_adr,source);
 ret= Imvmover_cmd_moveto(mover,target,2|(flag&32));
 r_ret= Imvmover_restore_current_adr(mover,&saved_adr,0);
 if(r_ret<ret)
   ret= r_ret;
 if(ret<=0)
   goto ex;
 history_sub= mover->history->sub;
 if(history_sub!=NULL) {
   mover->history->sub= NULL;
   Imvhistory_destroy(&(mover->history),0);
   mover->history= history_sub;
   mover->history->boss= NULL;
 }
 ret= 1;
ex:;
 Sregex_string(&source,NULL,0);
 Sregex_string(&saved_adr,NULL,0);
 return(1);
}


int Imvmover_cmd_replace_double(mover,flag)
struct ImvmoveR *mover;
int flag;
{
 int ret,r_ret;
 char *saved_adr= NULL,*source= NULL,*target= NULL;

 ret= Asagent_get_permission(mover->agent,"move","replace_double",0);
 if(ret<=0)
   goto ex;

 ret= Asagent_get_first_double(mover->agent,mover->current_adr,&source,0);
 if(ret==0)
   {ret= 2; goto ex;}
 if(ret<0)
   goto ex;
 ret= Imvmover_save_current_adr(mover,&saved_adr,0);
 if(ret<=0)
   goto ex;
 strcpy(mover->current_adr,source);
 ret= Imvmover_copy_to_trash(mover,"_replace_",2|256);
 if(ret<=0)
   return(ret);
 ret= Imvguard_check_out(mover->app->guard,source,0);
 if(ret<=0)
   goto ex;
 ret= Imvmover_restore_current_adr(mover,&saved_adr,0);
 if(ret<=0)
   goto ex;
 ret= Imvmover_cmd_moveto(mover,source,32|128);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 Sregex_string(&target,NULL,0);
 Sregex_string(&source,NULL,0);
 Sregex_string(&saved_adr,NULL,0);
 return(ret);
}





int Imvmover_determine_current(mover,idx,count,flag)
struct ImvmoveR *mover;
int *idx;
int *count;
int flag;
{
 int ret;

 *count= mover->files->count;
 ret= Imvfiles_get_idx(mover->files,idx,0);
 if(ret<=0) {
   ret= 2; /* no files */
   goto ex;
 } else {
   ret= Imvfiles_get_adr(mover->files,mover->current_adr,0);
   if(ret<=0)
     goto ex;
   if(mover->current_adr[0]==0) {
     ret= 0;
     goto ex;
   }
 }
 ret= 1;
ex:;
 return(ret);
}


int Imvmover_determine_index(mover,text,reply_idx,flag)
struct ImvmoveR *mover;
char *text;
int *reply_idx;
int flag;
/*
flag:
 bit0= reply relative to current item's index 
 bit1= do not accept absolute index <0 
 bit2= do not accept absolute index >=files->count
 bit3= clip to 0 <= index <files->count  (do not accept if files->count<=0)
return:
 <=0 = error
   1 = index was given absolute 
   2 = index was given relative
*/
{
 int ret,jump_rel= 0,jump_target=0,idx,absolute_target;

 ret= Imvfiles_get_idx(mover->files,&idx,0);
 if(ret<=0) 
   goto ex;
 if(text[0]=='+') {
   jump_rel= 1;
   sscanf(text+1,"%ld",&jump_target);
 } else {
   sscanf(text,"%ld",&jump_target);
 }
 if(jump_target<=0)
   jump_rel= 1;

 if(jump_rel)
   absolute_target= idx+jump_target;
 else
   absolute_target= jump_target-1;         /* absolute jump_target is base 1 */
 if((flag&1) && absolute_target<0)
   {ret= 0; goto ex;}
 if((flag&2) && absolute_target>=mover->files->count)
   {ret= 0; goto ex;}
 if(flag&8){
   if(mover->files->count<=0)
     {ret= 0; goto ex;}
   if(absolute_target<0)
     absolute_target= 0; 
   else if(absolute_target>=mover->files->count)
     absolute_target= mover->files->count- 1;
 }
 if(flag&1)
   *reply_idx= absolute_target-idx;
 else
   *reply_idx= absolute_target;
 ret= 1+jump_rel;
ex:;
 return(ret);
}


int Imvmover_cmd_jump(mover,target,flag)
struct ImvmoveR *mover;
char *target;
int flag;
{
 int ret,idx;

 ret= Asagent_get_permission(mover->agent,"navigation","jump",0);
 if(ret<=0)
   goto ex;
 ret= Imvmover_determine_index(mover,target,&idx,1);
 if(ret<=0)
   goto ex;
 mover->fstep= idx;
 if(mover->fstep==0)
   mover->agent->keep_views= 1;
 ret= 1;
ex:;
 return(ret);
}


int Imvmover_cmd_list(mover,in_cmd,flag)
struct ImvmoveR *mover;
char *in_cmd;
int flag;
{
 int ret,start_idx,end_idx,i,line_numbers= 1;
 char *start_text= NULL,*end_text= NULL,*option= NULL,*cpt,fileadr[SfileadrL];
 char index_text[81];

 ret= Asagent_get_permission(mover->agent,"info","list",0);
 if(ret<=0)
   goto ex;

 cpt= in_cmd;
 ret= Sregex_cut_neck(cpt,':',&start_text,&cpt,0);
 if(ret<0)
   goto ex;
 if(ret==0)
   if(Sregex_string(&end_text,"+0",0)<=0)
     {ret= -1; goto ex;}
 ret= Sregex_cut_neck(cpt,':',&end_text,&cpt,2);
 if(ret<0)
   goto ex;
 if(ret==0)
   if(Sregex_string(&end_text,start_text,0)<=0)
     {ret= -1; goto ex;}
 ret= Sregex_cut_neck(cpt,':',&option,&cpt,2);
 if(ret<0)
   goto ex;
 ret= Imvmover_determine_index(mover,start_text,&start_idx,8);
 if(ret<=0)
   goto ex;
 ret= Imvmover_determine_index(mover,end_text,&end_idx,8);
 if(ret<=0)
   goto ex;
 if(option!=NULL) {
   if(strcmp(option,"raw")==0)
     line_numbers= 0;
   else if(strcmp(option,"count")==0)
     line_numbers= 2;
 }   
 for(i= start_idx;i<=end_idx;i++) {
   ret= Imvfiles_get_idxadr(mover->files,i,fileadr,0);
   if(ret<=0)
     goto ex;
   if(line_numbers){
     if(line_numbers>1)
       sprintf(index_text,"%ld %ld ",i+1,mover->files->count);
     else
       sprintf(index_text,"%ld ",i+1);
     Asagent_stdout(mover->agent,index_text,0);
   }
   Asagent_stdout(mover->agent,fileadr,1);
 }
 ret= 1;
ex:;
 mover->agent->keep_views= 1;
 Sregex_string(&start_text,NULL,0);
 Sregex_string(&end_text,NULL,0);
 Sregex_string(&option,NULL,0);
 return(ret);
}


int Imvmover_cmd_for(mover,in_cmd,flag)
struct ImvmoveR *mover;
char *in_cmd;
int flag;
/*
 bit1= do not complain loudely about unknown command

return: see Imvmover_interpreter()
*/
{
 int ret,start_idx,end_idx,i,idx,old_idx,hret,count,pre_idx;
 char *cmd= NULL,*start_text= NULL,*end_text= NULL,*cpt;
 struct AsagenT *agent;
 struct LstrinG *mark_pt;

 agent= mover->agent;
 ret= Imvfiles_get_idx(mover->files,&old_idx,0);
 if(ret<=0) 
   goto ex;

 /* start_idx:end_idx:cmd */
 ret= Asagent_get_permission(agent,"loop","for",0);
 if(ret<=0)
   return(ret);
 cpt= in_cmd;
 ret= Sregex_cut_neck(cpt,':',&start_text,&cpt,0);
 if(ret<0)
   goto ex;
 if(ret==2 || strlen(cpt)==0) {
malformed:;
   Asagent_stderr(agent,"--- malformed options to -for : ",0);
   Asagent_stderr(agent,in_cmd,1);
   ret= 0; goto ex;
 }
 ret= Sregex_cut_neck(cpt,':',&end_text,&cpt,0);
 if(ret<0)
   goto ex;
 if(ret!=1)
   goto malformed;
 if(Sregex_string(&cmd,cpt,0)<=0)
   {ret= -1; goto ex;}

 ret= Imvmover_determine_index(mover,start_text,&start_idx,4|8);
 if(ret<=0)
   goto ex;
 ret= Imvmover_determine_index(mover,end_text,&end_idx,2|8);
 if(ret<=0)
   goto ex;
 mark_pt= agent->markstack->next;
 for(i=start_idx;i<=end_idx;i++){
   Imvfiles_get_idx(mover->files,&pre_idx,0);
   Imvfiles_set_idx(mover->files,i,0);
   ret= Imvmover_determine_current(mover,&idx,&count,0);
   if(ret<=0)
     goto ex;
   if(pre_idx!=idx) {
     Asagent_stdout(agent,"",1);
     Asagent_stdout_bar(agent,0);
     Asagent_print_fileinfo(agent,mover->current_adr,0);
   }
   Asagent_notify_mark(agent,"",mark_pt,2|4);
   ret= Asagent_execute_subcmd(agent,cmd,flag&2);
   if(ret!=1)
     goto ex;
 }
ex:;
 hret= Imvfiles_get_idx(mover->files,&idx,0);
 if(hret>0 && idx!=old_idx)
   agent->keep_views= 0;
 else if(ret>0 && cmd[0]==0)
   agent->keep_views= 1;
 Sregex_string(&cmd,NULL,0);
 Sregex_string(&start_text,NULL,0);
 Sregex_string(&end_text,NULL,0);
 return(ret);
}


int Imvmover_cmd_where(mover,in_cmd,flag)
struct ImvmoveR *mover;
char *in_cmd;
int flag;
/*
 bit1= do not complain loudely about unknown command

return: see Imvmover_interpreter()
*/
{
 int ret,start_idx,end_idx,i,idx,old_idx,hret,count,last_idx= -1,mode= 0;
 char *cmd= NULL,*expr_text= NULL,*cpt,*mode_text= NULL;
 char adr[SfileadrL],dummy[SfileadrL];
 struct AsagenT *agent;
 struct LstrinG *mark_pt;
 struct SregeX *re= NULL;

 agent= mover->agent;
 ret= Imvfiles_get_idx(mover->files,&old_idx,0);
 if(ret<=0) 
   goto ex;

 /* mode:regular_expression:cmd */
 ret= Asagent_get_permission(agent,"loop","where",0);
 if(ret<=0)
   return(ret);
 cpt= in_cmd;
 ret= Sregex_cut_neck(cpt,':',&mode_text,&cpt,0);
 if(ret<0)
   goto ex;
 if(ret==2 || strlen(cpt)==0) {
malformed:;
   Asagent_stderr(agent,"--- malformed options to -where : ",0);
   Asagent_stderr(agent,in_cmd,1);
   ret= 0; goto ex;
 }
 ret= Sregex_cut_neck(cpt,':',&expr_text,&cpt,0);
 if(ret<0)
   goto ex;
 if(ret!=1)
   goto malformed;
 if(Sregex_string(&cmd,cpt,0)<=0)
   {ret= -1; goto ex;}
 if(strcmp(mode_text,"adr")==0)
   mode= 0;
 else if(strcmp(mode_text,"name")==0)
   mode= 1;
 else if(strcmp(mode_text,"dir")==0)
   mode= 2;

 ret=  Sregex_new(&re,expr_text,0);
 if(ret<=0)
   goto ex;
 /* memorize mark stack position and in loop release up to that position */
 mark_pt= agent->markstack->next;
 for(i=0;i<mover->files->count;i++){
   Imvfiles_set_idx(mover->files,i,0);
   ret= Imvmover_determine_current(mover,&idx,&count,0);
   if(ret<=0)
     goto ex;
   if(mode==1)
     Sfile_split_adr(mover->current_adr,dummy,adr,0);
   else if(mode==2)
     Sfile_split_adr(mover->current_adr,adr,dummy,0);
   else
     strcpy(adr,mover->current_adr);
   if(!Sregex_is_match(re,adr,0))
 continue;
   last_idx= i;
   Asagent_stdout(agent,"",1);
   Asagent_stdout_bar(agent,0);
   Asagent_print_fileinfo(agent,mover->current_adr,0);
   Asagent_notify_mark(agent,"",mark_pt,2|4);
   ret= Asagent_execute_subcmd(agent,cmd,flag&2);
   if(ret!=1)
     goto ex;
 }
 if(last_idx>=0)
   Imvfiles_set_idx(mover->files,last_idx,0);
ex:;
 hret= Imvfiles_get_idx(mover->files,&idx,0);
 if(hret>0 && idx!=old_idx)
   agent->keep_views= 0;
 else if(ret>0 && cmd[0]==0)
   agent->keep_views= 1;
 Sregex_destroy(&re,0);
 Sregex_string(&cmd,NULL,0);
 Sregex_string(&expr_text,NULL,0);
 Sregex_string(&mode_text,NULL,0);
 return(ret);
}


int Imvmover_addfile(mover,filename,flag)
struct ImvmoveR *mover;
char *filename;
int flag;
/*
 bit2= for Imvfiles_add() : do not notify_list_change
*/
{
 int ret;

 ret= Imvfiles_add(mover->files,filename,1|2|(flag&4)); 
                   /* beware in case we ever add change of working directory */
 if(ret<=0) {
   Imvmover_stderr(mover,"creation of file item failed : ",0);
   Imvmover_stderr(mover,filename,1);
   goto ex;
 } else if(ret==2) {
   Imvmover_stderr(mover,"illegal filename given with -addfile* :",1);
   Imvmover_stderr(mover,filename,1);
   ret= 0; goto ex;
 } else if(ret==3) {
   Imvmover_stderr(mover,"file given with -addfile* is not a data file :",1);
   Imvmover_stderr(mover,filename,1);
   ret= 0; goto ex;
 }
 ret= 1;
ex:
 return(ret);
}


int Imvmover_cmd_visit(mover,filename,flag)
struct ImvmoveR *mover;
char *filename;
int flag;
{
 int ret,added= 0,old_idx,idx;
 struct AsagenT *agent;

 Imvfiles_get_idx(mover->files,&old_idx,0);
 ret= Asagent_get_permission(mover->agent,"visit",filename,4);
 if(ret<=0)
   goto ex;
 Imvmover_noop(mover,0);
 ret= Imvmover_addfile(mover,filename,4);
 if(ret<=0)
   goto ex;
 added= 1;
 Imvfiles_get_idx(mover->files,&idx,0);

 /* make the usual refresh show */
 agent= mover->agent;
 Asagent_stdout(agent,"",1);
 Asagent_stdout_bar(agent,0);
 Asagent_print_fileinfo(agent,filename,0);
 ret= Asagent_show_doubles(agent,filename,1);
 if(ret<=0)
   goto ex;
 ret= Asagent_show_current(agent,filename,idx,0);
 if(ret<=0)
   goto ex;

 mover->agent->keep_views= 1;
 ret= 1;
ex:
 if(added) {
   Imvfiles_get_idx(mover->files,&idx,0);
   if(idx!=old_idx && idx==Imvfiles_return_count(mover->files,0)-1)
     Imvfiles_delete_last(mover->files,0);
 }
 Imvfiles_set_idx(mover->files,old_idx,0);
 return(ret);
}


int Imvmover_cmd_addfile_list(mover,filename,flag)
struct ImvmoveR *mover;
char *filename;
int flag;
{
 int ret,success= 0,failed= 0;
 FILE *fp= NULL;
 char line[2*SfileadrL];

 ret= Asagent_get_permission(mover->agent,"filelist","addfile_list",0);
 if(ret<=0)
   goto ex;
 Imvmover_noop(mover,0);
 fp= Sfile_fopen(filename,"rb");
 if(fp==NULL){
   sprintf(line,"-addfile_list : failed to open %s",filename);
   Imvmover_stderr(mover,line,0);
   return(0);
 }
 while(Sfile_fgets(line,SfileadrL-1,fp)!=NULL){
   ret= Imvmover_addfile(mover,line,0);
   if(line[0]==0)
 continue;
   if(ret<0)
     goto ex;
   else if(ret>0)
     success++;  
   else
     failed++;
 }
 ret= (!!success)||(!failed);
ex:;
 if(fp!=NULL)
   fclose(fp);
 return(ret);
}


int Imvmover_interpreter(mover,flag)
struct ImvmoveR *mover;
int flag;
/*
 bit0= quiet : do not report command
 bit1= do not complain loudely about unknown command
 reserved:  bit2= client mode
 reserved:  bit3= forced local mode
*/
/*
 return values: 
  <=0 = error
  1 = all went normal
  2 = redo user interaction
  4= unknown command
*/
{
 char *cmd= NULL,line[SfileadrL],*adr;
 int ret,idx,no_files= 0,yes= 0,count,i;

 static unsigned int hop_delay= 0;

 mover->recursion_level++;
 if(mover->recursion_level>mover->recursion_max){
   Imvmover_stderr(mover,
             "Command recursion too deep, probably endless loop.",1);
   ret= 0;
   goto ex;
 }

 /* get index (and test wether valid at all) */
 ret= Imvmover_determine_current(mover,&idx,&count,0);
 if(ret<=0)
   goto ex;
 else if(ret==2)
   no_files= 1;
 ret= Asevent_get_cmd(mover->event,&cmd,0);
 if(ret<=0)
   goto ex;
 if(strncmp(cmd,"--imvmover",10)==0) {
   /* remove own module prefix if there */
   for(i=10;cmd[i]!=0;i++)
     cmd[i-10]= cmd[i];
   cmd[i]= 0;
 }
 adr= mover->current_adr;
 Imvhistory_destroy(&(mover->history),0);
 Imvhistory_new(&(mover->history),idx,adr,adr,NULL,mover->event,0);
 if(ret<=0)
   goto ex;

 ret= 0;
 mover->fstep= 0;
 if(!(flag&1))
   Imvmover_stdout(mover,cmd,1);
 if(cmd[0]==0||cmd[0]=='#'||cmd[0]=='!') {
   Imvmover_noop(mover,0);
 } else if(strncmp(cmd,"-addfile",8)==0 && cmd[8]!=0 && cmd[8]!='_') {
   /* -addfile?address */
   ret= Asagent_get_permission(mover->agent,"filelist","addfile",0);
   if(ret<=0)
     goto ex;
   Imvmover_noop(mover,0);
   ret= Imvmover_addfile(mover,cmd+9,0);
   if(ret<=0)
     goto ex;
 } else if(strncmp(cmd,"-addfile_list",13)==0 && cmd[13]!=0) {
   ret= Imvmover_cmd_addfile_list(mover,cmd+14,0);
   if(ret<=0)
     goto ex;
 } else if(strcmp(cmd,"-back")==0){
   ret= Asagent_get_permission(mover->agent,"navigation","back",0);
   if(ret<=0)
     goto ex;
   Sfile_microsleep(hop_delay,0);
   Imvmover_noop(mover,0);
   mover->fstep= -1;
 } else if(strncmp(cmd,"-doubleto",9)==0 && cmd[9]!=0){     /* -doubleto?dir */
   ret= Asagent_get_permission(mover->agent,"move","doubleto",0);
   if(ret<=0)
     goto ex;
   if(no_files) goto empty_hands;
   ret= Imvmover_cmd_doubleto(mover,cmd+10,0);
   if(ret<=0)
     goto ex;
 } else if(strncmp(cmd,"-for",4)==0 && cmd[4]!=0){
   ret= Imvmover_cmd_for(mover,cmd+5,0);
   if(ret<=0)
     goto ex;
   Imvmover_noop(mover,0);
 } else if(strcmp(cmd,"-jpgb")==0){
   ret= Asagent_get_permission(mover->agent,"modify","jpgb",2);
   if(ret<=0)
     goto ex;
   if(no_files) goto empty_hands;
   ret= Imvmover_cmd_jpgb(mover,0);
   if(ret<=0)
     goto ex;
 } else if(strncmp(cmd,"-jump",5)==0 && cmd[5]!=0){
   ret= Imvmover_cmd_jump(mover,cmd+6,0);
   if(ret<0)
     goto ex;
   Imvmover_noop(mover,0);
 } else if(strncmp(cmd,"-list",5)==0 && cmd[5]!=0){
   ret= Imvmover_cmd_list(mover,cmd+6,0);
   if(ret<0)
     goto ex;
   Imvmover_noop(mover,0);
 } else if(strncmp(cmd,"-moveto",7)==0 && cmd[7]!=0){         /* -moveto?dir */
   ret= Asagent_get_permission(mover->agent,"move","moveto",0);
   if(ret<=0)
     goto ex;
   if(no_files) goto empty_hands;
   ret= Imvmover_cmd_moveto(mover,cmd+8,0);
   if(ret<=0)
     goto ex;
   if(ret==2)
     goto ex;
   mover->fstep= 1;
 } else if(strcmp(cmd,"-preserve")==0){
   if(no_files) goto empty_hands;
   ret= Imvmover_cmd_preserve(mover,0);
   if(ret<=0)
     goto ex;
 } else if(strcmp(cmd,"-replace_double")==0){
   if(no_files) goto empty_hands;
   ret= Imvmover_cmd_replace_double(mover,0);
   if(ret<=0)
     goto ex;
   if(ret==1)
     mover->fstep= 1;


#ifdef NOT_ACTIVE
 } else if(strcmp(cmd,"-rm")==0){
   if(no_files) goto empty_hands;
   ret= Imvmover_cmd_rm(mover,0);
   if(ret<=0)
     goto ex;
   mover->fstep= 1;
#endif /* NOT_ACTIVE */

 } else if(strcmp(cmd,"-skip")==0){
   ret= Asagent_get_permission(mover->agent,"navigation","skip",0);
   if(ret<=0)
     goto ex;
   Sfile_microsleep(hop_delay,0);
   Imvmover_noop(mover,0);
   mover->fstep= 1;
 } else if(strcmp(cmd,"-undo")==0) {
   ret= Imvmover_cmd_undo(mover,0);
   if(ret<=0)
     goto ex;
 } else if(strncmp(cmd,"-visit",6)==0 && cmd[6]!=0){
   ret= Imvmover_cmd_visit(mover,cmd+7,0);
   if(ret<=0)
     goto ex;
 } else if(strncmp(cmd,"-where",6)==0 && cmd[6]!=0){
   ret= Imvmover_cmd_where(mover,cmd+7,0);
   if(ret<=0)
     goto ex;
   Imvmover_noop(mover,0);
 } else {
   if(!(flag&2)){
     sprintf(line,"--- command not recognized : '%s'",cmd);
     Imvmover_stderr(mover,line,1);
     Imvmover_noop(mover,0);
   }
   ret= 4; goto ex;
 }
 if(! (Imvhistory_nothing_happened(mover->history,0) 
       || Imvhistory_no_undo_possible(mover->history,0) ) ) {
   ret= Imvfiles_add_history(mover->files,mover->history,0);
   if(ret<=0)
     goto ex;
   mover->history= NULL;
 }
 if(ret==0) ret= 1;
ex:;
 Imvhistory_destroy(&(mover->history),0);
 Sregex_string(&cmd,NULL,0);
 mover->recursion_level--;
 return(ret);
empty_hands:;
 sprintf(line,"--- no file object in list with command : '%s'",cmd);
 Imvmover_stderr(mover,line,1);
 ret= 2;
 goto ex;
}

