
/*
  cc -g -c sfile.c
*/

#include <sys/time.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <sys/wait.h>
#include <errno.h>
#include <utime.h>


#ifdef Sfile_use_smeM

#include "smem.h"

#else /* Sfile_use_smeM */

#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));
#define Smem_malloC malloc
#define Smem_freE free

#endif /* ! Sfile_use_smeM */


#include "sfile.h"


/* ------------------------------------------------------------------------- */


int Sfile_getcwd(handle,flag)
char **handle;
int flag;
/*
 bit0= append '/'
*/
{
 char dir[SfileadrL],*h;

 if(getcwd(dir,SfileadrL-1)==NULL)
   return(0);
 if((flag&1) && strlen(dir)>1)
   strcat(dir,"/");
 
 h= TSOB_FELD(char,strlen(dir)+1);
 if(h==NULL)
   return(-1);
 strcpy(h,dir);
 if((*handle)!=NULL) {
   Smem_freE(*handle); 
   *handle= NULL;
 }
 *handle= h;
 return(1);
} 


int Sfile_try_local_adr(adr,locadr,flag)
char *adr,**locadr;
int flag;
{
 char *dir= NULL;
 int ret;

 *locadr= adr;
 ret= Sfile_getcwd(&dir,1);
 if(ret<=0)
   goto ex;
 if(strncmp(adr,dir,strlen(dir))!=0)
   {ret= 0; goto ex;}
 *locadr= adr+strlen(dir);
 ret= 1;
ex:;
 if(dir!=NULL)
   {Smem_freE(dir);dir= NULL;}
 return(ret);
}


int Sfile_unique_adr(adr,flag)
char *adr;
int flag;
{
 int i,testcount= 0;
 char *cpt,*wpt,infix[16];
 struct stat stbuf;
 
 while(lstat(adr,&stbuf)!=-1) {
   if(testcount==0){
     /* preserve suffix */
     cpt= strrchr(adr,'.');
     if(cpt!=NULL) {
       for(wpt= adr+strlen(adr);wpt>=cpt;wpt--)
          *(wpt+8)= *wpt;
     } else {
       cpt= adr+strlen(adr);
       *(cpt+8)= 0;
     }
   }
   sprintf(infix,"_%-7.7ld",testcount);
   for(i=0;i<8;i++)
     *(cpt+i)= infix[i];
   testcount++;
 }
 return(1);
}


FILE *Sfile_fopen(filename,mode,flag)
char *filename;
char *mode;
int flag;
/*
 bit0= do not print error message on failure
*/
{
 FILE *fp= NULL;
 if(strcmp(filename,"-")==0) {
   if(mode[0]=='a') fp= stdout;
   else             fp= stdin;
 } else if(strncmp(filename,"tcp:",4)==0){
   fprintf(stderr,"sorry - TCP/IP service isn't implemented yet.\n");
 } else if(strncmp(filename,"file:",5)==0){
   fp= fopen(filename+5,mode);
 } else {
   fp= fopen(filename,mode);
 }
 if(fp==NULL){
   if(!(flag&1))
     fprintf(stderr,"failed to open %s in %s mode\n",filename,mode);
   return(NULL);
 }
 return(fp);
}


int Sfile_fclose(fp)
FILE *fp;
{
 if(fp!=stdout && fp!=stdin)
   fclose(fp); 
 return(1);
}


char *Sfile_fgets(line,maxl,fp)
char *line;
int maxl;
FILE *fp;
{
int l;
char *ret;
 ret= fgets(line,maxl,fp);
 if(ret==NULL)
   return(NULL);
 l= strlen(line);
 if(l>0) if(line[l-1]=='\r') line[--l]= 0;
 if(l>0) if(line[l-1]=='\n') line[--l]= 0;
 if(l>0) if(line[l-1]=='\r') line[--l]= 0;
 return(ret);
}


int Sfile_read(fd,buf,bufsize,count,flag)
int fd;
char *buf;
int bufsize;
int *count;
int flag;
{
 int ret= -1;

 *count= 0;
 ret= read(fd,buf,bufsize);
 if(ret==-1) {
   if(errno==EAGAIN)
     ret= 1;
   goto ex;
 } else if (ret==0) /* EOF */
   goto ex;
 *count= ret;
ex:;
 return(ret);
}


int Sfile_write(fd,buf,count,microsleep,flag)
int fd;
char *buf;
int count;
unsigned int microsleep;
int flag;
/*
 bit0= if would-block : return(2) rather than wait
*/
{
 int ret= -1;

 if(microsleep==0)
   microsleep= 100000;
try_it:;
 ret= write(fd,buf,count);
 if(ret<=0) {
   if(errno==EAGAIN) {
     if(flag&1) {
       ret= 2;
       goto ex;
     }
     Sfile_microsleep(microsleep,0);
     goto try_it;
   }
   goto ex;
 }
ex:;
 return(ret);
}


int Sfile_fd_is_ready(fd,microsec,flag)
int fd;
unsigned int microsec;
int flag;
/*
 bit0= check for reading
 bit1= check for writing
 bit2= check for exception
return-bits are set according to matching flag-bits
*/
{
 struct timeval wt;
 fd_set rds,wts,exs;
 int ready,ret;

 wt.tv_sec=  microsec/1000000;
 wt.tv_usec= microsec%1000000;
 FD_ZERO(&rds);
 FD_ZERO(&wts);
 FD_ZERO(&exs);
 if(flag&1)
   FD_SET(fd,&rds);
 if(flag&2)
   FD_SET(fd,&wts);
 if(flag&4)
   FD_SET(fd,&exs);
 ready= select(fd+1,&rds,&wts,&exs,&wt);
 if(ready<=0)
   return(0);
 ret= 0;
 if(flag&1)
   if(FD_ISSET(fd,&rds))
     ret|=1;
 if(flag&2)
   if(FD_ISSET(fd,&wts))
     ret|=2;
 if(flag&4)
   if(FD_ISSET(fd,&exs))
     ret|=4;
 return(ret);
}


int Sfile_identical(file1,file2,flag)
char *file1,*file2;
int flag;
/*
 bit0= check wether it is the same  i-node 
*/
{
 FILE *fp1,*fp2;
 struct stat stbuf;
 int ret,c1,c2;
 ino_t ino1,ino2;

 ret= 0;

 if(flag&1){
   if(stat(file1,&stbuf)==-1) return(0);
   ino1= stbuf.st_ino;
   if(stat(file2,&stbuf)==-1) return(0);
   ino2= stbuf.st_ino;
   return(ino1==ino2);
 }

 fp1= Sfile_fopen(file1,"rb",0);
 fp2= Sfile_fopen(file2,"rb",0);
 if(fp1==NULL||fp2==NULL) goto fertig;
 while(1){
   c1= fgetc(fp1);
   c2= fgetc(fp2);
   if(c1==EOF && c2==EOF){
     ret= 1;
 break;
   }
   if(c1!=c2)
 break;
 }
fertig:;
 if(fp1!=NULL) fclose(fp1);
 if(fp2!=NULL) fclose(fp2);
 return(ret);
}


off_t Sfile_size(filename,flag)
char *filename;
int flag;
/*
 bit0= return -1 if file is missing 
*/
{
 struct stat stbuf;
 if(stat(filename,&stbuf)==-1) {
   if(flag&1) return(-1);
   else       return(0);
 }
 return(stbuf.st_size);
}


int Sfile_line_count(filename,flag)
char *filename;
int flag;
/*
 bit0-3= mode : 0=count all , 1=count nonempty lines , 2=count empty lines
*/
{
 int count= 0,mode;
 FILE *fp;
 char line[SfileadrL];

 mode= flag&15;
 fp= fopen(filename,"r");
 if(fp==NULL)
   return(-1);
 while(Sfile_fgets(line,sizeof(line),fp)!=NULL) {
   switch (mode) {
   case 1:
     if(line[0]==0)
 continue;
   break; case 2:
     if(line[0]!=0)
 continue;
   }
   count++;
 }
 fclose(fp);
 return(count);
}


int Sfile_type(filename,flag)
char *filename;
int flag;
/*
 bit0= return -1 if file is missing 
 bit1= return a hardlink with siblings as type 5
*/
/*
 return:  0=unknown  1=regular  2=directory  3=symbolic link  4=named pipe
          5=multiple hardlink (with bit1)
*/
{
 struct stat stbuf;
 if(lstat(filename,&stbuf)==-1) {
   if(flag&1) return(-1);
   else       return(0);
 }
 if(S_ISREG(stbuf.st_mode)) {
   if(flag&2)
     if(stbuf.st_nlink>1)
       return(5);
   return(1);
 }
 if(S_ISDIR(stbuf.st_mode))
   return(2);
 if(S_ISLNK(stbuf.st_mode))
   return(3);
 if(S_ISFIFO(stbuf.st_mode))
   return(4);
 return(0);
}


int Sfile_readlink(filename,buff,buff_size,flag)
char *filename;
char *buff;
int buff_size;
int flag;
/*
 bit0= return -1 if file is missing 
*/
{
 int l;

 if(Sfile_type(filename,0)!=3)
   return(0);
 l= readlink(filename,buff,buff_size-1);
 if(l<=0)
   return(0);
 buff[l]= 0;
 return(1);
}


int Sfile_make_fifo(filename,flag)
char *filename;
int flag;
/*
 bit0= allow read for group
 bit1= allow write for group
 bit2= allow read for others
 bit3= allow write for others
 bit8= remove existing file
*/
{
 int ret;
 mode_t permission;

 if(flag&256) {
   char *newname;
   newname= Smem_malloC(strlen(filename)+8);
   if(newname==NULL)
     return(-1);
   sprintf(newname,"%s.nofifo",filename);
   ret= rename(filename,newname);
   Smem_freE((char *) newname);
   if(ret==-1)
     return(-1);
 }
 permission= S_IRUSR|S_IWUSR;
 if(flag&1)
   permission|= S_IRGRP;
 if(flag&2)
   permission|= S_IWGRP;
 if(flag&4)
   permission|= S_IROTH;
 if(flag&8)
   permission|= S_IWOTH;

 ret= mknod(filename,S_IFIFO,0);
 if(ret==-1) 
   return(-1);
 ret= chmod(filename,S_IRUSR|S_IWUSR);
 if(ret==-1) 
   return(0);
 return(1);
}


int Sfile_bin_to_file(filename,data,size,flag)
char *filename;
char *data;
int size;
int flag;
/*
 bit0= permission to overwrite existing file
 bit1= append to file
 bit2= add newline character
*/
{
 FILE *fp;
 struct stat stbuf;
 int ret;

 if(!(flag&3))
   if(stat(filename,&stbuf)!=-1)
     return(0);
 if(flag&2)
   fp= Sfile_fopen(filename,"ab",0);
 else
   fp= Sfile_fopen(filename,"wb",0);
 if(fp==NULL)
   return(-1);
 if(size>0)
   ret= fwrite(data,size,1,fp);
 else
   ret= 1;
 if(ret==1 && (flag&4))
   ret= (fputc('\n',fp)!=EOF);
 Sfile_fclose(fp);
 return(ret);
}


int Sfile_string_to_file(filename,text,flag)
char *filename;
char *text;
int flag;
/*
 bit0= permission to overwrite existing file
 bit1= append to file
 bit2= add newline character
*/
{
 int ret;

 ret= Sfile_bin_to_file(filename,text,strlen(text),flag&7);
 return(ret);
}


int Sfile_file_to_text(filename,text,textsize,flag)
char *filename;
char *text;
int textsize;
int flag;
/*
 bit0= return(2) if file is empty (0 else)
 bit1= Sfile_fopen:bit0 : no error messages 
*/
{
 FILE *fp;
 int ret;

 if(textsize<=0)
   return(-1);
 fp= Sfile_fopen(filename,"rb",(flag&2)>>1);
 if(fp==NULL)
   return(0);
 if(Sfile_fgets(text,textsize,fp)==NULL)
   ret= 2*(flag&1);
 else
   ret= 1;
ex:;
 Sfile_fclose(fp);
 return(ret);
}


int Sfile_file_to_double(filename,num,flag)
/*
 bit0= return(2) if file is empty (0 else)
 bit1= Sfile_fopen:bit0 : no error messages 
*/
char *filename;
double *num;
int flag;
{
 char text[SfileadrL];
 double dv;
 int ret;

 if(Sfile_file_to_text(filename,text,sizeof(text),flag&2)<=0)
   return(0);
 ret= sscanf(text,"%lf",&dv);
 if(ret<=0)
   return(0);
 *num= dv;
 return(1);
}


int Sfile_file_to_int(filename,num,flag)
char *filename;
int *num;
int flag;
/*
 bit1= Sfile_fopen:bit0 : no error messages 
*/
{
 double dv;
 int ret;

 if(Sfile_file_to_double(filename,&dv,flag&2)<=0)
   return(0);
 if(dv<-0x7FFFFFFF || dv>=0x7FFFFFFF)
   return(0);
 *num= dv;
 return(1);
}


int Sfile_file_to_bin(filename,data,datasize,flag)
char *filename;
char *data;
int datasize;
int flag;
/*
 bit0= return(2) if file does not deliver enough bytes (0 else)
 bit1= Sfile_fopen:bit0 : no error messages 
*/
{
 FILE *fp;
 int ret;

 if(datasize<=0)
   return(-1);
 fp= Sfile_fopen(filename,"rb",(flag&2)>>1);
 if(fp==NULL)
   return(0);
 ret= fread(data,1,datasize,fp);
 if(ret<datasize){
   if(flag&1)
     return(2);
   return(0);
 }
ex:;
 Sfile_fclose(fp);
 return(1);
}


int Sfile_get_filetimes(fileadr,atime,mtime,flag)
char *fileadr;
time_t *atime;
time_t *mtime;
int flag;
{
 struct stat stbuf;

 if(lstat(fileadr,&stbuf)==-1)
   return(flag&1);
 if(!S_ISREG(stbuf.st_mode))
   return(0);
 *atime= stbuf.st_atime;
 *mtime= stbuf.st_mtime;
 return(1);
}


int Sfile_set_filetimes(fileadr,mtime,atime,flag)
char *fileadr;
time_t mtime;
int flag;
{
 struct stat stbuf;
 struct utimbuf buf;
 int ret;

 if(lstat(fileadr,&stbuf)==-1)
   return(flag&1);
 if(!S_ISREG(stbuf.st_mode))
   return(0);
 buf.actime= atime;
 buf.modtime= mtime;
 ret= utime(fileadr,&buf);
 if(ret<=0)
   return(0);
 return(1);
}


int Sfile_older_than(fileadr,agelimit,flag)
char *fileadr;
time_t agelimit;
int flag;
/*
 bit0= return 1 if file age cannot be determined
*/
{
 struct stat stbuf;

 if(lstat(fileadr,&stbuf)==-1)
   return(flag&1);
 if(!S_ISREG(stbuf.st_mode))
   return(flag&1);
 if(stbuf.st_mtime<agelimit)
   return(1);
 return(0);
}


int Sfile_remove_old_files(filec,filev,agelimit,flag)
int filec;
char **filev;
time_t agelimit;
int flag;
/*
 bit0=  verbous
 bit15= file Sfile_autoremove_olD needed in parent directory rather than local
*/
{
 char permfile[SfileadrL];
 int ret,i,removed= 0,l,max_fpl;
 struct stat stbuf;

 if(flag&(1<<15)) {
   sprintf(permfile,"../%s",Sfile_autoremove_olD);
 } else {
   strcpy(permfile,Sfile_autoremove_olD);
 }
 if(lstat(permfile,&stbuf)==-1)
   return(0);
 l= 0;
 for(i=0;i<filec;i++)
   if(l<strlen(filev[i])) l= strlen(filev[i]);
 max_fpl= 79/(l+1);
 if(max_fpl<=0) max_fpl= 1;
 for(i=0;i<filec;i++){
   if(filev[i][0]=='/' || filev[i]==0) /* only local file addresses allowed */
 continue;
   if(Sfile_older_than(filev[i],agelimit,0)){
     unlink(filev[i]);
     removed++;
     if(flag&1){
       if(removed==1) 
         printf("removed (because too old) :\n");
       printf(" %s",filev[i]);
       if((removed % max_fpl)==0)
         printf("\n");
     }
   }
 }
 if((removed % max_fpl))
   printf("\n");
 return(1);
}


int Sfile_read_file_line(fileadr,search,search_l,line,line_l,errn,errmsg,flag)
char *fileadr,*search;
int search_l;
char *line;
int line_l;
int *errn;
char *errmsg;
int flag;
/*
 bit1= Sfile_fopen:bit0 : no error messages 
*/
{
 int ret,i,l;
 FILE *fp;
 
 if(search_l<=0)
   search_l= strlen(search);
 fp= Sfile_fopen(fileadr,"rb",(flag>>1)&1);
 if(fp==NULL) {
   *errn= errno;
   sprintf(errmsg,"Error %ld on opening input file : %s\n%s",
                  errno,fileadr,strerror(errno));
   return(-1);
 }
 while(Sfile_fgets(line,line_l,fp)!=NULL)
   if(strncmp(line,search,search_l)==0)
     break;
 if(strncmp(line,search,search_l)!=0) {
   line[0]= 0;
   ret= 0;
 } else {
   l= strlen(line);
   for(i= search_l;i<=l;i++)
     line[i-search_l]= line[i];
   ret= 1;
 }
ex:;
 if(fp!=NULL) fclose(fp);
 return(ret);
}


int Sfile_change_file_lines(fileadr,search,search_l,subst_line,
                            errn,errmsg,flag)
char *fileadr;
char *search;
int search_l;
char *subst_line;
int *errn;
char *errmsg;
int flag;
/*
 bit0-7= mode : 
         0=look for content of 'search' at start of text lines
*/
{
 int ret;
 char newadr[SfileadrL],line[SfileadrL];
 FILE *fp= NULL,*fw= NULL;
 struct utimbuf old_times;
 struct stat stbuf;

 if(search_l<=0)
   search_l= strlen(search);
 sprintf(newadr,"%s.new",fileadr);
 if(stat(fileadr,&stbuf)==-1){
   old_times.actime= old_times.modtime= time(0);
 }else{
   old_times.actime= stbuf.st_atime;
   old_times.modtime= stbuf.st_mtime;
 }
 fp= fopen(fileadr,"rb");
 if(fp==NULL) {
   *errn= errno;
   sprintf(errmsg,"Error %ld on opening input file : %s\n%s",
                  errno,fileadr,strerror(errno));
   return(0);
 }
 fw= fopen(newadr,"wb");
 if(fw==NULL) {
   *errn= errno;
   sprintf(errmsg,"Error %ld on opening output file : %s\n%s",
                  errno,newadr,strerror(errno));
   ret= 0;
   goto ex;
 }
 while(fgets(line,sizeof(line),fp)!=NULL){
   if(strncmp(line,search,search_l)==0)
     fprintf(fw,"%s",subst_line);
   else
     fprintf(fw,"%s",line);
 }
 ret= 1;
ex:;
 if(fp!=NULL) fclose(fp);
 if(fw!=NULL) fclose(fw);
 if(ret>0) {
   unlink(fileadr);
   rename(newadr,fileadr);
   utime(fileadr,&old_times);
 }else{
   unlink(newadr);
 }
 return(ret);
}


int Sfile_del_file_lines(fileadr,start_mark,stml,end_mark,enml,
                         errn,errmsg,flag)
char *fileadr;
char *start_mark;
int stml;
char *end_mark;
int enml;
int *errn;
char *errmsg;
int flag;
/*
 bit0= delete inclusive start_mark-line (else begin deleting after start_mark)
 bit1= delete inclusive end_mark-line (else stop deleting before end_mark)
*/
{
 int ret,del_mode= 0;
 char newadr[SfileadrL],line[SfileadrL];
 FILE *fp= NULL,*fw= NULL;
 struct utimbuf old_times;
 struct stat stbuf;

 if(stml<=0)
   stml= strlen(start_mark);
 if(enml<=0)
   enml= strlen(end_mark);
 sprintf(newadr,"%s.new",fileadr);
 if(stat(fileadr,&stbuf)==-1){
   old_times.actime= old_times.modtime= time(0);
 }else{
   old_times.actime= stbuf.st_atime;
   old_times.modtime= stbuf.st_mtime;
 }
 fp= fopen(fileadr,"rb");
 if(fp==NULL) {
   *errn= errno;
   sprintf(errmsg,"Error %ld on opening input file : %s\n%s",
                  errno,fileadr,strerror(errno));
   return(0);
 }
 fw= fopen(newadr,"wb");
 if(fw==NULL) {
   *errn= errno;
   sprintf(errmsg,"Error %ld on opening output file : %s\n%s",
                  errno,newadr,strerror(errno));
   ret= 0;
   goto ex;
 }
 while(fgets(line,sizeof(line),fp)!=NULL){
   if(strncmp(line,start_mark,stml)==0 && !del_mode){
     del_mode= 1;
     if(!(flag&1))
       fprintf(fw,"%s",line);
   } else if(strncmp(line,end_mark,enml)==0 && del_mode){
     del_mode= 0;
     if(!(flag&2))
       fprintf(fw,"%s",line);
   } else if(! del_mode) {
     fprintf(fw,"%s",line);
   }
 }
 ret= 1;
ex:;
 if(fp!=NULL) fclose(fp);
 if(fw!=NULL) fclose(fw);
 if(ret>0) {
   unlink(fileadr);
   rename(newadr,fileadr);
   utime(fileadr,&old_times);
 }else{
   unlink(newadr);
 }
 return(ret);
}


int Sfile_copy_file(source,target,errn,errmsg,flag)
char *source;
char *target;
int *errn;
char errmsg[SfileadrL];
int flag;
{
 FILE *fp,*fw;
 char *line[1024];
 int ret;

 if(Sfile_identical(source,target,1)>0)
   return(2);
 fp= fopen(source,"r");
 if(fp==NULL) {
   *errn= errno;
   sprintf(errmsg,"Error %ld on opening input file : %s\n%s",
                  errno,source,strerror(errno));
   return(0);
 }
 fw= fopen(target,"w");
 if(fw==NULL) {
   *errn= errno;
   sprintf(errmsg,"Error %ld on opening output file : %s\n%s",
                  errno,target,strerror(errno));
   ret= 0;
   goto ex;
 }
 while((ret= fread(line,1,sizeof(line),fp))>0){
   fwrite(line,1,ret,fw);
   if(ret<sizeof(line))
 break;
 }
 ret= 1;
ex:;
 if(fp!=NULL) fclose(fp);
 if(fw!=NULL) fclose(fw);
 return(ret);
}


int Sfile_move_file(source,target,errn,errmsg,flag)
char *source;
char *target;
int *errn;
char *errmsg;
int flag;
/*
 bit0= do not unlink target if source cannot be unlinked
 bit7= use overwrite+delete rather than rename
 bit8= like bit7 but do not unlink source (i.e. copy >>> with properties)
*/
{
 int ret,ovw_del=0;
 struct stat stbuf;

 if(flag&(128|256))
   ovw_del= 1;
 if(!ovw_del)
   if(rename(source,target)!=-1)
     return(1);
 if(errno==EXDEV || ovw_del) {
   /* try copy and unlink */;
   ret= Sfile_copy_file(source,target,errn,errmsg,0);
   if(ret<=0)
     return(ret);
   if(!(flag&128)) {

     /* >>> transfer as many properties as possible */;

   }
   if(flag&256)
     return(3);
   if(unlink(source)==-1) {
     *errn= errno;
     sprintf(errmsg,
             "Error %ld on removing file %s\nafter copying to %s\n\"%s\"",
             errno,source,target,strerror(errno));
     if(lstat(source,&stbuf)!=-1 && !(flag&1)) {
       strcat(errmsg,"\n(Operation failed. Target file has been removed now)");
       unlink(target);
     }
     return(0);
   }
   return(2);
 }
 *errn= errno;
 sprintf(errmsg,"Error %ld on renaming\n   %s\nto %s\n\"%s\"",
                errno,source,target,strerror(errno));
 return(0);
}


int Sfile_mk_hashdir(dir,hashsize,flag)
char *dir;
int hashsize;
int flag;
/*
 bit0= directory may be already existent
*/
{
 int ret,i;
 char adr[SfileadrL],text[80];
 struct stat stbuf;

 if(hashsize<=0 || hashsize>255)
   return(-1);
 if(stat(dir,&stbuf)!=-1) {
   if(!(flag&1))
     return(0);
 } else {
   ret= mkdir(dir,0x1FF);
   if(ret==-1)
     return(0);
 }
 sprintf(adr,"%s/%s",dir,"sfile_hashed_directory");
 sprintf(text,"%ld",hashsize);
 ret= Sfile_string_to_file(adr,text,1);
 if(ret<=0)
   return(0);
 for(i=0;i<hashsize;i++) {
   sprintf(adr,"%s/%-2.2X",dir,i);
   if(stat(adr,&stbuf)==-1) {
     ret= mkdir(adr,0x1FF);
     if(ret==-1)
       return(0);
   }
 }
 return(1);
}


int Sfile_is_hashdir(dir,flag)
char *dir;
int flag;
{
 char adr[SfileadrL];
 int hashsize,ret;

 sprintf(adr,"%s/%s",dir,"sfile_hashed_directory");
 ret= Sfile_file_to_int(adr,&hashsize,2);
 if(ret<=0 || hashsize<0)
   hashsize= 0;
 return(hashsize);
}


int Sfile_get_hashadr(dir,name,adr,flag)
char *dir,*name,*adr;
int flag;
/*
 bit0= case insensitive
 bit1= create hash directory structure (hashsize=251)
*/
{
 char hname[SfileadrL],hdir[SfileadrL];
 int i,hashsize,ret,hashcode; 

 hashsize= Sfile_is_hashdir(dir,0);
 if(hashsize<=0) {
   if(!(flag&2))
     return(hashsize);
   hashsize= 251;
   ret= Sfile_mk_hashdir(dir,hashsize,1);
   if(ret<=0)
     return(ret);
 }
 if(hashsize>255)
   hashsize= 255;

 for(i=0;name[i]!=0;i++) {
   if((flag&1) && (isupper(name[i])||islower(name[i])))
     hname[i]= toupper(name[i]);
   else
     hname[i]= name[i];
 }
 hname[i]= 0;
 hashcode= Sfile_hashcode(hname,hashsize,1);
 sprintf(hdir,"%s/%-2.2X",dir,hashcode);
 ret= Sfile_get_hashadr(hdir,hname,adr,flag&~(1|2));
 if(ret<0)
   return(ret);
 if(ret>0)       /* hdir was a hashed directory itself */
   return(ret);
 sprintf(adr,"%s/%s",hdir,hname);
 return(1);
}


int Sfile_string_to_hashdir(dirlist,name,text,flag)
char *dirlist,*name,*text;
int flag;
/*
 bit0= permission to overwrite existing file
 bit1= ignore -return_false
 bit2= only interpret -options , simulate failed creation of name
 bit3= delete file rather than create
 bit4= with bit2 , simulate successful creation of name
*/
{
 char **argv= NULL,adr[SfileadrL],*npt;
 int ret=1,argc= 0,i,case_insensitive= 1,expl_ret= -1;

 npt= strrchr(name,'/');
 if(npt==NULL)
   npt= name;
 else
   npt++;
 ret= Sfile_make_argv("",dirlist,&argc,&argv,1);
 if(ret<=0 || argc<=0)
   goto ex;
 for(i=0;i<argc;i++) {
   if(strcmp(argv[i],"-case")==0) {
     case_insensitive= 0;
   } else if(strcmp(argv[i],"-nocase")==0) {
     case_insensitive= 1;
   } else if(strcmp(argv[i],"-return_true")==0) {
     expl_ret= 1;
   } else if(strcmp(argv[i],"-return_false")==0) {
     if(!(flag&2))
       expl_ret= 0;
   } else {
     if(flag&4) {
       if(flag&16)
         ret= 1;
       else
         ret= 0;
 continue;
     } else {
       ret= Sfile_get_hashadr(argv[i],npt,adr,case_insensitive|2);
       if(ret<=0)
 continue;
     }
     if(flag&8){
       ret= !unlink(adr);

       /* <<< only for debug purposes */
       fprintf(stderr,"Sfile_string_to_hashdir() removed '%s'\n",adr);

     } else {
       ret= Sfile_string_to_file(adr,text,flag&1);
     }
   }
 }
ex:;
 Sfile_make_argv("",dirlist,&argc,&argv,2); /* release */
 if(expl_ret>=0) 
   ret= expl_ret;
 return(ret);
}


int Sfile_search_file(dirlist,name,flag)
char *name;
char *dirlist;
int flag;
{
 char *cpt,*npt;
 char dir[SfileadrL],adr[SfileadrL],**filev=NULL,add[5];
 int ret,i,l,case_insensitive=1,found= 0,filec= 0;
 struct stat stbuf;
 
 cpt= npt= dirlist;

 while((*cpt)!=0) {
   npt= strchr(cpt+1,' ');
   if(npt==NULL)
     npt= cpt+strlen(cpt);
   l= npt-cpt;
   strncpy(dir,cpt,l);
   dir[l]= 0;
   cpt= npt;
   if(*cpt!=0) cpt++;
   if(strcmp(dir,"-nocase")==0){
     case_insensitive= 1;
 continue;
   } else if(strcmp(dir,"-case")==0){
     case_insensitive= 0;
 continue;
   }

   ret= Sfile_get_hashadr(dir,name,adr,case_insensitive);
   if(ret>0) {
     if(stat(adr,&stbuf)!=-1) {
       strcpy(name,adr);
       return(1);
     }
   } else if(case_insensitive) {
     strcpy(adr,dir);
     strcat(adr,"/");
     for(i=0;name[i]!=0;i++) {
       if(isupper(name[i])||islower(name[i])) {
         add[0]='[';
         add[1]= tolower(name[i]);
         add[2]= toupper(name[i]);
         add[3]= ']';
         add[4]= 0;
       } else {
         add[0]= name[i];
         add[1]= 0;
       }
       strcat(adr,add);
     }
     Sfile_ls(adr,&filec,&filev,1|4);
     if(filec>0) {
       strcpy(name,filev[0]);
       found= 1;
     }
     Sfile_ls(adr,&filec,&filev,2); /* release data */
     if(found>0) 
       return(1);
   }else{
     sprintf(adr,"%s/%s",dir,name);
     if(stat(adr,&stbuf)!=-1) {
       strcpy(name,adr);
       return(1);
     }
   }
 }
 return(0);
}


/* ------------------------------------------------------------------------- */


int Sfile_esc_filename(name,namsiz,flag)
unsigned char *name;
int namsiz;
int flag;
/*
 bit0= escape bad characters as +XX rather than XX and eventually prepend '_' 
 bit1= escape non-alphanum as '_'
 bit2= escape '/'
 bit3= escape blank as '_'
 bit4= do not precede address with '_' at bit0 if '+' has been used
 bit5= do not escape leading '-' and leading '.'
 bit6= replace non-printable-7-bit-ASCII by '_'
*/
{
 unsigned char *buff,*wpt,*cpt,infix[80];
 static unsigned char badchar[]= Sfile_bad_name_charS;
 int i,k,bad,l,ret= 1,testcount;
 struct stat stbuf;

 if(flag&64) goto narrow_ascii;
 if(flag&2) goto alnum;

 if(namsiz<=0)
   return(-1);
 buff= TSOB_FELD(unsigned char,namsiz+10);
 if(buff==NULL)
   return(-1);

 wpt= buff;
 l= strlen(name);
 for(i=0;i<l;i++){
   bad= 0;
   if(name[i]<=Sfile_bad_name_loW || name[i]>=Sfile_bad_name_higH){
     bad= 1;
   } else if((!(flag&32)) && (name[i]=='-' || name[i]=='.')) {
     if(i==0)
       bad= 1;
     else if(name[i-1]=='/' && !(flag&4))
       bad= 1;
   } else if((flag&1) && name[i]=='+') {
     bad= 1;
   } else if((flag&4) && name[i]=='/') {
     bad= 1;
   } else if(!(flag&64)){
     for(k=0;badchar[k]!=0;k++)
       if(badchar[k]==name[i])
         break;
     if(badchar[k]!=0)
       bad= 1;
   }
   if((flag&8) && name[i]==' ') {
     *(wpt++)= '_';
   } else if(bad) {
     if(flag&1)
       *(wpt++)= '+';
     sprintf(wpt,"%-2.2X",(unsigned int) name[i]);
     wpt+= strlen(wpt);
   } else {
     *(wpt++)= name[i];
   }
   if(wpt-buff>=namsiz){
     wpt== buff+namsiz-1;
     ret= 0;
 break;
   }
 }
 *wpt= 0;
 name[0]= 0;
 if((flag&1) && !(flag&16) && strchr(buff,'+')!=NULL)
   strcat(name,"_");
 strcat(name,buff);
 Smem_freE(buff);
 return(ret);

alnum:;
 l= strlen(name);
 for(i=0;i<l;i++){
   if(name[i]<'0' || (name[i]>'9' && name[i]<'A') ||
                     (name[i]>'Z' && name[i]<'a') || name[i]>'z')
     name[i]= '_';
 }
 return(1);

narrow_ascii:;
 l= strlen(name);
 for(i=0;i<l;i++){
   if(name[i]==' ')
 continue;
   if(name[i]<=Sfile_bad_name_loW || name[i]>=Sfile_bad_name_higH)
     name[i]= '_';
 }
 return(1);
}


int Sfile_make_delimiter(filename,delimiter,delimiter_size,flag)
char *filename;
char *delimiter;
int delimiter_size;
int flag;
/*
 bit0= start with empty delimiter
*/
{
 int dcount;

 dcount= strlen(delimiter);
 if(dcount==0 || (flag&1)) {
   strcpy(delimiter,"+");
   dcount= 1;
 }
 while(1) {
   if(strstr(filename,delimiter)==NULL) 
 break;
   dcount++;
   if(dcount>=delimiter_size-1)
     return(0);
   strcat(delimiter,"+");
 }
 return(1);
}


int Sfile_split_adr(adr,dir,name,flag)
char *adr;
char dir[SfileadrL];
char name[SfileadrL];
int flag;
/*
 bit0= set dir="." if empty
*/
{
 char *cpt;

 cpt= strrchr(adr,'/');
 if(cpt!=NULL) {
   strncpy(dir,adr,cpt-adr);
   dir[cpt-adr]= 0;
   strcpy(name,cpt+1);
   return(2);
 }
 if(flag&1)
   strcpy(dir,".");
 else
   dir[0]= 0;
 strcpy(name,adr);
 return(1);
}


int Sfile_is_local_adr(adr,flag)
char *adr;
int flag;
{
 if(adr[0]=='/')
   return(0);
 /* ??? recognize URLs ??? */
 if(adr[0]=='.')
   if(adr[1]=='/')
     return(2);
 return(1);
}


int Sfile_make_adr_global(adr,flag)
char *adr;
int flag;
{
 int ret;
 char *lpt,buffer[SfileadrL],dir[SfileadrL];

 ret= Sfile_is_local_adr(adr,0);
 if(ret==0)
   return(2);
 if(getcwd(dir,SfileadrL-strlen(adr)-1)==NULL) 
   return(0);
 lpt= adr;
 if(ret==2)
   lpt+= 2;
 sprintf(buffer,"%s/%s",dir,lpt);
 strcpy(adr,buffer);
 return(1);
}


int Sfile_addto_filename(adr,addition,flag)
char *adr;
char *addition;
int flag;
{
 char *cpt,*spt,*wpt;
 int l,i;

 l= strlen(addition);
 cpt= strrchr(adr,'.');
 spt= strrchr(adr,'/');
 if(cpt==NULL)
   cpt= adr+strlen(adr);
 else if(spt!=NULL && cpt<spt)
   cpt= adr+strlen(adr);
 wpt= cpt;
 for(wpt= cpt+strlen(cpt);wpt>=cpt;wpt--)
   *(wpt+l)= *wpt;
 for(i=0;i<l;i++)
   cpt[i]= addition[i];
 return(1);
}


int Sfile_hexval(x,flag)
unsigned char x;
int flag;
{
 int ret;

 ret= ( x<='9' ? x-'0' : x-'A' );
 if(ret<0||ret>15)
   ret= 0;
 return(ret);
}


int Sfile_unescape_filename(name,flag)
unsigned char *name;
int flag;
{
 int i,l,w=0,e;

 l= strlen(name);
 for(i=0;i<=l;i++){
   if(name[i]=='+'&&i<l-2){
     e= Sfile_hexval(name[i+1],0)*16+Sfile_hexval(name[i+2],0);
     name[w++]= e;
     i+= 2;
   } else 
     name[w++]= name[i];
 }
 return(1);
}


int Sfile_hashcode(name,hashsize,flag)
char *name;
int hashsize;
int flag;
/*
 bit0-7= algorithm
     0= flat sum
     1= rotating shift (0 to 22)
*/
{
 int sum=0,l,i,mode;

 l= strlen(name);
 mode= flag&255;
 if(mode==1){
   for(i=0;i<l;i++)
     sum= (sum+(((unsigned) name[i]))<<(i%23))%hashsize;
 }else{
   for(i=0;i<l;i++)
     sum= (sum+((unsigned) name[i]))%hashsize;
 }
 return(sum);
}


Sfile_append_crlf(line,flag)
char *line;
int flag;
{
 int l;
 l= strlen(line);
 if(l>0)
   if(line[l-1]=='\n') 
     line[--l]= 0;
 if(l>0)
   if(line[l-1]=='\r')
     line[--l]= 0;
 strcat(line,"\r\n");
 return(1);
}


/* ------------------------------------------------------------------------- */


int Sfile_get_lock(adr,objid,microsec,confirmation,flag)
char *adr;
/* void *objid; not a string but just a unique identifier within the process */
char *objid;
unsigned int microsec;
char *confirmation;
int flag;
{
 int ret;
 char ret_text[SfileadrL],line[SfileadrL];
 struct stat stbuf;

 if(lstat(adr,&stbuf)!=-1)
   return(0);
 if(gethostname(ret_text,sizeof(ret_text))==-1)
   strcpy(ret_text,"localhost");
 sprintf(line,"%s,%ld,%X",ret_text,getpid(),objid);
 ret= Sfile_string_to_file(adr,line,0);
 if(ret<=0)
   return(0);
 Sfile_microsleep(microsec,0);
 /* check wether string is still there (no race condition occured) */
 ret= Sfile_file_to_text(adr,ret_text,sizeof(ret_text),0);
 if(ret<=0)
   return(0);
 if(strcmp(ret_text,line)!=0)
   return(0);
 strcat(line," : ");
 strcat(line,confirmation);
 strcat(line,"\n");
 Sfile_string_to_file(adr,line,1);
 return(1);
}

/* ------------------------------------------------------------------------- */

int Sfile_make_argv(progname,line,argc,argv,flag)
char *progname;
char *line;
int *argc;
char ***argv;
int flag;
/*
 bit0= read progname as first argument from line
 bit1= just release argument list argv and return
*/
{
 int i,pass,maxl=0,l,argzaehl=0,bufl;
 char *cpt,*start;
 char buf[SfileadrL];

 Sfile_destroy_argv(argc,argv,0);
 if(flag&2) return(1);

 for(pass=0;pass<2;pass++) {
   cpt= line-1;
   if(!(flag&1)){
     argzaehl= 1;
     if(pass==0)
       maxl= strlen(progname);
     else
       strcpy((*argv)[0],progname);
   } else {
     argzaehl= 0;
     if(pass==0) maxl= 0;
   }
   while(*(++cpt)!=0){
   if(isspace(*cpt)) continue;
     start= cpt;
     buf[0]= 0;
     cpt--;
     while(*(++cpt)!=0) {
     if(isspace(*cpt)) break;
       if(*cpt=='"'){
         l= cpt-start; bufl= strlen(buf);
         if(l>0) {strncat(buf,start,l);buf[bufl+l]= 0;}
         l= strlen(buf);
         start= cpt+1;
         while(*(++cpt)!=0) if(*cpt=='"') break;
         l= cpt-start; bufl= strlen(buf);
         if(l>0) {strncat(buf,start,l);buf[bufl+l]= 0;}
         start= cpt+1;
       }else if(*cpt=='\''){
         l= cpt-start; bufl= strlen(buf);
         if(l>0) {strncat(buf,start,l);buf[bufl+l]= 0;}
         l= strlen(buf);
         start= cpt+1;
         while(*(++cpt)!=0) if(*cpt=='\'') break;
         l= cpt-start; bufl= strlen(buf);
         if(l>0) {strncat(buf,start,l);buf[bufl+l]= 0;}
         start= cpt+1;
       }
     if(*cpt==0) break;
     }
     l= cpt-start;
     bufl= strlen(buf);
     if(l>0) {strncat(buf,start,l);buf[bufl+l]= 0;}
     l= strlen(buf);
     if(pass==0){
       if(l>maxl) maxl= l;
     }else{
       strcpy((*argv)[argzaehl],buf);
     }
     argzaehl++;
   if(*cpt==0) break;
   }
   if(pass==0){
     *argc= argzaehl;
     *argv= (char **) Smem_malloC(argzaehl*sizeof(char *));
     for(i=0;i<*argc;i++)
       (*argv)[i]= (char *) Smem_malloC((maxl+1));
   }
 }
 return(1);
}


int Sfile_destroy_argv(argc,argv,flag)
int *argc;
char ***argv;
int flag;
{
 int i;

 if(*argc>0 && *argv!=NULL){
   for(i=0;i<*argc;i++){
     if((*argv)[i]!=NULL) 
       Smem_freE((*argv)[i]);
   }
   Smem_freE((char *) *argv);
 }
 *argc= 0;
 *argv= NULL;
 return(1);
}


int Sfile_make_argline(progname,argc,argv,line,flag)
char *progname;
int argc;
char **argv;
char **line;
int flag;
/*
 bit0= write progname as first argument
 bit1= write -cd $getwd() before writing arguments 
*/
{
 int i,l=0;
 char *cpt,*rpt,wdir[SfileadrL];

 if(flag&1) l= strlen(progname)+1;
 if(flag&2) if(getcwd(wdir,sizeof(wdir))) l+= strlen(wdir)+5;
 for(i=0;i<argc;i++) l+= strlen(argv[i])*3+1;
 cpt= *line= Smem_malloC(l+1);
 if(flag&1) {
   strcpy(cpt,progname);
   strcat(cpt," ");
   cpt+= strlen(cpt);
 }
 if(flag&2) {
   strcpy(cpt,"-cd ");
   strcat(cpt,wdir);
   strcat(cpt," ");
   cpt+= strlen(cpt);
 }
 for(i=0;i<argc;i++) {
   rpt= argv[i];
   while((*rpt)!=0){
     if(isspace(*rpt)){
       (*cpt++)= '"';
       (*cpt++)= *rpt;
       (*cpt++)= '"';
     } else if((*rpt)<32) {
       (*cpt++)= 27;
       (*cpt++)= *rpt;
     } else {
       (*cpt++)= *rpt;
     }
     rpt++;
   }
   if(i<argc-1) (*cpt++)= ' ';
 }
 *cpt= 0;
 return(1);
}


/* ------------------------------------------------------------------------- */


char *Sfile_datestr(tim,flag)
time_t tim;
short int flag; /* bit0=with hours+minutes  bit1=with seconds */
{
 static char zeitcode[80]={"000000"};
 char puff[80];
 struct tm *azt ;

 azt = gmtime(&tim);
 if(azt->tm_year>99)
   sprintf(zeitcode,"%c%01.1d%02.2d%02.2d",
           'A'+(azt->tm_year-100)/10,azt->tm_year%10,
           azt->tm_mon+1,azt->tm_mday);
 else
   sprintf(zeitcode,"%02.2d%02.2d%02.2d",
           azt->tm_year,azt->tm_mon+1,azt->tm_mday);
 if(flag&1){
   sprintf(puff,".%02.2d%02.2d",azt->tm_hour,azt->tm_min);
   strcat(zeitcode,puff);
 }
 if(flag&2){
   sprintf(puff,"%02.2d",azt->tm_sec);
   strcat(zeitcode,puff);
 }

 return(zeitcode);
}


int Sfile_rel_datestr(tim,rel_time,flag)
time_t tim;
char *rel_time;
short int flag;
/*
 bit0= use %6.lf (%9.lf in case of bit1)
 bit1= always in seconds 
 bit2= invert sign of difference
*/
{
 double diff,adiff;

 diff= ((double) tim)-((double) time(0));
 if(flag&4)
   diff= -diff;
 if(flag&2) {
   if(flag&1) 
     sprintf(rel_time,"%9.lfs",diff);
   else
     sprintf(rel_time,"%.lfs",diff);
 }
 if(diff<0) adiff= -diff;
 else       adiff= diff;
 if(adiff<7200) {
   if(flag&1) 
     sprintf(rel_time,"%6.lfs",diff);
   else
     sprintf(rel_time,"%.lfs",diff);
 } else if(adiff<2*86400) {
   if(flag&1) 
     sprintf(rel_time,"%6.lfh",diff/3600.0);
   else
     sprintf(rel_time,"%.lfh",diff/3600.0);
 } else if(adiff<3*7*86400) {
   if(flag&1) 
     sprintf(rel_time,"%6.lfd",diff/86400.0);
   else
     sprintf(rel_time,"%.lfd",diff/86400.0);
 } else {
   if(flag&1) 
     sprintf(rel_time,"%6.lfw",diff/(7*86400.0));
   else
     sprintf(rel_time,"%.lfw",diff/(7*86400.0));
 }
 return(1);
}


int Sfile_decode_date_input(reply,text,flag)
/* MMDDhhmm[[CC]YY[.ss]] */
struct tm *reply;
char *text;
int flag;
{
 int i,l,year;
 time_t current_time;
 struct tm *now;
 char line[1024];

 current_time= time(0);
 now= localtime(&current_time);
 for(i=0;i<sizeof(struct tm);i++)
   ((char *) reply)[i]= ((char *) now)[i];

 l= strlen(text);
 for(i=0;i<l;i++)
   if(text[i]<'0'||text[i]>'9')
     break;
 if(i!=8 && i!=10 && i!=12)
   return(0);
 if(text[i]==0)
   goto decode;
 if(text[i]!='.' || l!=15)
   return(0);
 i++;
 if(text[i]<'0'||text[i]>'9')
   return(0);
 i++;
 if(text[i]<'0'||text[i]>'9')
   return(0);

decode:;
 /* MMDDhhmm[[CC]YY][.ss]] */
 i= 0;
 reply->tm_mon=  10*(text[0]-'0')+text[1]-'0'-1;
 reply->tm_mday= 10*(text[2]-'0')+text[3]-'0';
 reply->tm_hour= 10*(text[4]-'0')+text[5]-'0';
 reply->tm_min=  10*(text[6]-'0')+text[7]-'0';
 reply->tm_sec= 0;
 if(l==8)
   return(1);
 if(l>10){
   year= 1000*(text[8]-'0')+100*(text[9]-'0')+10*(text[10]-'0')+(text[11]-'0');
 }else{
   year= 1900+10*(text[8]-'0')+(text[9]-'0');
   if(year<1970)
     year+= 100;
 }
 reply->tm_year= year-1900;
 if(l<=12)
   return(1);
 reply->tm_sec=  10*(text[13]-'0')+text[14]-'0';
 return(1);
}


int Sfile_decode_datestr(reply,text,flag)
/* YYMMDD[.hhmm[ss]] */
struct tm *reply;
char *text;
int flag;
{
 int i,l,year;
 time_t current_time;
 struct tm *now;
 char line[1024];

 current_time= time(0);
 now= localtime(&current_time);
 for(i=0;i<sizeof(struct tm);i++)
   ((char *) reply)[i]= ((char *) now)[i];

 if(text[0]<'0'|| (text[0]>'9' && text[0]<'A') || text[0]>'Z')
   return(0);
 l= strlen(text);
 for(i=1;i<l;i++)
   if(text[i]<'0'||text[i]>'9')
     break;
 if(i!=6)
   return(0);
 if(text[i]==0)
   goto decode;
 if(text[i]!='.' || (l!=11 && l!=13))
   return(0);
 for(i++;i<l;i++)
   if(text[i]<'0'||text[i]>'9')
     break;
 if(i!=l)
   return(0);

decode:;
 reply->tm_hour= 0;
 reply->tm_min= 0;
 reply->tm_sec= 0;
 i= 0;
 if(text[0]>='A') 
   reply->tm_year= 100+(text[i]-'A')*10+text[1]-'0';
 else
   reply->tm_year= 10*(text[0]-'0')+text[1]-'0';
 reply->tm_mon=  10*(text[2]-'0')+text[3]-'0'-1;
 reply->tm_mday= 10*(text[4]-'0')+text[5]-'0';
 if(l==6)
   return(1);
 reply->tm_hour= 10*(text[7]-'0')+text[8]-'0';
 reply->tm_min=  10*(text[9]-'0')+text[10]-'0';
 if(l==11)
   return(1);
 reply->tm_sec=  10*(text[11]-'0')+text[12]-'0';
 return(1);
}


int Sfile_decode_hpbasic(reply,text,flag)
/* DD Mth YYYY hh:mm:ss */
struct tm *reply;
char *text;
int flag;
{
#define SfilemaxdateleN 81
 char tday[SfilemaxdateleN],tmon[SfilemaxdateleN],tyear[SfilemaxdateleN];
 char ttime[SfilemaxdateleN],ttzone[SfilemaxdateleN];
 int i,year,mon,day,hour,min,sec,ret,tzr,sig,hcorr,mcorr;

 if(strlen(text)>=SfilemaxdateleN)
   return(0);
#undef SfilemaxdateleN
 
 ret= sscanf(text,"%s%s%s%s%s",tday,tmon,tyear,ttime,ttzone);
 if(ret<5){
   strcpy(ttzone,"+0000");
   ret= sscanf(text,"%s%s%s%s",tday,tmon,tyear,ttime);
   if(ret<4)
     return(0);
 }
 ret= sscanf(tyear,"%d",&year);
 if(ret<1)
   return(0);   

 for(i=0;tmon[i]!=0;i++) tmon[i]= islower(tmon[i])? toupper(tmon[i]): tmon[i];
 if(strcmp("JAN",tmon)==0) 
   mon= 1;
 else if(strcmp("FEB",tmon)==0)
   mon= 2;
 else if((strcmp("MAR",tmon)==0) || (strcmp("MRZ",tmon)==0))
   mon=3;
 else if(strcmp("APR",tmon)==0)
   mon=4;
 else if((strcmp("MAY",tmon)==0) || (strcmp("MAI",tmon)==0))
   mon=5;
 else if(strcmp("JUN",tmon)==0)
   mon=6;
 else if(strcmp("JUL",tmon)==0)
   mon=7;
 else if(strcmp("AUG",tmon)==0)
   mon=8;
 else if(strcmp("SEP",tmon)==0)
   mon=9;
 else if((strcmp("OCT",tmon)==0) || (strcmp("OKT",tmon)==0))
   mon=10;
 else if(strcmp("NOV",tmon)==0)
   mon=11;
 else if((strcmp("DEC",tmon)==0) || (strcmp("DEZ",tmon)==0))
   mon=12;
 else 
   return(0);

 ret= sscanf(tday,"%d",&day);
 if(ret<1)
   return(0);
 if(day<0 || day>31)
   return(0);

 if(ttime[2]!=':')
   return(0);
 if(ttime[5]!=':')
   return(0);
 ttime[2]= ttime[5]= 0;
 sscanf(ttime,"%ld",&hour);
 sscanf(ttime+3,"%ld",&min);
 sscanf(ttime+6,"%ld",&sec);
 if(hour<0||hour>23 || min<0||min>59 || sec<0||sec>59)
   return(0);
 
 tzr= 0;
 sig= 1;
 hcorr= 0;
 mcorr= 0;
 if(ttzone[tzr]=='-'){
   tzr++;
   sig= -1;
 }else if(ttzone[tzr]=='+') 
   tzr++;
 if(isdigit(ttzone[tzr]) && isdigit(ttzone[tzr+1])) {
   hcorr= 10*(ttzone[tzr]-'0')+ttzone[tzr+1]-'0';
   tzr+= 2;
   if(isdigit(ttzone[tzr]) && isdigit(ttzone[tzr+1])) 
     mcorr= 10*(ttzone[tzr]-'0')+ttzone[tzr+1]-'0';
 }
 hcorr*= sig;
 mcorr*= sig;

 reply->tm_year= year-1900;
 reply->tm_mon= mon-1;
 reply->tm_mday= day;
 reply->tm_hour= hour-hcorr;
 reply->tm_min= min-mcorr;
 reply->tm_sec= sec;
 return(1);
}


int Sfile_decode_newstime(reply,text,flag)
/* Wkd, DD Mth YYYY hh:mm:ss +/-ZZZZ  */ /* ZZZZ is ignored up to now */
struct tm *reply;
char *text;
int flag;
{
 int ret; 
 char *cpt;
 
 cpt= strchr(text,',');
 if(cpt==NULL) cpt= text;
 else          cpt++;
 ret= Sfile_decode_hpbasic(reply,cpt,0);
 return(ret);
}


time_t Sfile_mktime(inp_tm,flag)
struct tm *inp_tm;
int flag;
{
 struct tm loc_tm;
 double corr= 0.0;
 time_t ret;
 int i;

 for(i=0;i<sizeof(struct tm);i++)
   ((char *) &loc_tm)[i]= ((char *) inp_tm)[i];

 if(loc_tm.tm_sec<0) {
   corr+= loc_tm.tm_sec;
   loc_tm.tm_sec= 0;
 } else if (loc_tm.tm_sec>=60) {
   corr+= loc_tm.tm_sec-(loc_tm.tm_sec%60);
   loc_tm.tm_sec= loc_tm.tm_sec%60;
 }
 if(loc_tm.tm_min<0) {
   corr+= loc_tm.tm_min*60;
   loc_tm.tm_min= 0;
 } else if (loc_tm.tm_min>=60) {
   corr+= (loc_tm.tm_min-(loc_tm.tm_min%60))*60;
   loc_tm.tm_min= loc_tm.tm_min%60;
 }
 if(loc_tm.tm_hour<0) {
   corr+= loc_tm.tm_hour*60*24;
   loc_tm.tm_hour= 0;
 } else if (loc_tm.tm_hour>=24) {
   corr+= (loc_tm.tm_hour-(loc_tm.tm_hour%24))*60*24;
   loc_tm.tm_hour= loc_tm.tm_hour%24;
 }
 ret= mktime(&loc_tm);
 return(ret);
}


int Sfile_decode_time(code,valid,reply,flag)
char *code;
int *valid;
time_t *reply;
int flag;
{
 char *cpt,fakchr;
 double value,timepoint;
 struct tm res_tm;

 *valid= 0;
 *reply= 0;
 cpt= code;
 if(code[0]=='-' || code[0]=='+'){
   if(code[1]==0)
     return(1);
   if(code[1]>='0' && code[1]<='9') {
     timepoint= time(0);
     value= -1;
     fakchr= code[strlen(code)-1];
     if(isalpha(fakchr)) {
       fakchr= tolower(fakchr);
       code[strlen(code)-1]= 0;
     }
     sscanf(code,"%lf",&value);
     if     (fakchr=='s') timepoint+= 1.0*value;
     else if(fakchr=='h') timepoint+= 3600.0*value;
     else if(fakchr=='d') timepoint+= 86400.0*value;
     else if(fakchr=='w') timepoint+= 86400.0*7.0*value;
     else if(fakchr=='m') timepoint+= 86400.0*31.0*value;
     else if(fakchr=='y') timepoint+= 86400.0*(365.25*value+1.0);
     else                 timepoint+= 1.0*value;
     goto fertig;
   }
   cpt= code+1;
 } else if(Sfile_decode_datestr(&res_tm,code,0)>0) {
   /* YYMMDD[.hhmm[ss]] */
   timepoint= Sfile_mktime(&res_tm,0);
   goto fertig;
 } else if(Sfile_decode_date_input(&res_tm,code,0)>0) {
   /* MMDDhhmm[[CC]YY][.ss]] */
   timepoint= Sfile_mktime(&res_tm,0);
   goto fertig;
 } else if(Sfile_decode_hpbasic(&res_tm,code,0)>0) {
   /* DD Mth YYYY hh:mm:ss */
   timepoint= Sfile_mktime(&res_tm,0);
   goto fertig;
 } else if(Sfile_decode_newstime(&res_tm,code,0)>0) {
   /* Wkd, DD Mth YYYY hh:mm:ss +/-ZZZZ  */
   timepoint= Sfile_mktime(&res_tm,0);
   goto fertig;
 } else if(0) { /* ??? Mth dd hh:mm:ss Tzn YYYY */
   /* >>> */;
 }
 if(strcmp(cpt,"time")==0 || strcmp(cpt,"bigbang")==0 || strcmp(cpt,"off")==0)
   return(1);
 return(0);
fertig:;
 if(timepoint<0 || timepoint>=0x7FFFFFFF)
   return(0);
 *reply= timepoint;
 *valid= 1;
 return(1);
}


int Sfile_microsleep(microsec,flag)
/* sleep very fast by watching no file */
unsigned int microsec;
int flag;
{
 struct timeval wt;
 fd_set rds,wts,exs;
 int ready;

 wt.tv_sec=  microsec/1000000;
 wt.tv_usec= microsec%1000000;
 FD_ZERO(&rds);
 FD_ZERO(&wts);
 FD_ZERO(&exs);
 ready= select(1,&rds,&wts,&exs,&wt);
 if(ready>0)
   return(0); /* should never happen */
 return(1);
}


double Sfile_microtime(flag)
int flag;
{
 struct timeval tv;
 struct timezone tz;
 gettimeofday(&tv,&tz);
 return((double) (tv.tv_sec+1.0e-6*tv.tv_usec));
}


/* ------------------------------------------------------------------------- */

int Sfile_home_adr(name,fileadr,flag)
char *name,*fileadr;
int flag;
{
 char *home;

 strcpy(fileadr,name);
 home= getenv("HOME");
 if(home==NULL)
   return(0);
 strcpy(fileadr,home);
 if(name[0]!=0){
   strcat(fileadr,"/");
   strcat(fileadr,name);
 }
 return(1);
}


char *Sfile_esc_shell(in_text,out_text,flag)
char *in_text,*out_text;
int flag;
{
 int l,i,w=0;

 /* put everything in hard quotes */
 l= strlen(in_text);
 out_text[w++]= '\'';
 for(i=0;i<l;i++){
   if(in_text[i]=='\''){
     /* escape hard quotes */
     out_text[w++]= '\'';
     out_text[w++]= '"';
     out_text[w++]= '\'';
     out_text[w++]= '"';
     out_text[w++]= '\'';
   } else {
     out_text[w++]= in_text[i];
   }
 }
 out_text[w++]= '\'';
 out_text[w++]= 0;
 return(out_text);
}


int Sfile_print_fileinfo(adr,flag)
char *adr;
int flag;
/*
 bit0= print command text without executing
*/
{
 int ret;
 char line[SfileadrL],esc_file[SfileadrL],*fnpt;

 ret= Sfile_try_local_adr(adr,&fnpt,0);
 if(ret<0)
   return(ret);
 sprintf(line,"ls -l %s",Sfile_esc_shell(fnpt,esc_file,0));
 if(flag&1)
   printf("%s\n",line);
 else
   system(line);
 return(1);
}


int Sfile_popen_init(cmd,result,flag)
char *cmd;
struct Sfile_popeN *result;
int flag;
{
 FILE *fp;
 struct Sfile_linE {
   struct Sfile_linE *prev;
   struct Sfile_linE *next;
   char *text;
 };
 struct Sfile_linE *currentline,*freshline,*startline= NULL;
 char line[4096],*fret;
 int i,linecount=0,gesl,ret=0,l;

 /* Remove dead children if there are some */
 wait3(NULL,WNOHANG,NULL);

 strcpy(result->cmd,cmd);
 result->count= 0;
 result->lines= NULL;

 fp= popen(cmd,"r");
 if(fp==NULL){
   fprintf(stderr,"popen(\"%s\",\"r\") failed. errno=%ld",cmd,errno);
   return(0);
 }

 /* create a list of single lines */
 startline=  TSOB_FELD(struct Sfile_linE,1);
 startline->prev= startline->next= NULL;
 startline->text= NULL;
 /* read result lines */
 currentline= startline;
 while(1) {
   fret= Sfile_fgets(line,sizeof(line)-1,fp);
 if(fret==NULL) break;
   currentline->text= Smem_malloC(strlen(line)+1);
 if(currentline->text == NULL) break;
   strcpy(currentline->text,line);
   freshline= TSOB_FELD(struct Sfile_linE,1);
 if(freshline==NULL) break;
   currentline->next= freshline;
   freshline->prev= currentline;
   freshline->next= NULL;
   freshline->text= NULL;
   currentline= freshline;
   linecount++;
 }
 if(fret!=NULL) {
lack_of_memory:;
   fprintf(stderr,"out of memory while reading output of: %s\n",cmd);
   ret= -1;
   goto tmp_release;
 }

 pclose(fp);
 if(linecount==0) goto tmp_release;

 /* create result structure */
 result->count= linecount;
 result->lines= TSOB_FELD(char *,result->count);
 if(result->lines==NULL) goto lack_of_memory;
 for(i=0;i<result->count;i++)
   result->lines[i]= NULL;
 gesl= 0;
 currentline= startline;
 for(i=0;i<linecount;i++){
   l= strlen(currentline->text);
   result->lines[i]= TSOB_FELD(char,l+1);
   if(result->lines[i]==NULL) goto lack_of_memory;
   strcpy(result->lines[i],currentline->text);
   gesl+= l+1;
   currentline= currentline->next;
 }

 ret= 1;
tmp_release:;
 if(ret<=0)
    Sfile_popen_cleanup(result,0);
 currentline= startline;
 while(currentline!=NULL){
   if(currentline->text!=NULL) 
     Smem_freE(currentline->text);
   freshline= currentline->next;
   Smem_freE((char *) currentline);
   currentline= freshline;
 }
 return(ret);
}


int Sfile_popen_checkout(result,linec,linev,flag)
struct Sfile_popeN *result;
int *linec;
char ***linev;
int flag;
{
 *linev= result->lines;
 *linec= result->count;
 result->count= 0;
 result->lines= 0;
 return(1);
}


int Sfile_free_charvec(vec,count,flag)
char ***vec;
int count;
int flag;
{
 int i;

 if((*vec)!=NULL){
   for(i= 0;i<count;i++)
     if((*vec)[i]!=NULL) 
       Smem_freE((*vec)[i]);
   Smem_freE((char *) (*vec));
   *vec= NULL;
 }
 return(1);
}


int Sfile_popen_cleanup(result,flag)
struct Sfile_popeN *result;
int flag;
{
 int i;

 /* Remove dead children if there are some */
 wait3(NULL,WNOHANG,NULL);

 if(result==NULL) return(0);

 Sfile_free_charvec(&(result->lines),result->count,0);
 result->count= 0;
 return(1);
}


int Sfile_popen_resultstring(cmd,result,flag)
char *cmd;
char **result;
int flag;
/*
 bit0= omit last newline
*/
{
 struct Sfile_popeN pbuf;
 int i,l,ret;
 char *cbuf,*wpt;

 ret= Sfile_popen_init(cmd,&pbuf,0);
 l= 0;
 for(i=0;i<pbuf.count;i++)
   l+= strlen(pbuf.lines[i])+1;
 cbuf= Smem_malloC(l+1);
 if(cbuf==NULL)
   return(-1);
 wpt= cbuf;
 for(i=0;i<pbuf.count;i++){
   strcpy(wpt,pbuf.lines[i]);
   wpt+= strlen(pbuf.lines[i]);
   if(i+1<pbuf.count || !(flag&1))
     *(wpt++)= '\n';
 }
 *(wpt)= 0;
 Sfile_popen_cleanup(&pbuf,0);
 *result= cbuf;
 return(1);
}


int Sfile_ls(dir,filec,filev,flag)
char *dir;
int *filec;
char ***filev;
int flag;
/*
 bit0= ls option -d 
 bit1= just release list filev and return
 bit2= redirect stderr to /dev/null
 bit3= use   ls | grep 'dir'   rather than   ls dir
*/
{
 int ret;
 char cmd[SfileadrL];
 struct Sfile_popeN *result,result_data;
 
 result= &result_data;

 if(*filec>0 && *filev!=NULL){
   /* release previously created result structure */
   result_data.count= *filec;
   result_data.lines= *filev;
   Sfile_popen_cleanup(result,0);
   *filec= 0;
   *filev= NULL;
 }
 *filec= 0;
 if(flag&2)
   return(1);

 if(flag&8) {
   if(flag&1)
     sprintf(cmd,"ls -1ad | grep '%s'",dir);
   else
     sprintf(cmd,"ls -1a | grep '%s'",dir);
 } else {
   if(flag&1)
     sprintf(cmd,"ls -1ad %s",dir);
   else
     sprintf(cmd,"ls -1a %s",dir);
 }
 if(flag&4)
   strcat(cmd," 2>/dev/null");
 ret= Sfile_popen_init(cmd,result,0);
 if(ret<=0)
   return(ret);
 *filec= result->count;
 *filev= result->lines;
 return(1);
}


int Sfile_date(ret_date,flag)
char *ret_date;
int flag;
/*
 bit0= shorter format , GMT
*/
{
 int ret;
 char cmd[321],start_lang[SfileadrL];
 struct Sfile_popeN *result,result_data;

 strcpy(start_lang,getenv("LANG"));
 setenv("LANG","english",1);
 result= &result_data;
 ret_date[0]= 0;
 if(flag&1)
   strcpy(cmd,"date -u +'%d %b %Y %T'");
 else
   strcpy(cmd,"date +'%a, %d %b %Y %T %Z'");
 ret= Sfile_popen_init(cmd,result,0);
 if(ret<=0)
   goto ex;
 if(result->count<=0)
   {ret= 0; goto ex;}
 strcpy(ret_date,result->lines[0]);
 Sfile_popen_cleanup(result,0);
 ret= 1;
ex:;
 setenv("LANG",start_lang,1);
 return(ret);
}


int Sfile_read_var(line,name,content,flag)
char *line;
char **name;
char **content;
int flag;
/*
flag:
 bit2= do not interpret delimiter characters
 bit3= remove trailing blanks from content
 bit4= only look for delimiter '='
return:
 -1=error , >=0 number of variables read
*/
{
 int skip= 0;
 char *spt,*ept,*delimiters;
 static char std_delimiters[]= {"=,;:"},equal_delimiter[]= {"="};

 /* look for non-white */
 spt= line;
 while(*spt!=0 && isspace(*spt))
   spt++;
 /* if empty or # : return, do not read */
 if(*spt==0 || *spt=='#')
   return(0);
 /* look for white or delimiters */
 delimiters= std_delimiters;
 if(flag&16)
   delimiters= equal_delimiter;
 ept= spt;
 while(*ept!=0 && !isspace(*ept) 
       && ((flag&4) || strchr(delimiters,*ept)==NULL) )
   ept++;
 /* read name */
 if(Sregex_string_cut(name,spt,ept-spt,0)<=0)
   return(-1);
 if(*ept==0) {
   if(Sregex_string_cut(content,"",0)<=0)
     return(-1);
 } else {
   if(strchr(delimiters,*ept)!=NULL)
     skip= 1;
 }
 if(skip==0) {
   /* if not at "=,;:" : look for non-white or ("=,;:" and skip) */
   spt= ept+1;
   while(*spt!=0 && isspace(*spt) && strchr(delimiters,*spt)==NULL)
     spt++;
   if(*spt!=0)
     if(strchr(delimiters,*spt)!=NULL) 
       spt++;
 } else 
   spt= ept+skip;
 /* read remainder of line as content */
 if(Sregex_string(content,spt,0)<=0)
   return(-1);
 if(flag&3) {
   for(spt= *content+strlen(*content)-1;spt>=*content;spt--) {
     if(!isspace(*spt))
   break;
     *spt= 0;    
   }
 }
 return(1);
}


int Sfile_read_vars(fileadr,names,contents,count,flag)
char *fileadr;
char ***names; /* names and contents have to point to NULL or valid charvecs */
char ***contents;
int *count;
int flag;
/*
 bit2= for Sfile_read_var() : do not interpret delimiter characters
 bit15= release names and contents
*/
{
 int ret,widx,i;
 char line[2*SfileadrL];
 FILE *fp= NULL;

 if(*count>0) {
   if(*names!=NULL) 
     Sfile_free_charvec(names,*count,0);
   if(*contents!=NULL) 
     Sfile_free_charvec(contents,*count,0);
   *count= 0;
 }
 if(flag&(1<<15))
   {ret= 1; goto ex;}
     
 ret= Sfile_line_count(fileadr,1);
 if(ret<0)
   goto ex;
 *count= ret;
 if(ret==0)
   {ret= 1; goto ex;}

 *names= TSOB_FELD(char *,*count);
 if(*names==NULL)
   {ret= -1; goto ex;}
 for(i=0;i<*count;i++)
   (*names)[i]= NULL;
 *contents= TSOB_FELD(char *,ret);
 if(*contents==NULL)
   {ret= -1; goto ex;}
 for(i=0;i<*count;i++)
   (*contents)[i]= NULL;
 
 fp= Sfile_fopen(fileadr,"r",0);
 if(fp==NULL)
   {ret= 0; goto ex;}
 widx= 0;
 while(widx<*count) {
   if(Sfile_fgets(line,sizeof(line)-1,fp)==NULL)
 break;
   ret= Sfile_read_var(line,(*names)+widx,(*contents)+widx,flag&4);
   if(ret<0)
     goto ex;
   widx+= ret;
 }
 *count= widx;

 ret= 1;
ex:
 if(fp!=NULL)
   Sfile_fclose(fp);
 if(ret<=0) {
   *count= 1;
   Sfile_read_vars(fileadr,names,contents,count,1<<15);
 }
 return(ret);
}

int Sfile_find_name(search,names,count,flag)
char *search;
char **names;
int count;
int flag;
{
 int i;

 for(i=0;i<count;i++)
   if(names[i]!=NULL)
     if(strcmp(names[i],search)==0)
       return(i+1);
 return(0);
}


int Sfile_find_var(search,names,contents,count,result,result_size,flag)
char *search;
char **names;
char **contents;
int count;
union {char **ptpt; char *pt;} result;
int result_size;
int flag;
/*
 bit0= use result as target for Sregex_string (else strcpy)
 bit1= do not make result empty
*/
{
 int ret,l,idx;

 if(!(flag&2)) {
   if(flag&1)
     Sregex_string(result.ptpt,"",0);
   else 
     (result.pt)[0]= 0;
 }
 ret= Sfile_find_name(search,names,count,0);
 if(ret<=0) {
   /* >>> name not found */;
   return(0);
 }
 idx= ret-1;
 l= strlen(contents[idx]);
 if(l>result_size) {
   /* >>> result buffer overflow */; 
   return(0);
 }
 if(flag&1) {
   if(Sregex_string(result.ptpt,contents[idx],0)<=0)
     return(-1);
 } else {
   strcpy(result.pt,contents[idx]);
 }
 return(1);
}


int Sfile_find_var_direct(fp,searchlist,name,content,flag)
FILE *fp;
char **searchlist;
char **name,**content;          /* must be NULL or point to freeable address */
int flag;
/*
 bit0= return next valid name,content pair without refering to searchlist
 bit2= for Sfile_read_var() : do not interpret delimiter characters
 bit3= for Sfile_read_var() : remove trailing whitespace
*/
{
 int ret,i,l;
 char line[4*SfileadrL];

 while(1) {
   if(Sfile_fgets(line,sizeof(line)-1,fp)==NULL)
 break;
   ret= Sfile_read_var(line,name,content,flag&(8|4));
   if(ret<0)
     {ret= 0; goto ex;}
   if(ret==0)
 continue;
   /* check wether this line is to be extended by the next one */
   if((*content)[0]!=0) {
     l= strlen(*content);
     while((*content)[l-1]=='\\') {
       if(l>1)
         if((*content)[l-2]=='\\') {
           (*content)[l-1]= 0;
     break;    
         }
       if(Sfile_fgets(line,sizeof(line)-1,fp)==NULL)
     break;
       else {
         (*content)[l-1]= 0;
         if(Sregex_string(content,line,1)<=0)
           {ret= -1; goto ex;}
       }
     }
   }
   if(flag&1)
     {ret= 1; goto ex;}
   for(i=0;searchlist[i]!=NULL;i++) {
     if(strcmp(searchlist[i],*name)==0)
       {ret= 1; goto ex;}
   }
 }
 ret= 0;
ex:;
 return(ret);
}

/* ------------------------------------------------------------------------- */

