
#ifndef SfileadrL

/* Maximum Length of a file address */
#define SfileadrL 4096

struct Sfile_popeN {
 char cmd[SfileadrL];
 int count;
 char **lines;
};

struct Sfile_linE {
 struct Sfile_linE *prev;
 struct Sfile_linE *next;
 char *text;
};

char *Sfile_fgets();
FILE *Sfile_fopen();
char *Sfile_datestr();
double Sfile_microtime();

/* characters below _loW and above _higH (both including) get escaped */

#define Sfile_bad_name_loW      32
#define Sfile_bad_name_higH    128

/* these single characters get also escaped */

#define Sfile_bad_name_charS   "`~!@#$%^&*()|\\}]{[:;\"'?><,"

/* name of the file which allows Sfile_remove_old_files() */
#define Sfile_autoremove_olD "sfile_remove_old_files_allowed"


/*
                     Collection of misc helper routines 

  File system:

    FILE *Sfile_fopen(char *filename,char *mode,int flag)
        open a file like fopen(2) , 
        a special name is "-" : depending on mode it is stdin or stdout

    int Sfile_fclose(FILE *fp)
        close a file opened by Sfile_fopen()

    char *Sfile_fgets(char *line,int maxl,FILE *fp)
        read a line from a file, remove newline or carriage-return-newline
        from end of the line

    int Sfile_identical(char *file1,char *file2,int flag)
        test wether the content of two files is identical
        or test wether the names depict the same file object (i-node)

    int Sfile_string_to_file(char *filename,char *text,int flag)
        create a file and write text to it

    int Sfile_file_to_text(char *filename,char *text,int textsize,int flag)
        read the first text line from a file

    int Sfile_file_to_double(char *filename,double *num,int flag)
        read the first number from the first text line of a file.
        
    int Sfile_file_to_int(char *filename,int *num,int flag)
        read the first number from the first text line of a file
        if it is in the range of 32 bit.
   
    int Sfile_remove_old_files(int filec,char **filev,time_t agelimit,int flag)
        remove all files listed in filev[] which are older than agelimit
        since this function is very dangerous it is restricted to relative
        adresses in a directory where the file "sfile_remove_old_files_allowed"
        exists.

    int Sfile_change_file_lines(char *fileadr,char *search,int search_l,
                                char *subst_line,int *errn,char *errmsg,
                                int flag)
        filter the file given by filadr. Replace lines matching 'search' by
        subst_line. Preserve the files m_time and a_time.
        Report errors in *errn and errmsg.

    int Sfile_copy_file(char *source,char *target,
                        int *errn,char *errmsg,int flag)
        copy the content of source to target. 
        Report errors in *errn and errmsg.


  File names :

    int Sfile_esc_filename(char *name,int namsiz,int flag)
        replace bad characters in name by harmless ones
        caution: can make the name longer
         
    int Sfile_hashcode(char *name,int hashsize,int flag)
        calculates a hashnumber from filename
        useful for automatically distributing filenames over 'hashsize' many
        directories.


  Locking :

    int Sfile_get_lock(char *adr,unsigned int microsec,char *confirmation,
                       int flag)
        creates a lockfile if not already existing. Waits for microsec and
        checks wether a race condition has occured. If not, the confirmation
        text is appended to the lockfile's content and the lock becomes valid.


  Text parser :

    int Sfile_make_argv(char *progname,char *line,
                        int *argc,char **argv,int flag)
        split a text line into tokens and put them into an argument structure
        like for main() programs. Also release this structure if no longer
        needed.

    int Sfile_make_argline(char *progname,char *line,
                        int *argc,char **argv,int flag)
        compose a text line from tokens in a way that Sfile_make_argv() can 
        split it to the same list of tokens again.
  
  
  Time and date :

    char *Sfile_datestr(time_t tim,int flag)
        convert the time tim into a string of a system where chronological
        order matches alphabetical order: YYMMDD[.hhmm[ss]]

    int Sfile_decode_date_input(struct tm *reply,char *text,int flag)
        try to read a date definition from text in format MMDDhhmm[[CC]YY[.ss]]

    int Sfile_decode_datestr(reply,text,flag)
        try to read a date definition from text in format YYMMDD[.hhmm[ss]]

    int Sfile_decode_time(char i*code,int *valid,time_t *reply,int flag)
        read a time and date definition from code in various formats:
          datestr  : YYMMDD[.hhmm[ss]]
          date(1)  : MMDDhhmm[[CC]YY[.ss]]
          relative : -/+${number}[c] where c is one of {s,h,d,w,m,y}
                                     for second, hour, day, week, month, year
                                     relative to current time.

    int Sfile_microsleep(unsigned int microsec,int flag)
        sleep a number of microseconds (1000000 microseconds = 1 second)
        this function depends on the time granularity of select(2)


  Shell :

    int Sfile_home_adr(char *name,char *fileadr,int flag)
        get a directory name from the shell variable $HOME and compose
        fileadr from name and the directory.
        
    int Sfile_popen_init(char cmd,struct Sfile_popeN *result,int flag)
        execute cmd via popen(3) and store the standard output lines in result
        
    int Sfile_popen_cleanup(struct Sfile_popeN *result,int flag)
        release the memory related to result

    int Sfile_ls(char *dir,int *filec,char ***filev,intflag)
        get the standard output of  ls -1a ${dir}
        or release memory related to filec,filev

*/

#endif /* ! SfileadrL */

